////////////////////////////
//common.h
//stores common definitions, data types, prototypes
//throughout compiler

#define TARGET_I86

//special characters
//carriage return
#define CR	0x0D
//line feed
#define LF	0x0A
//tab
#define TAB	0x09

//symbol table size
#define SYMBOL_TABLE_HASH_SIZE 16384
//SYMBOL_TABLE_HASH_AND should always be SYMBOL_TABLE_HASH_SIZE-1
#define SYMBOL_TABLE_HASH_AND (16384-1)

//32 bit version
#define WORDSIZE 4
typedef long ATTRIB_ADDR;

//max size of in-code array initialization
#define MAXARRAYINITSIZE 512

#define MAXPARAMETERS 64

enum TypeSizes
{
	SIZE_CHAR		= 1,
	SIZE_SHORT		= 2,
	SIZE_LONG		= 4,
	SIZE_FLOAT		= 4,
	SIZE_DOUBLE		= 8,
	SIZE_MAXDATA	= 8,		//maximum size of all types

	SIZE_RPC		= 8,		//remote procedure call
};

#define NULL 0

template <class T>
T max(T x, T y)
{
	return (x > y ? x : y);
}

template <class T>
T min(T x, T y)
{
	return (x < y ? x : y);
}

#include <string.h>
#include <stdlib.h>

#include "token.h"
#include "compiler.h"


struct ASTFuncCall;	//prototype for ASTNode declaration

//node for abstract syntax tree
//unused children are NULL
//if op is an immediate (| T_IMMEDIATE), then child[0] is a pointer to immediate constant
//if T_IMMEDIATESTRING, can get by char *t = ((char **)curnode->data)[0];
struct ASTNode
{
	TOKEN_ID op;
	ATTRIB_TYPE type;
	long defnum;	//zero for all system stuff.  used for variabless
	unsigned char data[SIZE_MAXDATA];	//for data of node (immediates, etc.)
	ASTNode *parent;
	ASTNode *child[4];	//use 4 for  things like if for loop
	ASTFuncCall *funccall;	//used for function calls... put in all ASTNodes for future use (opperator overloading)
};

//node for function calling
//if user-type parameter, takes up two parameters, one for A_USERTYPE, and one for ID of user type -useful incase arrays of user type
//previous line needs to be updated...
struct ASTFuncCall
{
	TOKEN_ID op;	//id of function
	long defnum;	//defnum of function
	int numparameters;
	ASTNode *parameters[MAXPARAMETERS];	//MAXPARAMETERS is max number of parameters
};

//if operand is constant, then operand is set with T_IMMEDIATE flag, and target is pointer to data
//id = data type or variable name
//defnum = name unique number or address if address (***also redundantly stored in data)
//data = pointer to immediate data that it contains
//pointer = is it a pointer?
struct MIRInstr
{
	TOKEN_ID instr;

	//destination of instruction results (dest_id is relational operator for if statements)
	TOKEN_ID dest_id;
	unsigned long dest_defnum;
	bool dest_pointer;

	//operands of instruction (and data *'s for constants)
	TOKEN_ID op1_id;
	unsigned long op1_defnum;
	void *op1_data;
	bool op1_pointer;	//set to true if pointer to data

	TOKEN_ID op2_id;
	unsigned long op2_defnum;
	void *op2_data;
	bool op2_pointer;

	long target;	//used for jumps (if, goto, etc.), and wait_id's (so parallel calls can find their return value)

	MIRInstr *next;
	MIRInstr *prev;

};


//used for getting types, etc.
long getPrecedence(TOKEN_ID id);
ATTRIB_TYPE getGeneralType(ATTRIB_TYPE a, ATTRIB_TYPE b);

#include "tables.h"
#include "scanner.h"
#include "constrainer.h"
#include "parser.h"
#include "mir_transform.h"
#include "optimizer.h"

//target anchitecutre
#ifdef TARGET_I86
#include "i86codegenerator.h"
#endif

