#include "common.h"

#include <stdio.h>

int main(int argc, char* argv[])
{
	//check for parameters
	if(argc == 1)
	{
		printf("\nUsage: compiler textfile.txt flags\n");
		return 0;
	}

	char *SourceName = argv[1];
	char *Target = argv[2];

	// test stuff
	StringTable stringtable;
	SymbolTable symboltable;

	Parser parser(SourceName, &stringtable, &symboltable);
	bool okforcompile = parser.Parse();
	if(!okforcompile)
		return 1;

	//transform to intermediate form
	MIRTransform mirtrans(&stringtable, &symboltable);
	Optimizer o(&symboltable);
	CodeGenerator_i86 codegen(&stringtable, &symboltable);

	//set up functions for remote calling
	codegen.EnumerateFunctions();

	FILE *MIRout, *ASMout;
	MIRout = fopen("MIRoutputtest.txt", "w");

	FunctionType *curfunc = symboltable.getFirstFunction();
	while(curfunc != NULL)
	{

		//*** enable optimizations only by commandline flags
		o.ConstantFolding(curfunc->function);

		mirtrans.Transform(curfunc);

		o.ConstantFolding(curfunc->mircode);
		o.RemoveRedundantGOTOs(curfunc->mircode);

		mirtrans.PrintListing(MIRout, curfunc->mircode);

		codegen.GenerateCode(curfunc);

		curfunc = symboltable.getNextFunction();
	}
	fclose(MIRout);

	//for remote calling
	codegen.GenerateFunctionEnumeration();

	//peephole optimizing here
	codegen.PeepholeOptimization();

	ASMout = fopen("ASMtest.txt", "w");
	codegen.PrintListing(ASMout);
	fclose(ASMout);

	//pause
	char a;
	scanf("%c", &a);

	return 0;
}

