

///////////////////////////////////
//INSTRUCTION OPERANDS

//size flags
#define iDOUBLEWORD 1
#define iWORD 2
#define	iBYTE 4

#define OPERAND_i86 long

enum OPERANDS_i86
{
	//REGISTERS
	iNONE	=	0x0,

	iREGISTER = 0xF0,	//mask for register checking

	//format: number, partial (1 = full double word, 2 = lower word, 4 = lowest byte, (C = upper byte of lower word) )
	iA		=	0x10,	//register A
	iEAX	=	0x11,
	iAX		=	0x12,
	iAL		=	0x14,
	iAH		=	0x1C,

	iB		=	0x20,	//register B
	iEBX	=	0x21,
	iBX		=	0x22,
	iBL		=	0x24,
	iBH		=	0x2C,

	iC		=	0x30,	//register C
	iECX	=	0x31,
	iCX		=	0x32,
	iCL		=	0x34,
	iCH		=	0x3C,

	iD		=	0x40,	//register D
	iEDX	=	0x41,
	iDX		=	0x42,
	iDL		=	0x44,
	iDH		=	0x4C,

	iESI	=	0x51,
	iSI		=	0x52,

	iEDI	=	0x61,
	iDI		=	0x62,

	iESP	=	0x71,
	iSP		=	0x72,

	iEBP	=	0x81,
	iBP		=	0x82,

	iCS		=	0x90,
	iDS		=	0xA0,
	iES		=	0xB0,
	iFS		=	0xC0,
	iGS		=	0xD0,

	//can't use 2nd lowest byte (for registers) -must be 0

	//MEMORY
	iMEM	=	0x100,
	iMEMD	=	0x101,	//double word   offset(offset)
	iMEMW	=	0x102,	//word			offset(offset)
	iMEMB	=	0x104,	//byte			offset(offset)

	iIMM	=	0x300,
	iIMMD	=	0x301,	//double word (l_immediate)
	iIMMW	=	0x302,	//word			(l_immediate)
	iIMMB	=	0x304,	//byte			(l_immediate)

	//COMPLEX ADDRESSING MODES
	iROFF	=	0x400,
	iROFFD	=	0x401,	//register(opb) + offset(offset)
	iROFFW	=	0x402,	//register(opb) + offset(offset)
	iROFFB	=	0x404,	//register(opb) + offset(offset)

	iCOMP	=	0x500,
	iCOMPD	=	0x501,	//reg1(opb) + reg2(opc) * mult(multiplier) + offset(offset)
	iCOMPW	=	0x502,	//reg1(opb) + reg2(opc) * mult(multiplier) + offset(offset)
	iCOMPB	=	0x504,	//reg1(opb) + reg2(opc) * mult(multiplier) + offset(offset)

	iGLOBAL	 =	0x600,
	iGLOBALD =	0x601,	//func_id and func_defnum are the id and defnum of global
	iGLOBALW =	0x602,
	iGLOBALB =	0x604,


};

///////////////////////////////
//INSTRUCTIONS

enum INSTRUCTIONS_I86
{

//INTEGER
	//iNONE	=	0x0,	//used mostly for ignoring -already defined in registers
	iNOP	=	0x1,
	iMOV	=	0x2,
	iPUSH	=	0x3,
	iPOP	=	0x4,
	iPUSHF	=	0x5,
	iPOPF	=	0x6,
	iPUSHA	=	0x7,
	iPOPA	=	0x8,
	iPUSHAD	=	0x9,
	iPOPAD	=	0xA,
	iLAHF	=	0xB,
	iSAHF	=	0xC,
	iMOVSX	=	0xD,
	iMOVZX	=	0xE,
	iLEA	=	0xF,
	iLDS	=	0x10,
	iLES	=	0x11,
	iLFS	=	0x12,
	iLGS	=	0x13,
	iLSS	=	0x14,
	iADD	=	0x15,
	iSUB	=	0x16,
	iAND	=	0x17,
	iOR		=	0x18,
	iXOR	=	0x19,
	iADC	=	0x1A,
	iSBB	=	0x1B,
	iCMP	=	0x1C,
	iTEST	=	0x1D,
	iINC	=	0x1E,
	iDEC	=	0x1F,
	iNEG	=	0x20,
	iNOT	=	0x21,
	iMUL	=	0x22,
	iIMUL	=	0x23,
	iDIV	=	0x24,
	iIDIV	=	0x25,
	iCBW	=	0x26,
	iCWDE	=	0x27,
	iCWD	=	0x28,
	iCDQ	=	0x29,
	iSHR	=	0x2A,
	iSHL	=	0x2B,
	iSAR	=	0x2C,
	iSAL	=	0x2D,
	iSHLD	=	0x2E,
	iSHRD	=	0x2F,
	
	iSETC	=	0x30,
	iSETNC	=	0x31,
	iSETZ	=	0x32,
	iSETNZ	=	0x33,
	iSETS	=	0x34,
	iSETNS	=	0x35,
	iSETO	=	0x36,
	iSETNO	=	0x37,
	iSETP	=	0x38,
	iSETPE	=	0x39,
	iSETNP	=	0x3A,
	iSETPO	=	0x3B,
	iSETA	=	0x3C,
	iSETNBE	=	0x3D,
	iSETAE	=	0x3E,
	iSETNB	=	0x3F,
	iSETB	=	0x40,
	iSETNAE	=	0x41,
	iSETBE	=	0x42,
	iSETNA	=	0x43,
	iSETE	=	0x46,
	iSETNE	=	0x47,
	iSETG	=	0x48,
	iSETNLE	=	0x49,
	iSETGE	=	0x4A,
	iSETNL	=	0x4B,
	iSETL	=	0x4C,
	iSETNGE	=	0x4D,
	iSETLE	=	0x4E,
	iSETNG	=	0x50,

	iCALL	=	0x54,

	iJMP	=	0x55,
	iJC		=	0x56,
	iJNC	=	0x57,
	iJZ		=	0x58,
	iJNZ	=	0x59,
	iJS		=	0x5A,
	iJNS	=	0x5B,
	iJO		=	0x5C,
	iJNO	=	0x5D,
	iJP		=	0x5E,
	iJPE	=	0x5F,
	iJNP	=	0x60,
	iJPO	=	0x61,
	iJA		=	0x62,
	iJNBE	=	0x63,
	iJAE	=	0x64,
	iJNB	=	0x65,
	iJB		=	0x66,
	iJNAE	=	0x67,
	iJBE	=	0x68,
	iJNA	=	0x69,
	iJE		=	0x6C,
	iJNE	=	0x6D,
	iJG		=	0x6E,
	iJNLE	=	0x6F,
	iJGE	=	0x70,
	iJNL	=	0x71,
	iJL		=	0x72,
	iJNGE	=	0x73,
	iJLE	=	0x74,
	iJNG	=	0x75,

	iRET	=	0x79,

	iLABEL	=	0x80,	// 'instruction' to mark labels
	iSTRING	=	0x90,	// 'instruction' to put text into assembly (i.e. declarations, etc.)

/* to be added later...?
	LODS
	REP LODS
	STOS
	REP STOS
	MOVS
	REP MOVS
	SCAS
	REP(N)E SCAS
	CMPS
	REP(N)E CMPS
	CPUID
	RDTSC
*/


//FLOATING POINT
/*	FLD                  r/m32/m64         1         +              0      0
	FLD                  m80               3         np             0      0
	FBLD                 m80           48-58         np             0      0
	FST(P)               r                 1         np             0      0
	FST(P)               m32/m64           2 m)      np             0      0
	FST(P)               m80               3 m)      np             0      0
	FBSTP                m80         148-154         np             0      0
	FILD                 m                 3         np             2      2
	FIST(P)              m                 6         np             0      0
	FLDZ FLD1                              2         np             0      0
	FLDPI FLDL2E etc.                      5         np             0      0
	FNSTSW               AX/m16            6 q)      np             0      0
	FLDCW                m16               8         np             0      0
	FNSTCW               m16               2         np             0      0

	FADD(P)              r/m               3         +              2      2
	FSUB(R)(P)           r/m               3         +              2      2
	FMUL(P)              r/m               3         +              2      2 n)
	FDIV(R)(P)           r/m        19/33/39 p)      +             38 o)   2
	FCHS FABS                              1         +              0      0
	FCOM(P)(P) FUCOM     r/m               1         +              0      0
	FIADD FISUB(R)       m                 6         np             2      2
	FIMUL                m                 6         np             2      2
	FIDIV(R)             m          22/36/42 p)      np            38 o)   2
	FICOM                m                 4         np             0      0
	FTST                                   1         np             0      0
	FXAM                               17-21         np             4      0
	FPREM                              16-64         np             2      2
	FPREM1                             20-70         np             2      2
	FRNDINT                             9-20         np             0      0
	FSCALE                             20-32         np             5      0
	FXTRACT                            12-66         np             0      0

	FSQRT                                 70         np            69 o)   2
	FSIN FCOS                         16-126         np             2      2
	FSINCOS                           17-137         np             2      2
	F2XM1                              13-57         np             2      2
	FYL2X                             22-111         np             2      2
	FYL2XP1                           22-103         np             2      2
	FPATAN                            19-134         np             2      2
	FPTAN                             17-173         np            36 o)   0

	FNOP                                   2         np             0      0
	FXCH                 r                 1         np             0      0
	FINCSTP FDECSTP                        2         np             0      0
	FFREE                r                 2         np             0      0
	FNCLEX                               6-9         np             0      0
	FNINIT                             12-22         np             0      0
	FNSAVE               m           124-300         np             0      0
	FRSTOR               m             70-95         np             0      0
	WAIT                                   1         np             0      0
*/

};



struct i86instruct
{
	long instr;

	OPERAND_i86 op1a;
	OPERAND_i86 op1b;
	OPERAND_i86 op1c;

	OPERAND_i86 op2a;
	OPERAND_i86 op2b;
	OPERAND_i86 op2c;

	TOKEN_ID func_id;	//used for function calls
	long func_defnum;

	long offset;	//used for addressing
	long multiplier;	//used for complex addressing
	long l_immediate;	//used for immediate values and complex addressing (multiplier)

	char *string;	//for text markers (i.e. global, func, declarations)

	i86instruct *next;
	i86instruct *prev;
};


const char call_function_enumerator_code1[]
   = "  .align 4\n.globl call_enum_func__FllPl\n  .type call_enum_func__FllPl,@function\ncall_enum_func__FllPl:\n";
const char call_function_enumerator_code2[]
   = "  pushl %ebp\n  movl %esp, %ebp\n  subl $16, %esp\n  pushl %ecx\n movl 16(%ebp), %eax #*data\n  movl 12(%ebp), %ecx #datalen\n";
const char call_function_enumerator_code3[]
   = ".STACKCPY:\n  subl $1, %ecx\n  jl .STACKCPYEND\n  pushl (%eax,%ecx,4)\n  jmp .STACKCPY\n.STACKCPYEND:\n  movl 8(%ebp), %eax\n";
const char call_function_enumerator_code4[]
   = "  shll $2,%eax\n  movl func_enums(%eax),%ecx\n  call *%ecx\n  movl 12(%ebp),%ecx #datalen\n  shll $2, %ecx	#times 4 for long\n";
const char call_function_enumerator_code5[]
   = "  addl %ecx, %esp\n  popl %ecx\n  addl $16, %esp\n  movl %ebp, %esp\n  popl %ebp\n  ret\n.Lfecall_enum_func__FllPl:\n";
const char call_function_enumerator_code6[]
   = "  .size call_enum_func__FllPl,Lfecall_enum_func__FllPl-call_enum_func__FllPl\n";


