#include "common.h"

CodeGenerator_i86::CodeGenerator_i86(StringTable *st, SymbolTable *yt)
{
	stringtable = st;
	symboltable = yt;
	icode = NULL;

	//start from beginning
	parameter_stack_depth = 0;

	//start at beginning of functions
	num_enumerated_functions = 0;

	//put in data segment
	icode = new i86instruct;
	icode->instr = iSTRING;
	icode->next = NULL;
	icode->string = new char[16];
	strcpy(icode->string, ".section .data");

	//make current pointer, so icode is left as root node
	icur = icode;

	//////////////////////////
	//write out global variables
	symboltable->resetGlobalSelection();
	while(symboltable->selectNextGlobal())
	{
		TOKEN_ID g_id = symboltable->getCurGlobalID();
		long g_defnum = symboltable->getCurGlobalDefnum();

		TOKEN_ID g_type = symboltable->getType(g_id, g_defnum);
		if(g_type == T_CLASS || g_type == T_PCLASS)
			continue;

		//make sure not declared as external
		if(g_type & A_EXTERNAL)
			continue;

		//get variable name
		char *g_name = stringtable->getString(symboltable->getNameLookup(g_id, g_defnum));

		//allow variable to be exportable
		icur->next = new i86instruct;
		icur = icur->next;
		icur->instr = iSTRING;
		icur->next = NULL;
		icur->string = new char[256];
		sprintf(icur->string, "\n.globl %s", g_name);

		//alignment information (can make more complex and efficient)
		icur->next = new i86instruct;
		icur = icur->next;
		icur->instr = iSTRING;
		icur->next = NULL;
		icur->string = new char[256];
		strcpy(icur->string, ".align 4");

		//set type
		icur->next = new i86instruct;
		icur = icur->next;
		icur->instr = iSTRING;
		icur->next = NULL;
		icur->string = new char[256];
		sprintf(icur->string, "  .type %s,@object", g_name);
		//set size
		icur->next = new i86instruct;
		icur = icur->next;
		icur->instr = iSTRING;
		icur->next = NULL;
		icur->string = new char[256];
		sprintf(icur->string, "  .size %s,%d", g_name, symboltable->getSize(g_id, g_defnum));

		//make label
		icur->next = new i86instruct;
		icur = icur->next;
		icur->instr = iSTRING;
		icur->next = NULL;
		icur->string = new char[256];
		sprintf(icur->string, "%s:", g_name);

		//check what type (so know what to put)
		if( (g_type & A_ARRAY) && !(g_type & A_RESIZEABLE))
		{
			if(g_type & A_USERTYPE)
			{
				//create initializer/space saver
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iSTRING;
				icur->next = NULL;
				icur->string = new char[256];
				sprintf(icur->string, "  .zero %d", symboltable->getSize(g_id, g_defnum));
			}
			else
			{
				//will need the size
				long g_size = symboltable->getSize(g_id, g_defnum);

				switch(g_type & A_BASETYPES)
				{
				case T_CHAR:
					{
						for(int i = 0; i < g_size / SIZE_CHAR; i++)
						{
							//create initializer/space saver
							icur->next = new i86instruct;
							icur = icur->next;
							icur->instr = iSTRING;
							icur->next = NULL;
							icur->string = new char[256];
							if( *( (char **)symboltable->getInitVal(g_id, g_defnum) ) != NULL)
								sprintf(icur->string, "  .byte %d", (*((char **)symboltable->getInitVal(g_id, g_defnum)))[i] );
							else
								strcpy(icur->string, "  .byte 0");
						}
					}
					break;
				case T_SHORT:
					{
						for(int i = 0; i < g_size / SIZE_SHORT; i++)
						{
							//create initializer/space saver
							icur->next = new i86instruct;
							icur = icur->next;
							icur->instr = iSTRING;
							icur->next = NULL;
							icur->string = new char[256];
							if( *( (short **)symboltable->getInitVal(g_id, g_defnum) ) != NULL)
								sprintf(icur->string, "  .short %d", (*((short **)symboltable->getInitVal(g_id, g_defnum)))[i] );
							else
								strcpy(icur->string, "  .short 0");
						}
					}
					break;
				case T_LONG:
					{
						for(int i = 0; i < g_size / SIZE_LONG; i++)
						{
							//create initializer/space saver
							icur->next = new i86instruct;
							icur = icur->next;
							icur->instr = iSTRING;
							icur->next = NULL;
							icur->string = new char[256];
							if( *( (long **)symboltable->getInitVal(g_id, g_defnum) ) != NULL)
								sprintf(icur->string, "  .long %d", (*((long **)symboltable->getInitVal(g_id, g_defnum)))[i] );
							else
								strcpy(icur->string, "  .long 0");
						}
					}
					break;
				case T_FLOAT:
					{
						for(int i = 0; i < g_size / SIZE_FLOAT; i++)
						{
							//create initializer/space saver
							icur->next = new i86instruct;
							icur = icur->next;
							icur->instr = iSTRING;
							icur->next = NULL;
							icur->string = new char[256];
							if( *( (long **)symboltable->getInitVal(g_id, g_defnum) ) != NULL)
								sprintf(icur->string, "  .long %d", (*((long **)symboltable->getInitVal(g_id, g_defnum)))[i] );
							else
								strcpy(icur->string, "  .long 0");
						}
					}
					break;
				case T_DOUBLE:
					{
						for(int i = 0; i < g_size / SIZE_DOUBLE; i++)
						{
							//create initializer/space saver
							icur->next = new i86instruct;
							icur = icur->next;
							icur->instr = iSTRING;
							icur->next = NULL;
							icur->string = new char[256];
							if( *( (long **)symboltable->getInitVal(g_id, g_defnum) ) != NULL)
								sprintf(icur->string, "  .long %d,%d", (*((long **)symboltable->getInitVal(g_id, g_defnum)))[i], (*((long **)symboltable->getInitVal(g_id, g_defnum)))[i+1]);
							else
								strcpy(icur->string, "  .long 0");
						}
					}
					break;
				}
			}
		}
		else //non-array (or possibly resizable)
		{
			//create initializer/space saver
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iSTRING;
			icur->next = NULL;
			icur->string = new char[256];

			icur->string[0] = '\0';

			if(g_type & A_USERTYPE)
			{
				sprintf(icur->string, "  .zero %d", symboltable->getSize(g_id, g_defnum));
			}
			else
			{
				switch(g_type & A_BASETYPES)
				{
				case T_CHAR:
					sprintf(icur->string, "  .byte %d", *((char *)symboltable->getInitVal(g_id, g_defnum)) );
					break;
				case T_SHORT:
					sprintf(icur->string, "  .short %d", *((short *)symboltable->getInitVal(g_id, g_defnum)) );
					break;
				case T_LONG:
					sprintf(icur->string, "  .long %d", *((long *)symboltable->getInitVal(g_id, g_defnum)) );
					break;
				case T_FLOAT:
					sprintf(icur->string, "  .long %d", *((long *)symboltable->getInitVal(g_id, g_defnum)) );
					break;
				case T_DOUBLE:
					sprintf(icur->string, "  .long %d,%d", *((long *)symboltable->getInitVal(g_id, g_defnum)), ((long *)symboltable->getInitVal(g_id, g_defnum))[1] );
					break;
				}
			}
		}

		//don't need name anymore
		delete [] g_name;
	}	
}

CodeGenerator_i86::~CodeGenerator_i86()
{
	//delete listing
	while(icode != NULL)
	{
		i86instruct *temp = icode;
		icode = icode->next;
		if(temp->instr == iSTRING && temp->string != NULL)
			delete [] temp->string;
		delete temp;
	}
}

//gets the name used for assembler
char *CodeGenerator_i86::getFormalFunctionName(TOKEN_ID id, long defnum)
{
	//use standard gcc formatting for functions (so can link functions)
	//usage:  name__Fxxxx  where the x's are:
	//c = char, s = short, i = int, l = long, f = float, d = double
	//Px = pointer to x, Ux = unsigned x, P3xxx = pointer of type xxx with strlen of 3

	if(id == T_MAIN)
	{
		char *curfuncname = new char[6];
		return strcpy(curfuncname, "pmain");
	}		

	char *name = stringtable->getString(
		symboltable->getNameLookup(id, defnum) );
	char *parentname = NULL;

	TOKEN_ID parent = symboltable->getParent(id, defnum);
	if(parent != GLOBALSCOPE)
	{
		long parentdefnum = symboltable->getClassDefnum(parent);
		parentname = stringtable->getString(symboltable->getNameLookup(parent, parentdefnum));
	}

	char *formalname = new char[256];

	char params[256];
	strcpy(params, "__F");

	FunctionType *f = symboltable->getFunctionType(id, defnum);

	if(f->numarguements == 0)
		strcat(params, "v");

	for(int i = 0; i < f->numarguements; i++)
	{
		if(f->arguements[i] == A_USERTYPE)
		{
			i++;
			TOKEN_ID ut = f->arguements[i];
			long utdefnum = symboltable->getClassDefnum(ut);

			char *classname = stringtable->getString(symboltable->getNameLookup(ut, utdefnum));
			sprintf(&params[strlen(params)], "P%d%s", strlen(classname), classname);
		}
		else
		{
			if(f->arguements[i] & A_ARRAY)
				strcat(params, "P");

			if(f->arguements[i] & A_UNSIGNED)
				strcat(params, "U");

			switch(f->arguements[i] & A_BASETYPES)
			{
			case T_CHAR:
				strcat(params, "c");
				break;
			case T_SHORT:
				strcat(params, "s");
				break;
			case T_LONG:
				strcat(params, "l");
				break;
			case T_FLOAT:
				strcat(params, "f");
				break;
			case T_DOUBLE:
				strcat(params, "d");
				break;
			case T_IMMEDIATE_STR:
				strcat(params, "Pc");
				break;
			case T_ADDRESS:
				break;
			}
		}
	}


	if(parentname != NULL)
	{
		sprintf(formalname, "%s_%s", name, parentname);
		delete [] parentname;
	}
	else
	{
		sprintf(formalname, "%s", name, defnum);
	}

	strcat(formalname, params);

	delete [] name;
	return formalname;
}

void CodeGenerator_i86::GenerateCode(FunctionType *func)
{
	//don't write out external functions
	if(func->returnType & A_EXTERNAL)
		return;

	currentfunction = func;	//saved for a couple occurences of future use

	//get clean stack
	tempstack.Reset();
	parameter_stack_depth = 0;

	//print function label and export information
	char *curfuncname = getFormalFunctionName(symboltable->getCurFunctionID(),
									symboltable->getCurFunctionDefnum() );

	//tell it that it's a text segment
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[16];
	strcpy(icur->string, "\n.section .text");

	//set alignment
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[16];
	strcpy(icur->string, "  .align 4");

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	sprintf(icur->string, "\n.globl %s",  curfuncname);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	sprintf(icur->string, "  .type %s,@function", curfuncname);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	sprintf(icur->string, "%s:", curfuncname);

	//set up stack

	//save ebp
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPUSH;
	icur->op1a = iEBP;
	icur->next = NULL;

	//set up room for function's variables

	//use this as new stack base (to go up from)
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iMOV;
	icur->op1a = iESP;
	icur->op2a = iEBP;
	icur->next = NULL;

	//***for optimizing and better register allocation, this should be made much better
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSUB;
	icur->op1a = iIMMD;
	icur->l_immediate = -func->curvarpos - SIZE_LONG;	//curvarpos is negative, offset by return address and saved ebp
	icur->op2a = iESP;
	icur->next = NULL;


	//save other registers -not eax, because that's the return
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPUSH;
	icur->op1a = iEBX;
	icur->next = NULL;
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPUSH;
	icur->op1a = iECX;
	icur->next = NULL;
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPUSH;
	icur->op1a = iEDX;
	icur->next = NULL;

	//go through instructions
	for(MIRInstr *mir = func->mircode; mir != NULL; mir = mir->next)
	{
		switch(mir->instr)
		{
		case T_IF:
			{
				//cmp operand1, operand2
				i86instruct *icmp_original;
				i86instruct *icmp = icmp_original = new i86instruct;
				icmp->instr = iCMP;
				icmp->next = NULL;
				FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
								mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
								icmp, mir);

				bool opposite_compair = false;

				//can't have cmp operand, immediate
				if(icmp_original->op2a == iIMMD || icmp_original->op2a == iIMMW || icmp_original->op2a == iIMMB)
				{
					//swap operands
					OPERAND_i86 temp;
					temp = icmp_original->op1a;
					icmp_original->op1a = icmp_original->op2a;
					icmp_original->op2a = temp;
					temp = icmp_original->op1b;
					icmp_original->op1b = icmp_original->op2b;
					icmp_original->op2b = temp;
					temp = icmp_original->op1c;
					icmp_original->op1c = icmp_original->op2c;
					icmp_original->op2c = temp;

					opposite_compair = true;
				}

				icur->next = icmp;
				while(icur->next != NULL)
					icur = icur->next;

				icur->next = new i86instruct;
				icur = icur->next;
				icur->next = NULL;

				icur->op1a = mir->target;

				if(!opposite_compair)
				{
					switch(mir->dest_id)
					{
					case T_EQUALS:
						icur->instr = iJE;
						break;
					case T_LESS:
						icur->instr = iJLE;
						break;
					case T_LEQUAL:
						icur->instr = iJL;
						break;
					case T_GREATER:
						icur->instr = iJGE;
						break;
					case T_GEQUAL:
						icur->instr = iJG;
						break;
					case T_NOTEQUAL:
						icur->instr = iJNE;
						break;
					}
				}
				else //opposite compair
				{
					switch(mir->dest_id)
					{
					case T_EQUALS:
						icur->instr = iJE;
						break;
					case T_LESS:
						icur->instr = iJGE;
						break;
					case T_LEQUAL:
						icur->instr = iJG;
						break;
					case T_GREATER:
						icur->instr = iJLE;
						break;
					case T_GEQUAL:
						icur->instr = iJL;
						break;
					case T_NOTEQUAL:
						icur->instr = iJNE;
						break;
					}
				} //else opposite compair
			}
			break;

		case T_RETURN:
			if(mir->op1_id != (T_IMMEDIATE | T_VOID))
			{
				//put return value in EAX
				i86instruct *iret = new i86instruct;
				iret->next = NULL;
				iret->instr = iMOV;
				iret->op2a = iEAX;
				FetchOperand(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer, iret, mir, iNONE);

				icur->next = iret;
				while(icur->next != NULL)
					icur = icur->next;
			}

			//restore other registers -not eax, because that's the return
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iPOP;
			icur->op1a = iEDX;
			icur->next = NULL;
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iPOP;
			icur->op1a = iECX;
			icur->next = NULL;
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iPOP;
			icur->op1a = iEBX;
			icur->next = NULL;

			//fix stack
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iMOV;
			icur->op1a = iEBP;
			icur->op2a = iESP;
			icur->next = NULL;

			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iPOP;
			icur->op1a = iEBP;
			icur->next = NULL;

			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iRET;
			icur->next = NULL;
			break;

		case T_PUSHPARAMETER:
			{
				i86instruct *ipush = new i86instruct;
				ipush->next = NULL;
				ipush->instr = iPUSH;

				//save eax first (because that will be return) if first parameter
				if(parameter_stack_depth == 0)
				{
					bool need_to_save;
					tempstack.allocateRegister(-1, need_to_save, iEAX);
					tempstack.ReleaseRegister(-1);	//won't need anymore

					FetchOperand(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer, ipush, mir, iNONE);

					if(need_to_save)
					{
						i86instruct *temp = new i86instruct;
						temp->instr = iPUSH;
						temp->op1a = iEAX;
						temp->next = ipush;
						ipush = temp;
					}

				}
				else //just another parameter
				{
					FetchOperand(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer, ipush, mir, iNONE);
				}

				tempstack.AddFunctionParameter(SIZE_LONG);

				//keep track of size
				parameter_stack_depth += SIZE_LONG;

				icur->next = ipush;
				while(icur->next != NULL)
						icur = icur->next;
			}
			break;
		case T_PULLPARAMETER:
			{
				i86instruct *mov_original;
				i86instruct *mov = mov_original = new i86instruct;
				mov->instr = iMOV;
				mov->next = NULL;
				FetchOperand(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer, mov, mir, iEAX);
				mov_original->op2a = mov_original->op1a;
				mov_original->op2b = mov_original->op1b;
				mov_original->op2c = mov_original->op1c;
				mov_original->op1a = iEAX;

				icur->next = mov;
				while(icur->next != NULL)
					icur = icur->next;
			}
			break;

		case T_FUNCTIONCALL:
			{
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iCALL;
				icur->next = NULL;
				icur->func_id = mir->dest_id;
				icur->func_defnum = mir->dest_defnum;

				//put stack back where it was
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iADD;
				icur->next = NULL;
				icur->op1a = iIMMD;
				icur->l_immediate = parameter_stack_depth;
				icur->op2a = iESP;

				tempstack.ClearFunctionParameters();

				//reset stack
				parameter_stack_depth = 0;
			}
			break;
		case T_PFUNCTIONCALL:
			{
				//push stack size to copy over to other machine
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iPUSH;
				icur->op1a = iIMMD;
				icur->l_immediate = parameter_stack_depth;

				//push wait_id
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iPUSH;
				icur->op1a = iIMMD;
				icur->l_immediate = mir->target;

				//push function number (enumeration)
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iPUSH;
				icur->op1a = iIMMD;
				FunctionType *f = symboltable->getFunctionType(mir->dest_id, mir->dest_defnum);
				icur->l_immediate = f->enumeration;

				//call parallel_call
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iSTRING;
				icur->next = NULL;
				icur->string = new char[48];
				strcpy(icur->string, " call parallel_call__Flll");

				//put stack back where it was
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iADD;
				icur->next = NULL;
				icur->op1a = iIMMD;
				icur->l_immediate = parameter_stack_depth + 3 * SIZE_LONG; //stack size plus three for parallel call params
				icur->op2a = iESP;

				tempstack.ClearFunctionParameters();

				//reset stack
				parameter_stack_depth = 0;
			}
			break;
		case T_WAITRETURN:
			{
				//push wait_id
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iPUSH;
				icur->op1a = iIMMD;
				icur->l_immediate = mir->target;

				//call wait_for
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iSTRING;
				icur->next = NULL;
				icur->string = new char[48];
				strcpy(icur->string, " call wait_parallel_call__Fl");

				//put stack back where it was
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iADD;
				icur->next = NULL;
				icur->op1a = iIMMD;
				icur->l_immediate = SIZE_LONG; //stack size plus three for parallel call params
				icur->op2a = iESP;
			}
			break;

		///////////////
		//comunative operators -can do all the same
		case T_BAND:
		case T_BXOR:
		case T_BOR:
		case T_PLUS:
			{
				//choose of any of these
				long com_operation;
				switch(mir->instr)
				{
					case T_BAND:
						com_operation = iAND;
						break;
					case T_BXOR:
						com_operation = iXOR;
						break;
					case T_BOR:
						com_operation = iOR;
						break;
					case T_PLUS:
						com_operation = iADD;
						break;
				}
				

				//if operand 1 and operand 2 are different than destination, then do a move
				//otherwise collect into one of them
				if(mir->dest_id == mir->op1_id && mir->dest_defnum == mir->op1_defnum)
				{
					//add operand2, operand1|dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
									mir->dest_id, mir->dest_defnum, mir->op1_data, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
				else if(mir->dest_id == mir->op2_id && mir->dest_defnum == mir->op2_defnum)
				{
					//add operand1, operand2|dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
									mir->dest_id, mir->dest_defnum, mir->op2_data, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
				else
				{
					//mov operand1, dest
					i86instruct *imov = new i86instruct;
					imov->instr = iMOV;
					imov->next = NULL;
					FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
									mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
									imov, mir);

					//add operand2, dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
									mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = imov;
					while(icur->next != NULL)
						icur = icur->next;

					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
			}
			break;
		case T_MINUS:
			{
				//choose of any of these
				long com_operation;
				switch(mir->instr)
				{
					case T_MINUS:
						com_operation = iSUB;
						break;
				}

				//if operand 1 and operand 2 are different than destination, then do a move
				//otherwise collect into one of them
				if(mir->dest_id == mir->op1_id && mir->dest_defnum == mir->op1_defnum)
				{
					//sub operand2, operand1|dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
									mir->dest_id, mir->dest_defnum, mir->op1_data, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
				else if(mir->dest_id == mir->op2_id && mir->dest_defnum == mir->op2_defnum)
				{
					//sub operand1, operand2|dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
									mir->dest_id, mir->dest_defnum, mir->op2_data, mir->dest_pointer,
									com_instr, mir);

					//neg dest
					com_instr = new i86instruct;
					com_instr->instr = iNEG;
					com_instr->next = NULL;
					FetchOperand(mir->dest_id, mir->dest_defnum, mir->op2_data, mir->dest_pointer,
									com_instr, mir, iNONE);

					//concatenate new instructions onto end of list
					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
				else
				{
					//mov operand1, dest
					i86instruct *imov = new i86instruct;
					imov->instr = iMOV;
					imov->next = NULL;
					FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
									mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
									imov, mir);

					//sub operand2, dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
									mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = imov;
					while(icur->next != NULL)
						icur = icur->next;

					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
			}
			break;


		case T_MULTIPLY:
			{
				//***will have to be modified to support different types

				//can change this if have a signed multiplication
				long com_operation = iMUL;

				//mov operand1, eax
				i86instruct *imov1 = new i86instruct;
 				imov1->instr = iMOV;
				imov1->next = NULL;
				imov1->op2a = iEAX;
				FetchOperand(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
								imov1, mir, iEAX);

				//mov operand2, edx
				i86instruct *imov2 = new i86instruct;
				imov2->instr = iMOV;
				imov2->next = NULL;
				imov2->op2a = iEDX;
				FetchOperand(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
								imov2, mir, iEDX);

				//mov eax, dest
				i86instruct *imov3 = new i86instruct;
				imov3->instr = iMOV;
				imov3->next = NULL;
				imov3->op2a = iEAX;
				FetchOperand(mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
								imov3, mir, iEAX);

				bool save_edx = tempstack.isRegisterInUse(iEDX);

				//mul edx
				i86instruct *com_instr = new i86instruct;
				com_instr->instr = com_operation;
				com_instr->next = NULL;
				com_instr->op1a = iEDX;

				//concatenate new instructions onto end of list
				icur->next = imov1;
				while(icur->next != NULL)
					icur = icur->next;
				icur->next = imov2;
				while(icur->next != NULL)
					icur = icur->next;

				//lost edx (or portion of) -retrieve

				if(save_edx)
				{
					//push edx
					icur->next = new i86instruct;
					icur = icur->next;
					icur->next = NULL;
					icur->instr = iPUSH;
					icur->op1a = iEDX;
				}

				icur->next = com_instr;
				while(icur->next != NULL)
					icur = icur->next;

				if(save_edx)
				{
					//pop edx
					icur->next = new i86instruct;
					icur = icur->next;
					icur->next = NULL;
					icur->instr = iPOP;
					icur->op1a = iEDX;
				}

				icur->next = imov3;
				while(icur->next != NULL)
					icur = icur->next;
			}
			break;

		case T_DIVIDE:
		case T_MODULUS:
			{
			//can change this if have a signed division
				long com_operation = iDIV;

				//mov operand1, eax
				i86instruct *imov1 = new i86instruct;
 				imov1->instr = iMOV;
				imov1->next = NULL;
				imov1->op2a = iEAX;
				FetchOperand(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
								imov1, mir, iEAX);

				//mov operand2, ebx
				i86instruct *imov2 = new i86instruct;
				imov2->instr = iMOV;
				imov2->next = NULL;
				imov2->op2a = iEBX;
				FetchOperand(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
								imov2, mir, iEBX);

				//see if divide or modulus -use different registers
				OPERAND_i86 source;
				OPERAND_i86 registersave;
				switch(mir->instr)
				{
					case T_DIVIDE:
						source = iEAX;
						registersave = iEDX;
						break;
					case T_MODULUS:
						source = iEDX;
						registersave = iEAX;
						break;
				}
				//mov eax, dest
				i86instruct *imov3 = new i86instruct;
				imov3->instr = iMOV;
				imov3->next = NULL;
				imov3->op2a = source;
				FetchOperand(mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
								imov3, mir, source);

				//div edx
				i86instruct *com_instr = new i86instruct;
				com_instr->instr = com_operation;
				com_instr->next = NULL;
				com_instr->op1a = iEBX;

				//concatenate new instructions onto end of list
				icur->next = imov1;
				while(icur->next != NULL)
					icur = icur->next;
				icur->next = imov2;
				while(icur->next != NULL)
					icur = icur->next;

				//lost eax or edx register (or portion of) -retrieve

				//push edx or eax
				icur->next = new i86instruct;
				icur = icur->next;
				icur->next = NULL;
				icur->instr = iPUSH;
				icur->op1a = registersave;

				//mov 0,edx
				icur->next = new i86instruct;
				icur = icur->next;
				icur->next = NULL;
				icur->instr = iMOV;
				icur->op1a = iIMMD;
				icur->l_immediate = 0;
				icur->op2a = iEDX;

				icur->next = com_instr;
				while(icur->next != NULL)
					icur = icur->next;

				//pop edx or eax
				icur->next = new i86instruct;
				icur = icur->next;
				icur->next = NULL;
				icur->instr = iPOP;
				icur->op1a = registersave;

				icur->next = imov3;
				while(icur->next != NULL)
					icur = icur->next;
			}
			break;

		//***********AS OF NOW, THESE INSTRUCTIONS WILL ONLY WORK WITH IMMEDIATES
		case T_SHIFTLEFT:
		case T_SHIFTRIGHT:
			{
				//choose of any of these
				long com_operation;
				switch(mir->instr)
				{
					case T_SHIFTLEFT:
						com_operation = iSHL;
						break;
					case T_SHIFTRIGHT:
						com_operation = iSHR;
						break;
				}

				//if operand 1 and operand 2 are different than destination, then do a move
				//otherwise collect into one of them
				if(mir->dest_id == mir->op1_id && mir->dest_defnum == mir->op1_defnum)
				{
					//shl operand2, operand1|dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
									mir->dest_id, mir->dest_defnum, mir->op1_data, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
				else if(mir->dest_id == mir->op2_id && mir->dest_defnum == mir->op2_defnum)
				{
					//shl operand1, operand2|dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
									mir->dest_id, mir->dest_defnum, mir->op2_data, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
				else
				{
					//mov operand1, dest
					i86instruct *imov = new i86instruct;
					imov->instr = iMOV;
					imov->next = NULL;
					FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
									mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
									imov, mir);

					//shl operand2, dest
					i86instruct *com_instr = new i86instruct;
					com_instr->instr = com_operation;
					com_instr->next = NULL;
					FetchOperands(mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
									mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
									com_instr, mir);

					//concatenate new instructions onto end of list
					icur->next = imov;
					while(icur->next != NULL)
						icur = icur->next;

					icur->next = com_instr;
					while(icur->next != NULL)
						icur = icur->next;
				}
			}
			break;

		case T_EQUALS:
		case T_LESS:
		case T_LEQUAL:
		case T_GREATER:
		case T_GEQUAL:
		case T_NOTEQUAL:
			{
				//mov 0, dest
				i86instruct *imov = new i86instruct;
				imov->next = NULL;
				imov->instr = iMOV;
				//will set parameters after 'set'

				//cmp operand1, operand2
				i86instruct *icmp_original;
				i86instruct *icmp = icmp_original = new i86instruct;
				icmp->instr = iCMP;
				icmp->next = NULL;
				FetchOperands(mir->op1_id, mir->op1_defnum, mir->op1_data, mir->op1_pointer,
								mir->op2_id, mir->op2_defnum, mir->op2_data, mir->op2_pointer,
								icmp, mir);

				//set dest
				i86instruct *setoriginal;
				i86instruct *iset = setoriginal = new i86instruct;

				bool opposite_compair = false;

					//can't have cmp operand, immediate
				if(icmp_original->op2a == iIMMD || icmp_original->op2a == iIMMW || icmp_original->op2a == iIMMB)
				{
					//swap operands
					OPERAND_i86 temp;
					temp = icmp_original->op1a;
					icmp_original->op1a = icmp_original->op2a;
					icmp_original->op2a = temp;
					temp = icmp_original->op1b;
					icmp_original->op1b = icmp_original->op2b;
					icmp_original->op2b = temp;
					temp = icmp_original->op1c;
					icmp_original->op1c = icmp_original->op2c;
					icmp_original->op2c = temp;

					opposite_compair = true;
				}

				if(!opposite_compair)
				{
					switch(mir->instr)
					{
					case T_EQUALS:
						iset->instr = iSETE;
						break;
					case T_LESS:
						iset->instr = iSETLE;
						break;
					case T_LEQUAL:
						iset->instr = iSETL;
						break;
					case T_GREATER:
						iset->instr = iSETGE;
						break;
					case T_GEQUAL:
						iset->instr = iSETG;
						break;
					case T_NOTEQUAL:
						iset->instr = iSETNE;
						break;
					}
				}
				else //opposite compair
				{
					switch(mir->instr)
					{
					case T_EQUALS:
						iset->instr = iSETE;
						break;
					case T_LESS:
						iset->instr = iSETGE;
						break;
					case T_LEQUAL:
						iset->instr = iSETG;
						break;
					case T_GREATER:
						iset->instr = iSETLE;
						break;
					case T_GEQUAL:
						iset->instr = iSETL;
						break;
					case T_NOTEQUAL:
						iset->instr = iSETNE;
						break;
					}
				} //else opposite compair

				iset->next = NULL;
				FetchOperand(mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer,
								iset, mir, iNONE);

				imov->op1a = iIMMD;
				imov->l_immediate = 0;
				imov->op2a = setoriginal->op1a;
				imov->op2b = setoriginal->op1b;
				imov->op2c = setoriginal->op1c;

				//change set size to byte
				setoriginal->op1a = ((setoriginal->op1a & ~0xF) | iBYTE);

				icur->next = icmp;
				while(icur->next != NULL)
					icur = icur->next;

				icur->next = imov;
				while(icur->next != NULL)
					icur = icur->next;

				icur->next = iset;
				while(icur->next != NULL)
					icur = icur->next;
			}
			break;

		case T_LABEL:
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iLABEL;
			icur->op1a = mir->target;
			icur->next = NULL;
			break;
		case T_GOTO:
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iJMP;
			icur->op1a = mir->target;
			icur->next = NULL;
			break;
		case T_ARRAYRESIZE:
			break;
		case T_ADDRESSOF:
			{
				//lea global,reg
				i86instruct *original;
				i86instruct *ilea = original = new i86instruct;
				ilea->next = NULL;
				ilea->instr = iLEA;
				FetchOperand(mir->dest_id, mir->dest_defnum, NULL, mir->dest_pointer, ilea, mir, iNONE);

				//move to second operand
				original->op2a = original->op1a;
				original->op2b = original->op1b;
				original->op2c = original->op1c;

				//get global
				original->op1a = iGLOBALD;
				original->func_id = mir->op1_id;
				original->func_defnum = mir->op1_defnum;

				//append new instructions
				icur->next = ilea;
				while(icur->next != NULL)
					icur = icur->next;

				//very tricky workaround for global variables:
				//subtract ebp from original address
				//so when final address is calculated,
				//ebp is nullified :-)
				//sub %ebp, addr(from original->op2)
				icur->next = new i86instruct;
				icur = icur->next;
				icur->instr = iSUB;
				icur->next = NULL;
				icur->op1a = iEBP;
				icur->op2a = original->op2a;
				icur->op2b = original->op2b;
				icur->op2c = original->op2c;
				icur->offset = original->offset;
				icur->func_id = original->func_id;
				icur->func_defnum = original->func_defnum;
			}
			break;
		case T_GLOBALSPACE:
			{
				//add ebp, pointer
				i86instruct *original;
				i86instruct *global = original = new i86instruct;
				global->next = NULL;
				global->instr = iADD;
				FetchOperand(mir->op1_id, mir->op1_defnum, NULL, mir->op1_pointer, global, mir, iNONE);

				//move to second operand
				original->op2a = original->op1a;
				original->op2b = original->op1b;
				original->op2c = original->op1c;

				original->op1a = iEBP;

				//append new instructions
				icur->next = global;
				while(icur->next != NULL)
					icur = icur->next;
			}
			break;
		case T_PUSH_THIS:
			{
				icur->next = new i86instruct;
				icur = icur->next;
				icur->next = NULL;
				icur->instr = iPUSH;
				icur->op1a = iROFFD;
				icur->op1b = iEBP;
				icur->offset = currentfunction->parameterlength;

				tempstack.AddFunctionParameter(SIZE_LONG);

				//keep track of size
				parameter_stack_depth += SIZE_LONG;

			}
			break;
		default:
			break;

		} //switch mir instruction

		long stack_recovered = CleanStack(mir);
		if(stack_recovered)
		{
			icur->next = new i86instruct;
			icur = icur->next;
			icur->instr = iADD;
			icur->op1a = iIMMD;
			icur->l_immediate = stack_recovered;
			icur->op2a = iESP;
			icur->next = NULL;
		}
	}

	//restore other registers -not eax, because that's the return
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPOP;
	icur->op1a = iEDX;
	icur->next = NULL;
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPOP;
	icur->op1a = iECX;
	icur->next = NULL;
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPOP;
	icur->op1a = iEBX;
	icur->next = NULL;

	//throw on an extra return/exit statement incase not provided
	//-can be removed by dead code elimination optimizer
	//fix stack
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iMOV;
	icur->op1a = iEBP;
	icur->op2a = iESP;
	icur->next = NULL;

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iPOP;
	icur->op1a = iEBP;
	icur->next = NULL;

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iRET;
	icur->next = NULL;

	//put end of function info
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	sprintf(icur->string, ".Lfe%s:", curfuncname, curfuncname);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	sprintf(icur->string, "  .size %s,Lfe%s-%s", curfuncname, curfuncname, curfuncname);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[4];
	strcpy(icur->string, "\n");

	delete [] curfuncname;

}


//assigns a number to each function
void CodeGenerator_i86::EnumerateFunctions()
{
	FunctionType *curfunc;
	long i;
	//count functions (the order they will be tabled)
	//-start at 1, because pmain is also located in position 1
	for(curfunc = symboltable->getFirstFunction(), i = 1;
		curfunc != NULL; curfunc = symboltable->getNextFunction(), i++ )
	{
		//don't write out external functions, and don't count them either
		if(curfunc->returnType & A_EXTERNAL)
		{
			i--;
			continue;
		}

		curfunc->enumeration = i;
		num_enumerated_functions++;
	}
}

//write out table to be used for function enumeration calls
//in the form enum_func(long function_number, ...)
void CodeGenerator_i86::GenerateFunctionEnumeration()
{
	//create table of function labels
	//allow variable to be exportable
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	strcpy(icur->string, "\n.globl func_enums");

	//alignment information (can make more complex and efficient)
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	strcpy(icur->string, ".align 4");

	//set type
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	strcpy(icur->string, "  .type func_enums,@object");
	//set size
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	sprintf(icur->string, "  .size func_enums,%d", (num_enumerated_functions + 1) * SIZE_LONG); //plus 1 for extra pmain
	
	//make label
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	strcpy(icur->string, "func_enums:");

	//print pmain as first function
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[256];
	strcpy(icur->string, ".long pmain");

	//go through all functions
	FunctionType *curfunc;
	for(curfunc = symboltable->getFirstFunction(); curfunc != NULL; curfunc = symboltable->getNextFunction() )
	{
		//don't write out external functions
		if(curfunc->returnType & A_EXTERNAL)
			continue;

		//print function label and export information
		char *curfuncname = getFormalFunctionName(symboltable->getCurFunctionID(),
										symboltable->getCurFunctionDefnum() );

		//make a label and pointer for each function
		icur->next = new i86instruct;
		icur = icur->next;
		icur->instr = iSTRING;
		icur->next = NULL;
		icur->string = new char[256];
		sprintf(icur->string, ".long %s", curfuncname);

		delete [] curfuncname;
	}

	//tell it that it's a text segment
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[16];
	strcpy(icur->string, "\n.section .text");

	//set alignment

	//code for function -broken into pieces
	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[strlen(call_function_enumerator_code1)+5];
	strcpy(icur->string, call_function_enumerator_code1);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[strlen(call_function_enumerator_code2)+5];
	strcpy(icur->string, call_function_enumerator_code2);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[strlen(call_function_enumerator_code3)+5];
	strcpy(icur->string, call_function_enumerator_code3);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[strlen(call_function_enumerator_code4)+5];
	strcpy(icur->string, call_function_enumerator_code4);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[strlen(call_function_enumerator_code5)+5];
	strcpy(icur->string, call_function_enumerator_code5);

	icur->next = new i86instruct;
	icur = icur->next;
	icur->instr = iSTRING;
	icur->next = NULL;
	icur->string = new char[strlen(call_function_enumerator_code6)+5];
	strcpy(icur->string, call_function_enumerator_code6);
}

