

#include "i86architecture.h"

#define NUM_AVALIABLE_REGISTERS 4

struct TempStackNode
{
	long defnum;
	ATTRIB_ADDR size;	//size of data

	TempStackNode *next;
};

//this is used for allocating a register for a single thing
#define TEMPORARY_TEMP (-1)
//this is for function parameters
#define PARAMETER_TEMP (-2)

//keeps track of runtime stack
class TempStack
{
public:
	TempStack();
	~TempStack();

	//resets stack, etc.
	void Reset();

	//pass in defnum of temporary
	//returns 0xX0 where X is the number of register (or with size to use)
	//fills in need_to_save with whether or not need to save previous register
	//if reg is set to other than iNONE, it puts it in that register
	OPERAND_i86 allocateRegister(long defnum, bool &need_to_save, OPERAND_i86 reg);

	//these are used as placeholders on the stack
	void AddFunctionParameter(long size);
	void ClearFunctionParameters();

	//used to determine whether or not a register needs to be saved
	bool isRegisterInUse(OPERAND_i86 reg);

	//returns register temporary is in, iNONE if it's not in there
	OPERAND_i86 TempRegister(long defnum);

	//checks to see if free at top of stack (if so, then can just pop it off) -requires free register
	//if reg is not iNONE, then need to look at that register
	bool FreeAtTopOfStack(long defnum, OPERAND_i86 reg);

	//removes top node from stack --check AtTopOfStack to see if can do this
	void RemoveTopOfStack();

	//this will deallocate the register used by the temporary, returning it to the pool
	void ReleaseRegister(long defnum);

	//address of temporary relative to current stack position, -1 if not in stack
	ATTRIB_ADDR TempLocation(long defnum);

	//returns defnum of top of stack
	long stackTopDefnum();
	//returns number of bytes on top of stack that are done with
	long removeStackTop();

private:
	//stack to hold where temporaries are
	TempStackNode *root;

	long register_defnums[NUM_AVALIABLE_REGISTERS];	//temporary assigned to each register
	int previous_register;	//register used during last allocation (try to keep cycling)
};



class CodeGenerator_i86
{
public:
	CodeGenerator_i86(StringTable *st, SymbolTable *yt);
	~CodeGenerator_i86();

	void GenerateCode(FunctionType *func);
	void PrintListing(FILE *outfile);

	void PeepholeOptimization();

	//takes in target code (client, server, etc.)
	void GenerateFunctionEnumeration();	//generates function enumeration table and caller
	void EnumerateFunctions();	//enumerates functions for reference in parallel calling

private:
	char *getFormalFunctionName(TOKEN_ID id, long defnum);	//gets the name used for assembler
	void concatParam(char *buffer, OPERAND_i86 opa, OPERAND_i86 opb, OPERAND_i86 opc, i86instruct *instr);	//appends a parameter to instruction text string

	//used when creating an instruction
	//takes an instruction and puts in correct values for the given operands
	//possibly prepending the instruction with additional
	//instructions if dynamic fetching (i.e. stashed register) must be done
	//set anything unused to 0 or NULL
	void FetchOperands(TOKEN_ID op1_id, long op1_defnum, void *op1_data, bool op1_pointer, TOKEN_ID op2_id, long op2_defnum, void *op2_data, bool op2_pointer, i86instruct *&instr, MIRInstr *mir);

	//version of above, but for one operand instructions
	//if reg is set to other than iNONE, it puts it in that register
	void FetchOperand(TOKEN_ID op1_id, long op1_defnum, void *op1_data, bool op1_pointer, i86instruct *&instr, MIRInstr *mir, OPERAND_i86 reg);

	//used to retrieve temp
	//if unused, creates
	//if previously used, grabs off stack
	//returns register stored in
	//retrieve points to any instructions needed to load temp into register
	//if reg is set to other than iNONE, it puts it in that register
	OPERAND_i86 FetchTemp(long temp_defnum, i86instruct *&retrieve, OPERAND_i86 reg);

	//checks stack to make sure there's nothing on top that doesn't need to be
	//this ensures that the stack will always end up where it's supposed to be
	long CleanStack(MIRInstr *m);

	//used to check if last occurence of a temporary in mir code
	bool LastOccurenceOfTemp(MIRInstr *m, unsigned long temp_defnum);


	StringTable *stringtable;
	SymbolTable *symboltable;

	i86instruct *icode;
	i86instruct *icur;

	//this is used to keep track of parameters pushed on stack for function calling
	//clean stack after calling
	long parameter_stack_depth;

	//class to keep track of temporaries
	TempStack tempstack;

	//used to know how many enumerated functions exist
	long num_enumerated_functions;

	FunctionType *currentfunction;	//function currently transforming
};

