#include "common.h"

TempStack::TempStack()
{
	previous_register = 0;
	for(int i = 0; i < NUM_AVALIABLE_REGISTERS; i++)
		register_defnums[i] = 0;
	root = NULL;

}

TempStack::~TempStack()
{
	while(root != NULL)
	{
		TempStackNode *temp = root;
		root = root->next;
		delete temp;
	}
}

void TempStack::Reset()
{
	previous_register = 0;
	for(int i = 0; i < NUM_AVALIABLE_REGISTERS; i++)
		register_defnums[i] = 0;
	while(root != NULL)
	{
		TempStackNode *temp = root;
		root = root->next;
		delete temp;
	}
}

OPERAND_i86 TempStack::allocateRegister(long defnum, bool &need_to_save, OPERAND_i86 reg)
{
	//find next avaliable register
	int next_register;

	need_to_save = true;

	//if register not specified, find one
	if(reg == iNONE)
	{
		for(next_register = 0; next_register < NUM_AVALIABLE_REGISTERS; next_register++)
		{
			if(register_defnums[next_register] == 0)
			{
				need_to_save = false;
				break;
			}
		}

		if(need_to_save)
		{
			//mark next register for using next
			previous_register++;
			if(previous_register == NUM_AVALIABLE_REGISTERS)
				previous_register = 0;
			next_register = previous_register;

			//assume old register has been pushed onto stack -record this
			TempStackNode *n = new TempStackNode;
			n->defnum = register_defnums[next_register];
			n->size = 4;	//***ASSUME SIZE OF 4 FOR EVERYTHING FOR NOW
			n->next = root;
			root = n;
		}
		//put new temp in register list
		register_defnums[next_register] = defnum;

		return (next_register + 1) << 4;	//convert to 0xX0 format
	}

	//specified register
	next_register = (reg >> 4) - 1;
	if(register_defnums[next_register] == 0)
		need_to_save = false;

	if(need_to_save)
	{
		//assume old register has been pushed onto stack -record this
		TempStackNode *n = new TempStackNode;
		n->defnum = register_defnums[next_register];
		n->size = 4;	//***ASSUME SIZE OF 4 FOR EVERYTHING FOR NOW
		n->next = root;
		root = n;
	}

	//put new temp in register list
	register_defnums[next_register] = defnum;

//&&&this changes the feel of the register allocater
//	previous_register = next_register;

	return (next_register + 1) << 4;	//convert to 0xX0 format
}

void TempStack::AddFunctionParameter(long size)
{
	TempStackNode *n = new TempStackNode;
	n->defnum = PARAMETER_TEMP;
	n->size = size;	//***ASSUME SIZE OF 4 FOR EVERYTHING FOR NOW
	n->next = root;
	root = n;
}

void TempStack::ClearFunctionParameters()
{
	while(root != NULL && root->defnum == PARAMETER_TEMP)
	{
		TempStackNode *temp = root;
		root = root->next;
		delete temp;
	}
}

bool TempStack::isRegisterInUse(OPERAND_i86 reg)
{
	int r = (reg >> 4) - 1;
	if(register_defnums[r] == 0)
		return true;
	return false;
}

OPERAND_i86 TempStack::TempRegister(long defnum)
{
	for(int i = 0; i < NUM_AVALIABLE_REGISTERS; i++)
		if(register_defnums[i] == defnum)
			return (i + 1) << 4;

	return iNONE;
}

bool TempStack::FreeAtTopOfStack(long defnum, OPERAND_i86 reg)
{
	if(root == NULL)
		return false;
	if(root->defnum != defnum)
		return false;

	//if specified register, see if it's free
	if(reg != iNONE)
	{
		if(register_defnums[(reg >> 4) - 1] == 0)
			return true;
		return false;
	}

	//check for free register
	int free_registers;
	for(free_registers = 0; free_registers < NUM_AVALIABLE_REGISTERS; free_registers++)
	{
		if(register_defnums[free_registers] == 0)
			return true;
	}
	return false;
}

void TempStack::RemoveTopOfStack()
{
	if(root == NULL)
		return;
	TempStackNode *n = root;
	root = root->next;
	delete n;
}

void TempStack::ReleaseRegister(long defnum)
{
	for(int i = 0; i < NUM_AVALIABLE_REGISTERS; i++)
	{
		if(register_defnums[i] == defnum)
		{
			register_defnums[i] = 0;
		}
	}
}


ATTRIB_ADDR TempStack::TempLocation(long defnum)
{
	ATTRIB_ADDR total = 0;
	for(TempStackNode *t = root; t != NULL; t = t->next)
	{
		total += t->size;
		if(t->defnum == defnum)
			return total;
	}
	return -1;
}

long TempStack::stackTopDefnum()
{
	if(root == NULL)
		return 0;
	return root->defnum;
}

long TempStack::removeStackTop()
{
	if(root == NULL)
		return 0;
	long size = root->size;
	TempStackNode *temp = root;
	root = root->next;
	delete temp;
	return size;
}



//used to check if last occurence of a temporary
bool CodeGenerator_i86::LastOccurenceOfTemp(MIRInstr *m, unsigned long temp_defnum)
{
	//parameters will be cleared after function call
	if(temp_defnum == PARAMETER_TEMP)
		return false;

	for(;m != NULL; m = m->next)
	{
		if( (m->dest_id == T_TEMPREG && m->dest_defnum == temp_defnum)
			|| (m->op1_id == T_TEMPREG && m->op1_defnum == temp_defnum)
			|| (m->op2_id == T_TEMPREG && m->op2_defnum == temp_defnum) )
			return false;
	}
	return true;
}

long CodeGenerator_i86::CleanStack(MIRInstr *m)
{
	long defnum;
	long total_recovered = 0;
	while(defnum = tempstack.stackTopDefnum())
	{
		if(LastOccurenceOfTemp(m->next, defnum))
		{
			total_recovered += tempstack.removeStackTop();
		}
		else
			break;
	}
	return total_recovered;
}

OPERAND_i86 CodeGenerator_i86::FetchTemp(long temp_defnum, i86instruct *&retrieve, OPERAND_i86 reg)
{
	//main return variable
	OPERAND_i86 op;

	//check to see if already in register
	op = tempstack.TempRegister(temp_defnum);
	if(op != iNONE)
	{
		//if no register specified, this one will do
		if(reg == iNONE)
		{
			retrieve = NULL;
			return op | iDOUBLEWORD;
		}
		//need to move into a new register
		OPERAND_i86 temp = op;
		bool need_to_save;
		op = tempstack.allocateRegister(temp_defnum, need_to_save, reg);

		retrieve = new i86instruct;
		retrieve->instr = iMOV;
		retrieve->op1a = temp | iDOUBLEWORD;
		retrieve->op2a = op | iDOUBLEWORD;	//new register
		retrieve->next = NULL;

		//instruction to save register (prepended)
		if(need_to_save)
		{
			i86instruct *temp = new i86instruct;
			temp->instr = iPUSH;
			temp->op1a = op | iDOUBLEWORD;
			temp->next = retrieve;
			retrieve = temp;
		}
		return op | iDOUBLEWORD;
	}

	//not in register
	//check top of stack (and check for free register)
	if(tempstack.FreeAtTopOfStack(temp_defnum, reg))
	{
		//get rid of old one (with pop)
		tempstack.RemoveTopOfStack();

		bool need_to_save;
		op = tempstack.allocateRegister(temp_defnum, need_to_save, reg); //can ignore need to save

		retrieve = new i86instruct;
		retrieve->instr = iPOP;
		retrieve->op1a = op | iDOUBLEWORD;
		retrieve->next = NULL;

		//instruction to save register (prepended)
		if(need_to_save)
		{
			/*i86instruct *temp = new i86instruct;
			temp->instr = iPUSH;
			temp->op1a = op | iDOUBLEWORD;
			temp->next = retrieve;
			retrieve = temp;*/
			return 0;	//error -this should not happen
		}

		return op | iDOUBLEWORD;
	}

	//check deeper in stack
	ATTRIB_ADDR location = tempstack.TempLocation(temp_defnum);
	if(location != -1)
	{
		//it's in the stack -get a new spot for it and pull it out
		bool need_to_save;
		op = tempstack.allocateRegister(temp_defnum, need_to_save, reg);

		//location changed since push
		location = tempstack.TempLocation(temp_defnum);

		//pull temp off stack, then prepend instruction to save register if need to
		retrieve = new i86instruct;
		retrieve->instr = iMOV;
		retrieve->next = NULL;
		retrieve->op1a = iROFFD;
		retrieve->op1b = iESP;
		retrieve->offset = location - 4;	//because esp points to value on stack
		retrieve->op2a = op | iDOUBLEWORD;

		//instruction to save register (prepended)
		if(need_to_save)
		{
			i86instruct *temp = new i86instruct;
			temp->instr = iPUSH;
			temp->op1a = op | iDOUBLEWORD;
			temp->next = retrieve;
			retrieve = temp;
		}
		return op | iDOUBLEWORD;
	}

	//we haven't used this temporary before
	bool need_to_save;
	op = tempstack.allocateRegister(temp_defnum, need_to_save, reg);
	//need to save old one
	if(need_to_save)
	{
		retrieve = new i86instruct;
		retrieve->instr = iPUSH;
		retrieve->op1a = op | iDOUBLEWORD;
		retrieve->next = NULL;
	}

	return op | iDOUBLEWORD;
}

void CodeGenerator_i86::FetchOperands(TOKEN_ID op1_id, long op1_defnum, void *op1_data, bool op1_pointer, TOKEN_ID op2_id, long op2_defnum, void *op2_data, bool op2_pointer, i86instruct *&instr, MIRInstr *mir)
{
//***need to finish this
	//do any type conversions
	ATTRIB_TYPE op1type = symboltable->getType(op1_id, op1_defnum);
	ATTRIB_TYPE op2type = symboltable->getType(op2_id, op2_defnum);
	ATTRIB_TYPE gentype = getGeneralType(op1type, op2type);
//***

	i86instruct *attribute = NULL; //used to look up attribute (member variable)
	i86instruct *pre_instr = NULL; //used to switch over to a different register

	if(op1_id == T_TEMPREG)
	{
		instr->op1a = (FetchTemp(op1_defnum, pre_instr, iNONE) | iDOUBLEWORD);
		if(op1_pointer)
		{
			//this is for local variables
			instr->op1b = instr->op1a;
			instr->op1c = iEBP;
			instr->multiplier = 1;
			instr->op1a = iCOMPD;
			instr->offset = 0;
		}

		//check to see if ever used again
		if(LastOccurenceOfTemp(mir->next, op1_defnum))
			tempstack.ReleaseRegister(op1_defnum);
	}
	else if(op1_id & T_IMMEDIATE)
	{
		switch(op1_id & A_BASETYPES)
		{
		case T_CHAR:
			instr->op1a = iIMMB;
			instr->l_immediate = *(char *)(op1_data);
			break;
		case T_SHORT:
			instr->op1a = iIMMW;
			instr->l_immediate = *(short *)(op1_data);
			break;
		case T_LONG:
			instr->op1a = iIMMD;
			instr->l_immediate = *(long *)(op1_data);
			break;
		case T_ADDRESS:
			instr->op1a = iIMMD;
			instr->l_immediate = *(ATTRIB_ADDR *)(op1_data);
			break;
		case T_FLOAT:
			break;
		case T_DOUBLE:
			break;
		default:
			break;
		}
	}
	else //memory location
	{
		switch(symboltable->getType(op1_id, op1_defnum))
		{
		case T_CHAR:
			instr->op1a = iROFFB;
			break;
		case T_SHORT:
			instr->op1a = iROFFW;
			break;
		case T_LONG:
			instr->op1a = iROFFD;
			break;
		case T_ADDRESS:
			instr->op1a = iROFFD;
			break;
		case T_FLOAT:
			break;
		case T_DOUBLE:
			break;
		default:
			break;
		}

		//check to see if global or local (different addressing schemes)
		//get address
		TOKEN_ID varparent = symboltable->getParent(op1_id, op1_defnum);
		if(varparent != GLOBALSCOPE)
		{
			//see if member of a class
			if(symboltable->getClassDefnum(varparent) == INVALIDDEFNUM)
			{
				//not member of a class, so must be local
				instr->op1b = iEBP;
				instr->offset = symboltable->getAddress(op1_id, op1_defnum);
			}
			else
			{
				bool need_to_save;
				OPERAND_i86 op = tempstack.allocateRegister(TEMPORARY_TEMP, need_to_save, iNONE);
				tempstack.ReleaseRegister(TEMPORARY_TEMP);

				//mov parameter_size(ebp),reg
				attribute = new i86instruct;
				attribute->instr = iMOV;
				attribute->op1a = iROFFD;
				attribute->op1b = iEBP;
				attribute->offset = currentfunction->parameterlength;
				attribute->op2a = op | iDOUBLEWORD;	//new register
				attribute->next = NULL;

				instr->op1a = iROFFD;
				instr->op1b = op | iDOUBLEWORD;
				instr->offset = symboltable->getAddress(op1_id, op1_defnum);

				//instruction to save register (prepended)
				if(need_to_save)
				{
					i86instruct *temp = new i86instruct;
					temp->instr = iPUSH;
					temp->op1a = op | iDOUBLEWORD;
					temp->next = attribute;
					attribute= temp;
				}
			}
		}
		else //global
		{
			instr->op1a = iGLOBALD;
			instr->func_id = op1_id;
			instr->func_defnum = op1_defnum;
		}

		//check to see if other operand is a memory location also
		if(op2_id != T_TEMPREG || op2_pointer)
		{
			//need to load into a new register, then use that new register for next instruction
			pre_instr = new i86instruct;
			pre_instr->instr = iMOV;
			pre_instr->next = NULL;
			pre_instr->op1a = instr->op1a;
			pre_instr->op1b = instr->op1b;
			pre_instr->op1c = instr->op1c;
			pre_instr->offset = instr->offset;
			pre_instr->func_id = instr->func_id;
			pre_instr->func_defnum = instr->func_defnum;

			bool need_to_save;
			pre_instr->op2a = (tempstack.allocateRegister(-1, need_to_save, iNONE) | iDOUBLEWORD);
			tempstack.ReleaseRegister(-1);	//won't need anymore (since a one time use)

			if(need_to_save)
			{
				i86instruct *temp = new i86instruct;
				temp->instr = iPUSH;
				temp->op1a = pre_instr->op2a;
				temp->next = pre_instr;
				pre_instr = temp;
			}

			//now use this register
			instr->op1a = pre_instr->op2a;
		}
	} //if memory location

	//////////
	//do second operand

	i86instruct *attribute2 = NULL; //used to look up attribute (member variable)
	i86instruct *pre_instr2 = NULL; //used to switch over to a different register


	if(op2_id == T_TEMPREG)
	{
		instr->op2a = (FetchTemp(op2_defnum, pre_instr2, iNONE) | iDOUBLEWORD);

		if(op2_pointer)
		{
			//this is for local variables
			instr->op2b = instr->op2a;
			instr->op2c = iEBP;
			instr->multiplier = 1;
			instr->op2a = iCOMPD;
			instr->offset = 0;
		}

		//check to see if ever used again
		if(LastOccurenceOfTemp(mir, op2_defnum))
			tempstack.ReleaseRegister(op2_defnum);

		//also check if at top of stack -then change ESP
	}
	else if(op2_id & T_IMMEDIATE)
	{
		switch(op2_id & A_BASETYPES)
		{
		case T_CHAR:
			instr->op2a = iIMMB;
			instr->l_immediate = *(char *)(op2_data);
			break;
		case T_SHORT:
			instr->op2a = iIMMW;
			instr->l_immediate = *(short *)(op2_data);
			break;
		case T_LONG:
			instr->op2a = iIMMD;
			instr->l_immediate = *(long *)(op2_data);
			break;
		case T_ADDRESS:
			instr->op2a = iIMMD;
			instr->l_immediate = *(ATTRIB_ADDR *)(op2_data);
			break;
		case T_FLOAT:
			break;
		case T_DOUBLE:
			break;
		}
	}
	else //memory location
	{
		switch(symboltable->getType(op2_id, op2_defnum))
		{
		case T_CHAR:
			instr->op2a = iROFFB;
			break;
		case T_SHORT:
			instr->op2a = iROFFW;
			break;
		case T_LONG:
			instr->op2a = iROFFD;
			break;
		case T_ADDRESS:
			instr->op2a = iROFFD;
			break;
		case T_FLOAT:
			break;
		case T_DOUBLE:
			break;
		default:
			break;
		}

		//check to see if global or local (different addressing schemes
		//get address
		TOKEN_ID varparent = symboltable->getParent(op2_id, op2_defnum);
		if(varparent != GLOBALSCOPE)
		{
			//see if member of a class
			if(symboltable->getClassDefnum(varparent) == INVALIDDEFNUM)
			{
				//not member of a class, so must be local
				instr->op2b = iEBP;
				instr->offset = symboltable->getAddress(op2_id, op2_defnum);
			}
			else
			{
				bool need_to_save;
				OPERAND_i86 op = tempstack.allocateRegister(TEMPORARY_TEMP, need_to_save, iNONE);
				tempstack.ReleaseRegister(TEMPORARY_TEMP);

				//mov parameter_size(ebp),reg
				attribute = new i86instruct;
				attribute->instr = iMOV;
				attribute->op1a = iROFFD;
				attribute->op1b = iEBP;
				attribute->offset = currentfunction->parameterlength;
				attribute->op2a = op | iDOUBLEWORD;	//new register
				attribute->next = NULL;

				instr->op2a = iROFFD;
				instr->op2b = op | iDOUBLEWORD;
				instr->offset = symboltable->getAddress(op2_id, op2_defnum);

				//instruction to save register (prepended)
				if(need_to_save)
				{
					i86instruct *temp = new i86instruct;
					temp->instr = iPUSH;
					temp->op1a = op | iDOUBLEWORD;
					temp->next = attribute;
					attribute = temp;
				}
			}
		}
		else //global
		{
			instr->op2a = iGLOBALD;
			instr->func_id = op2_id;
			instr->func_defnum = op2_defnum;
		}

	}

	//put on preceding-setup instructions

	//prepend pre_instr
	i86instruct *ins;
	if(pre_instr != NULL)
	{
		for(ins = pre_instr; ins->next != NULL; ins = ins->next);
		ins->next = instr;
		instr = pre_instr;
	}

	//prepend attribute
	if(attribute != NULL)
	{
		for(ins = attribute; ins->next != NULL; ins = ins->next);
		ins->next = instr;
		instr = attribute;
	}

	//prepend pre_instr2
	if(pre_instr2 != NULL)
	{
		for(ins = pre_instr2; ins->next != NULL; ins = ins->next);
		ins->next = instr;
		instr = pre_instr2;
	}

	//prepend attribute2
	if(attribute2 != NULL)
	{
		for(ins = attribute2; ins->next != NULL; ins = ins->next);
		ins->next = instr;
		instr = attribute2;
	}
}

void CodeGenerator_i86::FetchOperand(TOKEN_ID op1_id, long op1_defnum, void *op1_data, bool op1_pointer, i86instruct *&instr, MIRInstr *mir, OPERAND_i86 reg)
{
	i86instruct *attribute = NULL; //used to look up attribute (member variable)
	i86instruct *pre_instr = NULL; //used to switch over to a different register

	if(op1_id == T_TEMPREG)
	{
		instr->op1a = (FetchTemp(op1_defnum, pre_instr, reg) | iDOUBLEWORD);

		if(op1_pointer)
		{
			//this is for local variables
			instr->op1b = instr->op1a;
			instr->op1c = iEBP;
			instr->multiplier = 1;
			instr->op1a = iCOMPD;
			instr->offset = 0;
		}

		//check to see if ever used again
		if(LastOccurenceOfTemp(mir->next, op1_defnum))
			tempstack.ReleaseRegister(op1_defnum);
	}
	else if(op1_id & T_IMMEDIATE)
	{
		switch(op1_id & A_BASETYPES)
		{
		case T_CHAR:
			instr->op1a = iIMMB;
			instr->l_immediate = *(char *)(op1_data);
			break;
		case T_SHORT:
			instr->op1a = iIMMW;
			instr->l_immediate = *(short *)(op1_data);
			break;
		case T_LONG:
			instr->op1a = iIMMD;
			instr->l_immediate = *(long *)(op1_data);
			break;
		case T_ADDRESS:
			instr->op1a = iIMMD;
			instr->l_immediate = *(ATTRIB_ADDR *)(op1_data);
			break;
		case T_FLOAT:
			break;
		case T_DOUBLE:
			break;
		default:
			break;
		}
		
		//check to see if requested in a certain register
		if(reg != iNONE)
		{
			bool need_to_save;
			OPERAND_i86 op = tempstack.allocateRegister(TEMPORARY_TEMP, need_to_save, reg);
			tempstack.ReleaseRegister(TEMPORARY_TEMP);

			pre_instr = new i86instruct;
			pre_instr->instr = iMOV;
			pre_instr->op1a = instr->op1a;
			pre_instr->l_immediate = instr->l_immediate;
			pre_instr->op2a = op | iDOUBLEWORD;	//new register
			pre_instr->next = NULL;

			//instruction to save register (prepended)
			if(need_to_save)
			{
				i86instruct *temp = new i86instruct;
				temp->instr = iPUSH;
				temp->op1a = op | iDOUBLEWORD;
				temp->next = pre_instr;
				pre_instr= temp;
			}

			//put this newly allocated register in instruction to return
			instr->op1a = op | iDOUBLEWORD;
		}
	}
	else //memory location
	{
		switch(symboltable->getType(op1_id, op1_defnum))
		{
		case T_CHAR:
			instr->op1a = iROFFB;
			break;
		case T_SHORT:
			instr->op1a = iROFFW;
			break;
		case T_LONG:
			instr->op1a = iROFFD;
			break;
		case T_ADDRESS:
			instr->op1a = iROFFD;
			break;
		case T_FLOAT:
			break;
		case T_DOUBLE:
			break;
		default:
			break;
		}

		//check to see if global or local (different addressing schemes
		//get address
		TOKEN_ID varparent = symboltable->getParent(op1_id, op1_defnum);
		if(varparent != GLOBALSCOPE)
		{
			//see if member of a class
			if(symboltable->getClassDefnum(varparent) == INVALIDDEFNUM)
			{
				//not member of a class, so must be local
				instr->op1b = iEBP;
				instr->offset = symboltable->getAddress(op1_id, op1_defnum);
			}
			else
			{
				bool need_to_save;
				OPERAND_i86 op = tempstack.allocateRegister(TEMPORARY_TEMP, need_to_save, iNONE);
				tempstack.ReleaseRegister(TEMPORARY_TEMP);

				//mov parameter_size(ebp),reg
				attribute = new i86instruct;
				attribute->instr = iMOV;
				attribute->op1a = iROFFD;
				attribute->op1b = iEBP;
				attribute->offset = currentfunction->parameterlength; //add on size of pointer
				attribute->op2a = op | iDOUBLEWORD;	//new register
				attribute->next = NULL;

				instr->op1a = iROFFD;
				instr->op1b = op | iDOUBLEWORD;
				instr->offset = symboltable->getAddress(op1_id, op1_defnum);

				//instruction to save register (prepended)
				if(need_to_save)
				{
					i86instruct *temp = new i86instruct;
					temp->instr = iPUSH;
					temp->op1a = op | iDOUBLEWORD;
					temp->next = attribute;
					attribute = temp;
				}
			}
		}
		else //global
		{
			instr->op1a = iGLOBALD;
			instr->func_id = op1_id;
			instr->func_defnum = op1_defnum;
		}

		//check to see if requested in a certain register
		if(reg != iNONE)
		{
			bool need_to_save;
			OPERAND_i86 op = tempstack.allocateRegister(TEMPORARY_TEMP, need_to_save, reg);
			tempstack.ReleaseRegister(TEMPORARY_TEMP);

			pre_instr = new i86instruct;
			pre_instr->instr = iMOV;
			pre_instr->op1a = instr->op1a;
			pre_instr->op1b = instr->op1b;
			pre_instr->offset = instr->offset;
			pre_instr->func_id = instr->func_id;
			pre_instr->func_defnum = instr->func_defnum;
			pre_instr->op2a = op | iDOUBLEWORD;	//new register
			pre_instr->next = NULL;

			//instruction to save register (prepended)
			if(need_to_save)
			{
				i86instruct *temp = new i86instruct;
				temp->instr = iPUSH;
				temp->op1a = op | iDOUBLEWORD;
				temp->next = pre_instr;
				pre_instr = temp;
			}

			//put this newly allocated register in instruction to return
			instr->op1a = op | iDOUBLEWORD;
		}
	}

	//put on preceding-setup instructions

	i86instruct *ins;

	//prepend pre_instr
	if(pre_instr != NULL)
	{
		for(ins = pre_instr; ins->next != NULL; ins = ins->next);
		ins->next = instr;
		instr = pre_instr;
	}

	//prepend attribute
	if(attribute != NULL)
	{
		for(ins = attribute; ins->next != NULL; ins = ins->next);
		ins->next = instr;
		instr = attribute;
	}
}


