#include "common.h"

#include <stdio.h>

//appends a parameter to instruction text string
void CodeGenerator_i86::concatParam(char *buffer, OPERAND_i86 opa, OPERAND_i86 opb, OPERAND_i86 opc, i86instruct *instr)
{
	switch(opa)
	{
	case iEAX:
		strcat(buffer, "%eax");
		break;
	case iAX:
		strcat(buffer, "%ax");
		break;
	case iAL:
		strcat(buffer, "%al");
		break;
	case iAH:
		strcat(buffer, "%ah");
		break;

	case iEBX:
		strcat(buffer, "%ebx");
		break;
	case iBX:
		strcat(buffer, "%bx");
		break;
	case iBL:
		strcat(buffer, "%bl");
		break;
	case iBH:
		strcat(buffer, "%bh");
		break;

	case iECX:
		strcat(buffer, "%ecx");
		break;
	case iCX:
		strcat(buffer, "%cx");
		break;
	case iCL:
		strcat(buffer, "%cl");
		break;
	case iCH:
		strcat(buffer, "%ch");
		break;

	case iEDX:
		strcat(buffer, "%edx");
		break;
	case iDX:
		strcat(buffer, "%dx");
		break;
	case iDL:
		strcat(buffer, "%dl");
		break;
	case iDH:
		strcat(buffer, "%dh");
		break;

	case iESI:
		strcat(buffer, "%esi");
		break;
	case iSI:
		strcat(buffer, "%si");
		break;

	case iEDI:
		strcat(buffer, "%edi");
		break;
	case iDI:
		strcat(buffer, "%di");
		break;

	case iESP:
		strcat(buffer, "%esp");
		break;
	case iSP:
		strcat(buffer, "%sp");
		break;

	case iEBP:
		strcat(buffer, "%ebp");
		break;
	case iBP:
		strcat(buffer, "%bp");
		break;

/*	case iCS:	-may not use these in linux
	case iDS:
	case iES:
	case iFS:
	case iGS:
*/
	//MEMORY
	case iMEMD:		//double word   offset(offset)
	case iMEMW:		//word			offset(offset)
	case iMEMB:		//byte			offset(offset)
	//***check to see if string instead -then will print string (incase global variable)... optional
		sprintf(&buffer[strlen(buffer)], "%d", instr->offset);
		break;

	case iIMMD:
	case iIMMW:
	case iIMMB:
		sprintf(&buffer[strlen(buffer)], "$%d", instr->l_immediate);
		break;

	//COMPLEX ADDRESSING MODES
	case iROFFD:	//register(opb) + offset(offset)
	case iROFFW:	//register(opb) + offset(offset)
	case iROFFB:	//register(opb) + offset(offset)
		if(instr->offset != 0)
			sprintf(&buffer[strlen(buffer)], "%d(", instr->offset);
		else
			strcat(buffer, "(");
		concatParam(&buffer[strlen(buffer)], opb, 0, 0, instr);
		strcat(buffer, ")");
		break;

	case iCOMPD:	//reg1(opb) + reg2(opc) * mult(multiplier) + offset(offset)
	case iCOMPW:	//reg1(opb) + reg2(opc) * mult(multiplier) + offset(offset)
	case iCOMPB:	//reg1(opb) + reg2(opc) * mult(multiplier) + offset(offset)
		if(instr->offset != 0)
			sprintf(&buffer[strlen(buffer)], "%d(", instr->offset);
		else
			strcat(buffer, "(");
		concatParam(&buffer[strlen(buffer)], opb, 0, 0, instr);
		strcat(buffer, ",");
		concatParam(&buffer[strlen(buffer)], opc, 0, 0, instr);
		strcat(buffer, ",");
		sprintf(&buffer[strlen(buffer)], "%d", instr->multiplier);
		strcat(buffer, ")");
		break;

	case iGLOBALD:	//func_id and func_defnum are the id and defnum of global
	case iGLOBALW:
	case iGLOBALB:
		strcat(buffer, stringtable->getString(symboltable->getNameLookup(instr->func_id, instr->func_defnum)) );
		break;

	default:
		opa = opa;
		break;
	}
}

void CodeGenerator_i86::PrintListing(FILE *outfile)
{
	char buffer[512];

	for(i86instruct *inst = icode; inst != NULL; inst = inst->next)
	{
		switch(inst->instr)
		{
			case iNOP:
				strcpy(buffer, " nop");
				break;
			case iMOV:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " movl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " movw ");
				else
					strcpy(buffer, " movb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iPUSH:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " pushl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " pushw ");
				else
					strcpy(buffer, " pushb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iPOP:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " popl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " popw ");
				else
					strcpy(buffer, " popb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iPUSHF:
				strcpy(buffer, " pushf");
				break;
			case iPOPF:
				strcpy(buffer, " popf");
				break;
			case iPUSHA:
				strcpy(buffer, " pusha");
				break;
			case iPOPA:
				strcpy(buffer, " popa");
				break;
			case iPUSHAD:
				strcpy(buffer, " pushal");
				break;
			case iPOPAD:
				strcpy(buffer, " popal");
				break;
			case iLAHF:
				strcpy(buffer, " lahf");
				break;
			case iSAHF:
				strcpy(buffer, " sahf");
				break;
			case iMOVSX:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " movsxl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " movsxw ");
				else
					strcpy(buffer, " movsxb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iMOVZX:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " movzxl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " movzxw ");
				else
					strcpy(buffer, " movzxb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iLEA:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " leal ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " leaw ");
				else
					strcpy(buffer, " leab ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
/*			linux doesn't use these at the application level
			case iLDS:
			case iLES:
			case iLFS:
			case iLGS:
			case iLSS:
*/
			case iADD:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " addl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " addw ");
				else
					strcpy(buffer, " addb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSUB:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " subl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " subw ");
				else
					strcpy(buffer, " subb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iAND:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " andl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " andw ");
				else
					strcpy(buffer, " andb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iOR:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " orl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " orw ");
				else
					strcpy(buffer, " orb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iXOR:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " xorl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " xorw ");
				else
					strcpy(buffer, " xorb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iADC:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " adcl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " adcw ");
				else
					strcpy(buffer, " adcb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSBB:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " sbbl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " sbbw ");
				else
					strcpy(buffer, " sbbb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iCMP:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " cmpl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " cmpw ");
				else
					strcpy(buffer, " cmpb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iTEST:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " testl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " testw ");
				else
					strcpy(buffer, " testb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iINC:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " incl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " incw ");
				else
					strcpy(buffer, " incb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iDEC:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " decl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " decw ");
				else
					strcpy(buffer, " decb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iNEG:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " negl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " negw ");
				else
					strcpy(buffer, " negb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iNOT:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " notl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " notw ");
				else
					strcpy(buffer, " notb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iMUL:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " mull ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " mulw ");
				else
					strcpy(buffer, " mulb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iIMUL:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " imull ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " imulw ");
				else
					strcpy(buffer, " imulb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iDIV:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " divl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " divw ");
				else
					strcpy(buffer, " divb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iIDIV:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " idivl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " idivw ");
				else
					strcpy(buffer, " idivb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iCBW:
				strcpy(buffer, " cbw");
				break;
			case iCWDE:
				strcpy(buffer, " cwde");
				break;
			case iCWD:
				strcpy(buffer, " cwd");
				break;
			case iCDQ:
				strcpy(buffer, " cdq");
				break;
			case iSHR:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " shrl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " shrw ");
				else
					strcpy(buffer, " shrb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSHL:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " shll ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " shlw ");
				else
					strcpy(buffer, " shlb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSAR:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " sarl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " sarw ");
				else
					strcpy(buffer, " sarb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSAL:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " sall ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " salw ");
				else
					strcpy(buffer, " salb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSHLD:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " shldl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " shldw ");
				else
					strcpy(buffer, " shldb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			case iSHRD:
				if(inst->op1a & iDOUBLEWORD)
					strcpy(buffer, " shrdl ");
				else if(inst->op1a & iWORD)
					strcpy(buffer, " shrdw ");
				else
					strcpy(buffer, " shrdb ");

				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				strcat(buffer, ", ");
				concatParam(buffer, inst->op2a, inst->op2b, inst->op2c, inst);
				break;
			
			case iSETC:
				strcpy(buffer, " setc ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNC:
				strcpy(buffer, " setnc ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETZ:
				strcpy(buffer, " setz ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNZ:
				strcpy(buffer, " setnz ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETS:
				strcpy(buffer, " sets ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNS:
				strcpy(buffer, " setns ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETO:
				strcpy(buffer, " seto ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNO:
				strcpy(buffer, " setno ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETP:
				strcpy(buffer, " setp ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETPE:
				strcpy(buffer, " setpe ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNP:
				strcpy(buffer, " setnp ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETPO:
				strcpy(buffer, " setpo ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETA:
				strcpy(buffer, " seta ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNBE:
				strcpy(buffer, " setnbe ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETAE:
				strcpy(buffer, " setae ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNB:
				strcpy(buffer, " setnb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETB:
				strcpy(buffer, " setb ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNAE:
				strcpy(buffer, " setnae ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETBE:
				strcpy(buffer, " setbe ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNA:
				strcpy(buffer, " setna ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETE:
				strcpy(buffer, " sete ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNE:
				strcpy(buffer, " setne ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETG:
				strcpy(buffer, " setg ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNLE:
				strcpy(buffer, " setnle ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETGE:
				strcpy(buffer, " setge ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNL:
				strcpy(buffer, " setnl ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETL:
				strcpy(buffer, " setl ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNGE:
				strcpy(buffer, " setnge ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETLE:
				strcpy(buffer, " setle ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;
			case iSETNG:
				strcpy(buffer, " setng ");
				concatParam(buffer, inst->op1a, inst->op1b, inst->op1c, inst);
				break;

			case iCALL:
				{
					char *funcname = getFormalFunctionName(inst->func_id, inst->func_defnum);
					sprintf(buffer, " call %s", funcname);
					delete [] funcname;
				}
				break;

			case iJMP:
				sprintf(buffer, " jmp L_%d", inst->op1a);
				break;
			case iJC:
				sprintf(buffer, " jc L_%d", inst->op1a);
				break;
			case iJNC:
				sprintf(buffer, " jnc L_%d", inst->op1a);
				break;
			case iJZ:
				sprintf(buffer, " jz L_%d", inst->op1a);
				break;
			case iJNZ:
				sprintf(buffer, " jnz L_%d", inst->op1a);
				break;
			case iJS:
				sprintf(buffer, " js L_%d", inst->op1a);
				break;
			case iJNS:
				sprintf(buffer, " jns L_%d", inst->op1a);
				break;
			case iJO:
				sprintf(buffer, " jo L_%d", inst->op1a);
				break;
			case iJNO:
				sprintf(buffer, " jno L_%d", inst->op1a);
				break;
			case iJP:
				sprintf(buffer, " jp L_%d", inst->op1a);
				break;
			case iJPE:
				sprintf(buffer, " jpe L_%d", inst->op1a);
				break;
			case iJNP:
				sprintf(buffer, " jnp L_%d", inst->op1a);
				break;
			case iJPO:
				sprintf(buffer, " jpo L_%d", inst->op1a);
				break;
			case iJA:
				sprintf(buffer, " ja L_%d", inst->op1a);
				break;
			case iJNBE:
				sprintf(buffer, " jnbe L_%d", inst->op1a);
				break;
			case iJAE:
				sprintf(buffer, " jae L_%d", inst->op1a);
				break;
			case iJNB:
				sprintf(buffer, " jnb L_%d", inst->op1a);
				break;
			case iJB:
				sprintf(buffer, " jb L_%d", inst->op1a);
				break;
			case iJNAE:
				sprintf(buffer, " jnae L_%d", inst->op1a);
				break;
			case iJBE:
				sprintf(buffer, " jbe L_%d", inst->op1a);
				break;
			case iJNA:
				sprintf(buffer, " jna L_%d", inst->op1a);
				break;
			case iJE:
				sprintf(buffer, " je L_%d", inst->op1a);
				break;
			case iJNE:
				sprintf(buffer, " jne L_%d", inst->op1a);
				break;
			case iJG:
				sprintf(buffer, " jg L_%d", inst->op1a);
				break;
			case iJNLE:
				sprintf(buffer, " jnle L_%d", inst->op1a);
				break;
			case iJGE:
				sprintf(buffer, " jge L_%d", inst->op1a);
				break;
			case iJNL:
				sprintf(buffer, " jnl L_%d", inst->op1a);
				break;
			case iJL:
				sprintf(buffer, " jl L_%d", inst->op1a);
				break;
			case iJNGE:
				sprintf(buffer, " jnge L_%d", inst->op1a);
				break;
			case iJLE:
				sprintf(buffer, " jle L_%d", inst->op1a);
				break;
			case iJNG:
				sprintf(buffer, " jng L_%d", inst->op1a);
				break;

			case iRET:
				strcpy(buffer, " ret");
				break;

			case iLABEL:
				sprintf(buffer, "L_%d:", inst->op1a);
				break;
			case iSTRING:
				if(inst->string != NULL)
					strcpy(buffer, inst->string);
				break;				

			default:
				break;
		}

		fprintf(outfile, "%s\n", buffer);
	}
}

