#!/usr/bin/perl -w
#####################
#
# launch.pl
# this program launches the specified program
# in the command line (full path required).
# it reads in nodes.dat and starts the program
# on all machines in the list

use IO::Socket;
use IO::Select;
require Auth;

$executable = $ARGV[0];
chomp $executable;

open FH, "nodes.dat";

my $instance = 0; #start with first instance in loop

my $launchport = 13013;
my $clientport = 13001; #port that connects with parallel modules
#this is the maximum number of nodes to allocate memory for -can adjust
my $MAX_NUMBER_OF_NODES = 14;

##########
#create server that will eventually connect all the paralel clients
$clientsock = new IO::Socket::INET( #LocalHost not used
			     LocalPort => $clientport,
			     Proto => 'tcp',
			     Reuse => 1,
			     Listen => $MAX_NUMBER_OF_NODES,
			     );
#array of client sockets
my @clientsockets = ();

die "socket could not be created: $!\n" unless $clientsock;



while($target = <FH>)
{
    #don't do blank lines
    next if $target !~ m/\w/;

    #don't do comments
    next if $target =~ m/\/\//;

    chomp $target;

    $instance++;
    print "connecting to $target, assigning instance $instance\n";

    $sock = new IO::Socket::INET( PeerAddr => "$target",
				  PeerPort => $launchport + 0,
				  Proto => 'tcp',
				  Reuse => 1,
				  );
    unless($sock)
    {
	print "Error opening socket on $target:$launchport: $!\n" unless $sock;
	next;
    }

    #do authentication -get value, 'encrypt/scramble', send
    $auth_test = <$sock>;
    my $auth_value = Auth::authenticate_value($auth_test);
    print $sock "$auth_value\n";

    #append network information to give to executable
    $mysockaddr = getsockname($sock);
    my ($myport, $myaddr) = sockaddr_in($mysockaddr);
    my $parameters = "$instance $clientport:" .
	inet_ntoa($myaddr) . ":" . $clientport;

    print "running $executable $parameters\n";

    print $sock "go $executable $parameters\n";
    close $sock;

    #create server and wait for new instance to be ready
    print "waiting for execution confirmation\n";
    $new_sock = $clientsock->accept();
    my $ready;
    read($new_sock, $ready, 1); #get ready character

    print "parallel module ready\n";
    #add to end of list
    push @clientsockets, $new_sock;
   
}

close FH;

#now have all parallel modules ready and connected to this program
#send out message for all modules to connect to each other
sleep 1;
foreach (@clientsockets)
{
    print $_ ('g');
}

#wait for connections to be complete on each one
foreach (@clientsockets)
{
    my $ready;
    read($_, $ready, 1); #get ready character
}

#launch pmain function from instance 1
#do so by sending out a go to all
#instance 1 will know it is supposed to call main
print "beginning program\n";
foreach (@clientsockets)
{
    print $_ ('g');
}



#create list for selecting socket
my $readwatch = new IO::Select();
foreach (@clientsockets)
{
    $readwatch->add($_);
}

#now monitor connections to do i/o and errors for program
for(;;)
{
    ($readlist) = IO::Select->select($readwatch, undef, undef, undef);
    foreach $readsock (@$readlist)
    {
	my $message;
	read($readsock, $message, 1); #get message

	#do something with message
	if($message =~ 's')
	{
	    #print string message
	    #read in byte containing length of string
	    read($readsock, $message, 1);
	    my $msglength = unpack 'C', $message;
	    #read in and print string
	    read($readsock, $message, $msglength);
	    print "|$message";
	}
	elsif($message =~ 'g')
	{
	    #get string message with prompt
	    #read in byte containing length of string
	    read($readsock, $message, 1);
	    my $msglength = unpack 'C', $message;
	    #read in and print string
	    read($readsock, $message, $msglength-1);
	    print "|$message\n";

	    #read in string from user
	    $message = <STDIN>;
	    chomp $message;
	    my $len = length $message;
	    #send byte of string length and string
	    print $readsock (pack('Ca*', (length $message)+0, $message),
			     (length $message) + 1);
	}
	elsif($message =~ 'r')
	{
	    printf("program complete -exiting\n");
	    exec "./stopall.pl";
	}
	else
	{
	    print "ERROR: recived unknown message: '$message'\n";
	}
    }

}

