//this is the stub that gets attached to every program
//it's command line looks like:
//progname instance(number 1..number of machines) local_recieving_port:launcher_ip:launcher_port
//it reads in the file connections.dat to find out its connections

#include "stub.h"

pthread_mutex_t num_activeconn_lock; //used for synchronizing between threads

//number of active connection pairs -needed global because of threads
long num_active_inconn; //number of active input threads
long num_active_outconn; //number of active output threads
//server won't start until num_active_inconn == num_active_outconn
//and both are equal to the total number of connections


//main global queues
ExecutionQueue *executionqueue;
WaitForQueue *waitforqueue;
PeerConnections *peerconn;

int loader_conn; //launcher connection

pthread_t jobmonitor[MAX_EXECQUEUE];
int jobnumber[MAX_EXECQUEUE]; //place to pass in jobmonitor's number
                              //must be global, because otherwise destroyed

//these following 2 functions are called by the parallel module
// to interface with the network
void parallel_call(long func_enum, long wait_id, long call_stack_size)
{
  printf("parallel_call called: func:  %d, %d: ", func_enum, wait_id);

  int i;
  for(i = 4; i <= call_stack_size; i += 4)
    {
      printf("%d ", *(long *)&(((unsigned char *)&call_stack_size)[i]));
    }
  printf("\n");

  //***more stuff here
  //send message to another node (after finding shortest queue length)
  //containing all this info

  //***just send to first connection for now -will add load balancing later
  char buffer[65536];
  buffer[0] = 'j';
  int bufferpos = 1;

  //get wait_id adjusted for threads
  long specific_wait_id = wait_id;
  for(int i = 0; i < MAX_EXECQUEUE; i++)
    if(pthread_equal(jobmonitor[i], pthread_self()))
      {
	wait_id |= (i << 16);
	break;
      }

  //send parameters
  *(long *)(&buffer[bufferpos]) = wait_id;
  bufferpos += 4;
  *(long *)(&buffer[bufferpos]) = func_enum;
  bufferpos += 4;
  *(long *)(&buffer[bufferpos]) = call_stack_size;
  bufferpos += 4;

  for(i = 4; i <= call_stack_size; i+= 4)
    {
      *(long *)(&buffer[bufferpos])
	= *(long *)&(((unsigned char *)&call_stack_size)[i]);
      bufferpos += 4;
    }
  int minsocket = peerconn->sendSocketWithShortQueue();
  printf("sent to socket handle %d\n", minsocket);
  write(peerconn->sendSocketWithShortQueue(), buffer, bufferpos);
}

long wait_parallel_call(long wait_id)
{
  //get wait_id adjusted for threads
  long specific_wait_id = wait_id;
  for(int i = 0; i < MAX_EXECQUEUE; i++)
    if(pthread_equal(jobmonitor[i], pthread_self()))
      {
	wait_id |= (i << 16);
	break;
      }

  printf("waiting... wait_id: %d\n", wait_id);

  while(!waitforqueue->inQueue(specific_wait_id));
  printf("data in queue, now fetching\n");
  return waitforqueue->getReturnData(specific_wait_id);
}

//this is the heart of the program
//it calls the parallel functions and returns out the respective
//values
void *ProcessQueues(void *arg)
{
  int jobnum = *(int *)arg;
  //run until we are stopped
  for(;;)
    {
      //get next job
      ExecQueueMember *j = executionqueue->getJob(jobnum);
      printf("jobmanager %d got job func: %d, datalen: %d\n",
	     jobnum, j->func_enum, j->data_length);

      //do job
      long retval = call_enum_func(j->func_enum, j->data_length, j->data);
      delete [] j->data;

      printf("got job finished.  retval: %d\n", retval);
      executionqueue->markJobComplete(jobnum);

      //send return value to where it came from
      char buffer[256];
      buffer[0] = 'r';
      *((long *)&buffer[1]) = j->wait_id;
      *((long *)&buffer[1+sizeof(long)]) = retval;

      int returnsocket = peerconn->matchingSendSocket(j->connection);
      printf("sending return to socket %d\n", returnsocket);

      //check if returning to launcher (program done)
      if(returnsocket == 0)
	returnsocket = loader_conn;

      write(returnsocket, buffer, 1 + 2*sizeof(long));
    }
  return NULL;
}


//this function accepts a string of local_port:dest_ip:dest_port
//and makes a connection to the distination server (ignoring local_port)
//and returns the socket
//remember to close(socket) for cleanup
int CreateClientConnection(char *descriptor)
{
  int clientSocket,
    remotePort,
    status = 0;
  struct hostent *hostPtr = NULL;
  struct sockaddr_in serverName = { 0 };
  char *remoteHost = NULL;

  //separate information from descriptor
  char buffer[256]; //this must be global with respect to function
  char *port_string; //pointer to get port from
  strcpy(buffer, descriptor);

  //get first colon
  remoteHost = strchr(descriptor, ':');
  remoteHost++; //move past colon
  //get second colon, and replace with 
  port_string = strchr(remoteHost, ':');
  *port_string = '\0'; //terminate remoteHost string
  port_string++; //go to beginning of port string

  remotePort = atoi(port_string);

  printf("creating connection to %s:%d\n", remoteHost, remotePort);

  //create socket
  clientSocket = socket(PF_INET, SOCK_STREAM, 
			IPPROTO_TCP);
  if(-1 == clientSocket)
    {
      perror("socket()");
      //      exit(1);
    }

  //need to resolve the remote server name or ip address
  hostPtr = gethostbyname(remoteHost);
  if(NULL == hostPtr)
    {
      hostPtr = gethostbyaddr(remoteHost, 
			      strlen(remoteHost), AF_INET);
      if(NULL == hostPtr)
        {
	  perror("Error resolving server address");
	  //	  exit(1);

        }
    }

  serverName.sin_family = AF_INET;
  serverName.sin_port = htons(remotePort);
  (void) memcpy(&serverName.sin_addr, 
                hostPtr->h_addr,
                hostPtr->h_length);
  
  status = connect(clientSocket, 
		   (struct sockaddr*) &serverName,
		   sizeof(serverName));
  if(-1 == status)
    {
      perror("connect()");
      //exit(1);
    }

    /*
     * Client application specific code goes here
     *
     * e.g. receive messages from server, respond,
     * etc. 
     while (0 < (status = read(clientSocket,
     buffer, sizeof(buffer) - 1)))
     {
     printf("%d: %s", status, buffer);
     }
     
     if (-1 == status)
     {
     perror("read()");
     }
    */

  //we now have a connection to here!
  

  return clientSocket;
}


//sub-thread that monitors a given connection for new jobs and return
//values from jobs sent out (writes into respective queue)
void MonitorConnectionForQueues(int socket)
{
  char buffer[65536];
  int bufferpos = 0; //buffer position
  int status = 0;

  //always monitor connections
  //may put a flag to stop monitoring connections? depending on communications?
  for(;;)
  {
    //get message type
    while(read(socket, buffer, 1) == 0);
    switch(buffer[0])
      {
      case 'j': //then recieve long wait_id, long func_enum, long data_length, then data
	{
	  printf("job recieved\n");
	  bufferpos = 0;
	  //read in 3 long-sized parameters
	  do
	    {
	      bufferpos += read(socket, &buffer[bufferpos], 3*sizeof(long) - bufferpos);
	    } while(bufferpos < 3*sizeof(long));

	  ExecQueueMember q;
	  q.connection = socket;
	  q.wait_id = ((long *)buffer)[0];
	  q.func_enum = ((long *)buffer)[1];
	  q.data_length = ((long *)buffer)[2];
	  printf("job info: func: %d, wait_id: %d, data_len %d\n",
		 q.func_enum, q.wait_id, q.data_length);

	  //get data
	  q.data = new long[q.data_length];
	  bufferpos = 0;
	  do
	    {
	      bufferpos += read(socket, &q.data[bufferpos],
				q.data_length - bufferpos);
	    } while(bufferpos < q.data_length);

	  //add new return value to queue
	  executionqueue->addNewJob(&q);

	}
	break;

      case 'r': //then recieve long wait_id, long return_data
	{
	  bufferpos = 0;
	  //read in 2 long-sized parameters
	  do
	    {
	      bufferpos += read(socket, &buffer[bufferpos], 2*sizeof(long) - bufferpos);
	    } while(bufferpos < 2*sizeof(long));
	  
	  //add new return value to queue
	  waitforqueue->StoreReturnValue( ((long *)buffer)[0],
					  ((long *)buffer)[1] );
	}
	break;

      case 'p':
	//this will not recieve 'p', because it is a response
	//sent back up the connection it came from
	break;

      case 'q': //then recieves nothing -respond back to same connection
	{
	  printf("recieved queue query\n");
	  buffer[0] = 'p'; //response
	  *(long *)(&buffer[1]) = executionqueue->getQueueLength();
	  write(socket, buffer, 1 + sizeof(long));
	}
	break;

      default:
	perror("recieved invalid message\n");
	break;
      }
  }
}

//this waits for server to be connected
//arguement is pointer to where to recieve serverSocket
//and where to put slaveSocket when done
//once connection is made, calls MonitorConnectionForQueues
//to keep track of new jobs coming in
void *AcceptServerConnection(void *arg)
{
  //get the connection number
  int serverSocket = *(int *)arg;

  struct sockaddr_in clientName = { 0 };
  int slaveSocket, clientLength = 
    sizeof(clientName);

  (void) memset(&clientName, 0, 
		sizeof(clientName));

  printf("monitoring port\n");
  printf("going into accept func\n");
  slaveSocket = accept(serverSocket,
		       (struct sockaddr *) &clientName, 
		       (socklen_t *)&clientLength);
  if(-1 == slaveSocket)
    {
      perror("accept()");
      //exit(1);
    }

  close(serverSocket);

  if(-1 == getpeername(slaveSocket,
		       (struct sockaddr *) &clientName, 
		       (socklen_t *)&clientLength))
    {
      perror("getpeername()");
    }
  else
    {
      printf("Connection request from %s\n",
	     inet_ntoa(clientName.sin_addr));
    }

  /* server application specific code
             * goes here, e.g. perform some 
             * action, respond to client etc.
             *
            write(slaveSocket, MESSAGE, 
			strlen(MESSAGE));
            close(slaveSocket);
            exit(0);
  */

  //need to put result here for external use (when destroying)
  *(int *)arg = slaveSocket;

  //add another active socket
  pthread_mutex_lock(&num_activeconn_lock);
  num_active_inconn++;
  pthread_mutex_unlock(&num_activeconn_lock);

  //once connection is made, we want to monitor it
  MonitorConnectionForQueues(slaveSocket);

  //need to return something
  return arg;
}



int main(int argc, char* argv[])
{
  //check for parameters
  if(argc == 1)
    {
      printf("\nUsage: should be executed by launcher, not manually\n");
      printf(" parameters: instance(number 1..number of machines)\n");
      printf("   local_recieving_port:launcher_ip:launcher_port\n");
      return 0;
    }

  char *instance_parameter = argv[1];
  char *connection_parameter = argv[2];
  int instance = atoi(instance_parameter);

  pthread_mutex_init(&num_activeconn_lock, NULL);

  //set up queues
  executionqueue = new ExecutionQueue;
  waitforqueue = new WaitForQueue;

  //get connections ready to listen
  peerconn = new PeerConnections(instance);

  //make connection to loader
  loader_conn = CreateClientConnection(connection_parameter);

  //send ready signal
  write(loader_conn, "g", 1);

  //wait for the rest of the modules to be loaded
  char buffer[256]; //make big just incase something weird
  while(read(loader_conn, buffer, 1) == 0);
  if(buffer[0] != 'g')
    printf("error value from loader while waiting on other modules\n");

  printf("recieved ready from server\n");
  printf("now making connections to other parallel modules\n");
  //make connections -each has its own monitoring thread
  //and increment num_activeconn for each connection made
  peerconn->createPeerConnections();

  //wait for all connections to be connected
  peerconn->waitForConnections();

  //announce all connections made
  write(loader_conn, "r", 1);
  printf("all connections ready\n");

  //wait for announcement of beginning of program
  while(read(loader_conn, buffer, 1) == 0);
  if(buffer[0] != 'g')
    printf("error value from loader while beginning program\n");

  //if it's instance 1, we must call pmain() by adding it to the job list
  //and specifying its caller as the launcher
  if(instance == 1)
    {
      //call pmain() by adding it to the job list
      ExecQueueMember pmain;
      pmain.connection = loader_conn;
      pmain.wait_id = 0;
      pmain.func_enum = 0;
      pmain.data_length = 0; //this could change if parameters???
      pmain.data = NULL;

      executionqueue->addNewJob(&pmain);
      printf("beginning program\n");
    }

  //run one execution thread,
  //and keep connection monitors going (they feed execution)
  //and wait for ending signal from loader (to know when done)
  for(int i = 1; i < MAX_EXECQUEUE; i++)
    {
      jobnumber[i] = i;
      pthread_create(&jobmonitor[i], NULL, ProcessQueues, &jobnumber[i]);
    }

  jobnumber[0] = 0;
  ProcessQueues(&jobnumber[0]);

  printf("exiting parallel module\n");

  close(loader_conn);

  delete peerconn;
  delete executionqueue;
  delete waitforqueue;
  pthread_mutex_destroy(&num_activeconn_lock);
  return 1;
}


//local replacement of gethostname() to aid portability
int _GetHostName(char *buffer, int length)
{
    struct utsname sysname = { 0 };
    int status = 0;

    status = uname(&sysname);
    if(-1 != status)
    {
        strncpy(buffer, sysname.nodename, length);
    }

    return status;
}
