#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <sys/socket.h>
#include <unistd.h>               /* read() */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <pthread.h>


//queue sizes _MUST_ be powers of 2
#define MAX_CONNECTIONS 256
#define MAX_EXECQUEUE 8
#define MAX_WAITQUEUE 256

const int BACK_LOG = 5;


//global variable prototypes (found in stub.cpp)
extern long num_active_inconn; //number of active input threads
extern long num_active_outconn; //number of active output threads

//launcher connection
extern int loader_conn;

//prints a message to the launcher
void putstring(char *str);


int _GetHostName(char *buffer, int length);

//for queue of list of things to do
struct ExecQueueMember
{
  pthread_mutex_t joblock; //used for synchronizing between threads
  int active; //boolean: weather or not it's currently in use
  int connection; //handle of connection to return to
  long wait_id;   //wait id to return with it
  long func_enum; //number of function to call
  long data_length; //length of data to be pushed on stack as number of longs
  long *data;  //data to be pushed onto stack for enumfunction call
};

class ExecutionQueue
{
public:
  ExecutionQueue();
  ~ExecutionQueue();

  //returns a pointer to the next job (waits for next job if needs to)
  //removes job from queue
  //note: the data must be freed from memory, otherwise it is a memory leak
  // but do not attempt to free the node itself
  ExecQueueMember *getJob(int jobnum);
  void markJobComplete(int jobnum);

  //called by connection monitor, which watches for new jobs
  //it recieves an ExecQueueMember structure, and copies data from it
  //but uses the same data pointer, so the structure can be deleted
  //if it's allocated, but not the data pointer
  //(this class will take care of freeing the memory)
  //places it in the next jobnum avaliable
  void addNewJob(ExecQueueMember *j);

  //returns queue length for load balancing
  long getQueueLength();

private:
  ExecQueueMember queue[MAX_EXECQUEUE];
  pthread_mutex_t broadcast_lock;
  pthread_cond_t broadcast;
};

//used by waitfor function -filled in
struct WaitQueueMember
{
  int active; //true if active data, false if empty
  long wait_id; //wait_id weighted with thread number
  long return_data; //function's return data
};


class WaitForQueue
{
public:
  WaitForQueue();
  ~WaitForQueue();

  //returns true if the waited for value is in the queue
  bool inQueue(long wait_id);

  //returns the data from the wait queue and then removes it
  //poll with inQueue() until it returns true, then call this function
  long getReturnData(long wait_id);

  //adds the return data to the list under wait_id
  //returns true if succedes, false if out of space (error condition)
  bool StoreReturnValue(long wait_id, long return_data);

private:

  WaitQueueMember queue[MAX_WAITQUEUE];
  pthread_mutex_t queuelock; //used for synchronizing between threads

};


class PeerConnections
{
public:
  PeerConnections() { }
  PeerConnections(int instance);
  ~PeerConnections();

  //waits until num_activeconn == num_conn
  void waitForConnections();

  //creates connections to all of peers
  void createPeerConnections();

  //returns the send socket pared with that read descriptor
  int matchingSendSocket(int socket);

  //returns a handle to the socket of the connection with
  //shortest queue
  int sendSocketWithShortQueue();

private:

  //creates a new server and stores it in connection array
  //also creates thread to monitor input, and to append to queue
  void createNewServer(char *descriptor);

  //number of connection pairs
  long num_conn;

  //connection (socket) of each connection pair
  //every 2 sockets is a connection
  //orderded as recieving, sending
  //allow up to MAX_CONNECTIONS*2 connections
  int connection[MAX_CONNECTIONS*2];
  //text descriptors to keep for creating client connections
  char *conn_descriptor[MAX_CONNECTIONS];

  //array of threads to monitor client connections (wait for input)
  //technically, it is a server, because it is connected to, however
  pthread_t clientmonitor[MAX_CONNECTIONS];

};

//global functions
int CreateClientConnection(char *descriptor);
void MonitorConnectionForQueues(int socket);
void *AcceptServerConnection(void *arg);

//global function from parallel code
long call_enum_func(long func_enum, long data_len, long *data);


