#include "common.h"
#include <stdio.h>


MIRTransform::MIRTransform(StringTable *st, SymbolTable *yt)
{
	stringtable = st;
	symboltable = yt;
	nexttemp = 0;
	nextlabel = 0;
	_zero = 0; //need a pointer to 0
	_neg_one = 0xFFFFFFFF;

	current_wait_id = 0;

	inparallel = false;	//don't want to begin by parallizing things -only where specified
}



MIRTransform::~MIRTransform()
{
}


void MIRTransform::Transform(FunctionType *func)
{
	currentfunction = func; //save this for other use (i.e. returning, etc.)

	//put in dummy in first position to make data manipulation easier (only one case then)
	func->mircode = new MIRInstr;
	memset(func->mircode, 0, sizeof(MIRInstr));

	nexttemp = 0;	//begin with a new function and new temps

	TOKEN_ID dummy;	//place holders for returning values from translateASTtoMIR if unused
	long dummy2;
	bool dummy3;

	func->mircode->next = translateASTtoMIR(func->function, dummy, dummy2, dummy3, false);
	//finish up things like connecting link list's other direction
	PolishMIR(func->mircode);

//***WILL NEED EXTRA INFO ABOUT STUFF LIKE STACK SPACE FROM FunctionType
}

void MIRTransform::MirPrint3PartInstr(MIRInstr *func, char *buffer, char *instr)
{
	char op1b[64];
	char op2b[64];
	char dstb[64];

	//write out each of the operands
	if(func->op1_id == T_TEMPREG)
		sprintf(op1b, "t%d", func->op1_defnum);
	else if(func->op1_id & T_IMMEDIATE)
	{
		switch(func->op1_id & A_BASETYPES)
		{
		case T_DOUBLE:
			sprintf(op1b, "%f", *(double *)func->op1_data);
			break;
		case T_FLOAT:
			sprintf(op1b, "%f", *(float *)func->op1_data);
			break;
		case T_ADDRESS:
			sprintf(op1b, "&%d", *(long *)func->op1_data);
			break;
		default:
			sprintf(op1b, "%ld", *(long *)func->op1_data);
		}
	}
	else if(func->op1_id >= (unsigned long)T_IDENTIFIERSTART)
		sprintf(op1b, "%s,%d", stringtable->getString(symboltable->getNameLookup(func->op1_id, func->op1_defnum)), func->op1_defnum);

	if(func->op1_pointer)
	{
		char tempbuffer[64];
		memcpy(tempbuffer, op1b, 64);
		sprintf(op1b, "[%s]", tempbuffer);
	}

	if(func->op2_id == T_TEMPREG)
		sprintf(op2b, "t%d", func->op2_defnum);
	else if(func->op2_id & T_IMMEDIATE)
	{
		switch(func->op2_id & A_BASETYPES)
		{
		case T_DOUBLE:
			sprintf(op2b, "%f", *(double *)func->op2_data);
			break;
		case T_FLOAT:
			sprintf(op2b, "%f", *(float *)func->op2_data);
			break;
		case T_ADDRESS:
			sprintf(op2b, "&%d", *(long *)func->op2_data);
			break;
		default:
			sprintf(op2b, "%ld", *(long *)func->op2_data);
		}
	}
	else if(func->op2_id >= (unsigned long)T_IDENTIFIERSTART)
		sprintf(op2b, "%s,%d", stringtable->getString(symboltable->getNameLookup(func->op2_id, func->op2_defnum)), func->op2_defnum);

	if(func->op2_pointer)
	{
		char tempbuffer[64];
		memcpy(tempbuffer, op2b, 64);
		sprintf(op2b, "[%s]", tempbuffer);
	}

	if(func->dest_id == T_TEMPREG)
		sprintf(dstb, "t%d", func->dest_defnum);
	else
		sprintf(dstb, "%s,%d", stringtable->getString(symboltable->getNameLookup(func->dest_id, func->dest_defnum)), func->dest_defnum);

	if(func->dest_pointer)
	{
		char tempbuffer[64];
		memcpy(tempbuffer, dstb, 64);
		sprintf(dstb, "[%s]", tempbuffer);
	}

	sprintf(buffer, " %s <- %s %s %s\n", dstb, op1b, instr, op2b);
}

void MIRTransform::PrintListing(FILE *outfile, MIRInstr *func)
{
	//print its name
	fprintf(outfile, "sub: %s\n", stringtable->getString(
		symboltable->getNameLookup(symboltable->getCurFunctionID(),
									symboltable->getCurFunctionDefnum() ) ) );
	if(func == NULL)
		return;

	for(func = func->next /*assume first dummy*/; func != NULL; func = func->next)
	{
		char buffer[512];

		switch(func->instr)
		{
		case T_IF:
			{
				char op1b[64];
				char op2b[64];
				char dest[64];

				//write out each of the operands
				if(func->op1_id == T_TEMPREG)
					sprintf(op1b, "t%d", func->op1_defnum);
				else if(func->op1_id & T_IMMEDIATE)
				{
					switch(func->op1_id & A_BASETYPES)
					{
					case T_DOUBLE:
						sprintf(op1b, "%f", *(double *)func->op1_data);
						break;
					case T_FLOAT:
						sprintf(op1b, "%f", *(float *)func->op1_data);
						break;
					case T_ADDRESS:
						sprintf(op1b, "&%d", *(long *)func->op1_data);
						break;
					default:
						sprintf(op1b, "%ld", *(long *)func->op1_data);
					}
				}
				else if((signed long)func->op1_id >= (signed long)T_IDENTIFIERSTART)
					sprintf(op1b, "%s,%d", stringtable->getString(symboltable->getNameLookup(func->op1_id, func->op1_defnum)), func->op1_defnum);

				if(func->op1_pointer)
				{
					char tempbuffer[64];
					memcpy(tempbuffer, op1b, 64);
					sprintf(op1b, "[%s]", tempbuffer);
				}

				if(func->op2_id == T_TEMPREG)
					sprintf(op2b, "t%d", func->op2_defnum);
				else if(func->op2_id & T_IMMEDIATE)
				{
					switch(func->op2_id & A_BASETYPES)
					{
					case T_DOUBLE:
						sprintf(op2b, "%f", *(double *)func->op2_data);
						break;
					case T_FLOAT:
						sprintf(op2b, "%f", *(float *)func->op2_data);
						break;
					case T_ADDRESS:
						sprintf(op2b, "&%d", *(long *)func->op2_data);
						break;
					default:
						sprintf(op2b, "%ld", *(long *)func->op2_data);
					}
				}
				else if((signed long)func->op2_id >= (signed long)T_IDENTIFIERSTART)
					sprintf(op2b, "%s,%d", stringtable->getString(symboltable->getNameLookup(func->op2_id, func->op2_defnum)), func->op2_defnum);

				if(func->op2_pointer)
				{
					char tempbuffer[64];
					memcpy(tempbuffer, op2b, 64);
					sprintf(op2b, "[%s]", tempbuffer);
				}

				switch(func->dest_id)
				{
					case T_PLUS:		strcpy(dest, "+");	break;
					case T_MINUS:		strcpy(dest, "-");	break;
					case T_MULTIPLY:	strcpy(dest, "*");	break;
					case T_DIVIDE:		strcpy(dest, "/");	break;
					case T_MODULUS:		strcpy(dest, "%");	break;
					case T_SHIFTLEFT:	strcpy(dest, "<<");	break;
					case T_SHIFTRIGHT:	strcpy(dest, ">>");	break;
					case T_BXOR:		strcpy(dest, "^");	break;
					case T_BOR:			strcpy(dest, "|");	break;
					case T_BAND:		strcpy(dest, "&");	break;
					case T_EQUALS:		strcpy(dest, "==");	break;
					case T_LESS:		strcpy(dest, "<");	break;
					case T_LEQUAL:		strcpy(dest, "<=");	break;
					case T_GREATER:		strcpy(dest, ">");	break;
					case T_GEQUAL:		strcpy(dest, ">=");	break;
					case T_NOTEQUAL:	strcpy(dest, "!=");	break;
				}
				sprintf(buffer, " if %s %s %s goto %d\n", op1b, dest, op2b, func->target);
			}
			break;

		case T_RETURN:
		case T_PUSHPARAMETER:
		case T_PULLPARAMETER:
		case T_GLOBALSPACE:
			{
				char op1b[64];
				if(func->op1_id == T_TEMPREG)
					sprintf(op1b, "t%d", func->op1_defnum);
				else if((signed long)func->op1_id >= (signed long)T_IDENTIFIERSTART)
					sprintf(op1b, "%s,%d", stringtable->getString(symboltable->getNameLookup(func->op1_id, func->op1_defnum)), func->op1_defnum);
				else
				{
					switch(func->op1_id & A_BASETYPES)
					{
					case T_DOUBLE:
						sprintf(op1b, "%f", *(double *)func->op1_data);
						break;
					case T_FLOAT:
						sprintf(op1b, "%f", *(float *)func->op1_data);
						break;
					case T_ADDRESS:
						sprintf(op1b, "&%d", *(long *)func->op1_data);
						break;
					case T_VOID:
						op1b[0] = '\0';
						break;
					default:
						sprintf(op1b, "%ld", *(long *)func->op1_data);
						break;
					}
				}

				if(func->op1_pointer)
				{
					char tempbuffer[64];
					memcpy(tempbuffer, op1b, 64);
					sprintf(op1b, "[%s]", tempbuffer);
				}

				if(func->instr == T_RETURN)
					sprintf(buffer, " return %s\n", op1b);
				else if(func->instr == T_PUSHPARAMETER)
					sprintf(buffer, " push %s\n", op1b);
				else if(func->instr == T_PULLPARAMETER)
					sprintf(buffer, " pull %s\n", op1b);
				else if(func->instr == T_GLOBALSPACE)
					sprintf(buffer, " globalspace %s\n", op1b);
			}
			break;
		case T_FUNCTIONCALL:
			{
				sprintf(buffer, " call %s,%d\n", stringtable->getString(symboltable->getNameLookup(func->dest_id, func->dest_defnum)), func->dest_defnum);
			}
			break;
		case T_PFUNCTIONCALL:
			{
				sprintf(buffer, " pcall %s,%d  wait_id: %d\n", stringtable->getString(symboltable->getNameLookup(func->dest_id, func->dest_defnum)), func->dest_defnum, func->target);
			}
			break;
		case T_WAITRETURN:
			{
				sprintf(buffer, " waitfor %s,%d  wait_id: %d\n", stringtable->getString(symboltable->getNameLookup(func->dest_id, func->dest_defnum)), func->dest_defnum, func->target);
			}
			break;
		case T_PLUS:
			MirPrint3PartInstr(func, buffer, "+");
			break;
		case T_MINUS:
			MirPrint3PartInstr(func, buffer, "-");
			break;
		case T_MULTIPLY:
			MirPrint3PartInstr(func, buffer, "*");
			break;
		case T_DIVIDE:
			MirPrint3PartInstr(func, buffer, "/");
			break;
		case T_MODULUS:
			MirPrint3PartInstr(func, buffer, "%");
			break;
		case T_SHIFTLEFT:
			MirPrint3PartInstr(func, buffer, "<<");
			break;
		case T_SHIFTRIGHT:
			MirPrint3PartInstr(func, buffer, ">>");
			break;
		case T_BXOR:
			MirPrint3PartInstr(func, buffer, "^");
			break;
		case T_BOR:
			MirPrint3PartInstr(func, buffer, "|");
			break;
		case T_BAND:
			MirPrint3PartInstr(func, buffer, "&");
			break;
		case T_EQUALS:
			MirPrint3PartInstr(func, buffer, "==");
			break;
		case T_LESS:
			MirPrint3PartInstr(func, buffer, "<");
			break;
		case T_LEQUAL:
			MirPrint3PartInstr(func, buffer, "<=");
			break;
		case T_GREATER:
			MirPrint3PartInstr(func, buffer, ">");
			break;
		case T_GEQUAL:
			MirPrint3PartInstr(func, buffer, ">=");
			break;
		case T_NOTEQUAL:
			MirPrint3PartInstr(func, buffer, "!=");
			break;
		case T_LABEL:
			sprintf(buffer, "%d:\n", func->target);
			break;
		case T_GOTO:
			sprintf(buffer, " goto %d\n", func->target);
			break;
		case T_ARRAYRESIZE:
			break;
		case T_ADDRESSOF:
			sprintf(buffer, " loadaddr t%d %s,%d\n", func->dest_defnum, stringtable->getString(symboltable->getNameLookup(func->op1_id, func->op1_defnum)), func->op1_defnum);
			break;
		case T_PUSH_THIS:
			sprintf(buffer, " push this\n");
			break;
		default:
			sprintf(buffer, "error\n");
			break;
		}
		fprintf(outfile, "%s", buffer);
	}
	fprintf(outfile, "\n");
}


//builds from right to left (or end to front)
MIRInstr *MIRTransform::translateASTtoMIR(ASTNode *curnode, TOKEN_ID &eval_id, long &eval_defnum, bool &eval_pointer, bool request_pointer)
{
	if(curnode == NULL)
		return NULL;

	if(curnode->op >= T_IDENTIFIERSTART)
	{
		//special case -must get what's at there, because it's a pointer
		if(curnode->type & A_RESIZEABLE)
		{
			MIRInstr *tgets = new MIRInstr;
			tgets->instr = T_PLUS;
			tgets->target = 0;
			tgets->next = NULL;
			//shortcut for gets: use x = y + 0
			//optimizer will take care of this redundancy if necesary

			//put pointer on left side of addition
			tgets->op1_id = curnode->op;
			tgets->op1_defnum = curnode->defnum;
			tgets->op1_pointer = false;

			//put zero to be added into op2, with size/type of op1
			tgets->op2_id = T_ADDRESS | T_IMMEDIATE;
			tgets->op2_defnum = 0;
			tgets->op2_data = &_zero;
			tgets->op2_pointer = false;

			eval_id = tgets->dest_id = T_TEMPREG;
			eval_defnum = tgets->dest_defnum = ++nexttemp;
			eval_pointer = tgets->dest_pointer = false;

			return tgets;
		}

		if(request_pointer || (curnode->type & A_USERTYPE) || (curnode->type & A_ARRAY) )
		{
			//check to see if global or local -if global, need to keep name
			if(symboltable->getParent(curnode->op, curnode->defnum) == GLOBALSCOPE)
			{
				MIRInstr *addr = new MIRInstr;
				addr->instr = T_ADDRESSOF;
				addr->target = 0;
				addr->next = NULL;

				addr->op1_id = curnode->op;
				addr->op1_defnum = curnode->defnum;
				addr->op1_pointer = false;

				eval_id = addr->dest_id = T_TEMPREG;
				eval_defnum = addr->dest_defnum = ++nexttemp;
				eval_pointer = addr->dest_pointer = false;

				return addr;
			}

			//local, so we can just grab address
			eval_id = T_ADDRESS | T_IMMEDIATE;
			eval_defnum = symboltable->getAddress(curnode->op, curnode->defnum);
			eval_pointer = false;
			return NULL;
		}

		eval_id = curnode->op;
		eval_defnum = curnode->defnum;
		eval_pointer = false;
		return NULL;
	}


	switch(curnode->op)
	{
	case T_SEMICOLON:
		{
			//traverse right first and append
			MIRInstr *nextblock = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);
			//if other side is NULL, we're done
			MIRInstr *frontblock = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			return appendMIRtoMIR(frontblock, nextblock);
		}

	case T_PARALLEL:
		{
			MIRInstr *func_precalls = translateASTFuncPrecallsToMIR(curnode->child[1]);
			inparallel = true;
			MIRInstr *parablock = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);
			inparallel = false;

			parablock = appendMIRtoMIR(func_precalls, parablock);

			MIRInstr *regblock = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);

			return appendMIRtoMIR(regblock, parablock);
		}

	case T_CHAR: 
	case T_SHORT:
	case T_LONG:
	case T_FLOAT:
	case T_DOUBLE:
		{
			//no additional code required for immediate, only value
			//must flag as immediate
			eval_id = curnode->op | T_IMMEDIATE;
			eval_defnum = 0;
			eval_pointer = false;
			return NULL;
		}

	case T_GETS:
		{
			MIRInstr *tgets = new MIRInstr;
			tgets->instr = T_PLUS;
			tgets->target = 0;
			tgets->next = NULL;
			//shortcut for gets: use x = y + 0
			//optimizer will take care of this redundancy if necesary

			//traverse r-side, then l-side (evaluate then assign)
			MIRInstr *r_side = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);
			tgets->op1_id = eval_id;
			tgets->op1_defnum = eval_defnum;
			if((signed long)tgets->op1_id < (signed long)T_TEMPREG)
			{
				tgets->op1_id |= T_IMMEDIATE;
				tgets->op1_data = (void *)curnode->child[1]->data;
			}
			tgets->op1_pointer = false;

			//put zero to be added into op2, with size/type of op1
			tgets->op2_id = curnode->child[1]->type | T_IMMEDIATE;
			tgets->op2_defnum = 0;
			tgets->op2_data = &_zero;
			tgets->op2_pointer = false;

			//traverse left side
			MIRInstr *l_side = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			tgets->dest_id = eval_id;
			tgets->dest_defnum = eval_defnum;
			tgets->dest_pointer = eval_pointer;
			//eval_xxx contains correct info

			return appendMIRtoMIR( appendMIRtoMIR(r_side, l_side), tgets);
		}

	////////unary ops
	case T_UNARYPLUS:
		if(curnode->child[0] != NULL)
			return translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
		else return NULL;

	case T_UNARYMINUS:
		{
			//build an instruction as t = 0 - x
			MIRInstr *result = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			MIRInstr *tmin = new MIRInstr;
			tmin->instr = T_MINUS;

			//op1 is zero
			tmin->op1_id = curnode->child[0]->type | T_IMMEDIATE;
			tmin->op1_defnum = 0;
			tmin->op1_data = &_zero;
			tmin->op1_pointer = false;

			//op2 is value
			tmin->op2_id = eval_id;
			tmin->op2_defnum = eval_defnum;
			if((signed long)tmin->op2_id < (signed long)T_TEMPREG)
			{
				tmin->op2_id |= T_IMMEDIATE;
				tmin->op2_data = (void *)curnode->child[0]->data;
			}
			tmin->op2_pointer = eval_pointer;

			tmin->target = 0;
			tmin->next = NULL;

			//destination is next avaliable "temporary" register
			eval_id = tmin->dest_id = T_TEMPREG;
			eval_defnum = tmin->dest_defnum = ++nexttemp;
			eval_pointer = tmin->dest_pointer = false;
			return appendMIRtoMIR(result, tmin);
		}

	case T_BNOT:
		{
			//binary not, same as t = 0xFFFFF... ^ x
			MIRInstr *result = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			MIRInstr *tnot = new MIRInstr;
			tnot->instr = T_BXOR;

			//op1 is zero
			tnot->op1_id = curnode->child[0]->type | T_IMMEDIATE;
			tnot->op1_defnum = 0;
			tnot->op1_data = &_neg_one;
			tnot->op1_pointer = false;

			//op2 is value
			tnot->op2_id = eval_id;
			tnot->op2_defnum = eval_defnum;
			if((signed long)tnot->op2_id < (signed long)T_TEMPREG)
			{
				tnot->op2_id |= T_IMMEDIATE;
				tnot->op2_data = (void *)curnode->child[0]->data;
			}
			tnot->op2_pointer = eval_pointer;

			tnot->target = 0;
			tnot->next = NULL;

			//destination is next avaliable "temporary" register
			eval_id = tnot->dest_id = T_TEMPREG;
			eval_defnum = tnot->dest_defnum = ++nexttemp;
			eval_pointer = tnot->dest_pointer = false;
			return appendMIRtoMIR(result, tnot);
		}

	case T_RETURN:
		{
			MIRInstr *tret = new MIRInstr;
			tret->instr = curnode->op;
			MIRInstr *result = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);

			//only need one child, op1
			tret->op1_id = eval_id;
			tret->op1_defnum = eval_defnum;
			if((signed long)tret->op1_id < (signed long)T_TEMPREG)
			{
				tret->op1_id |= T_IMMEDIATE;
				tret->op1_data = (void *)curnode->child[0]->data;
			}
			tret->op1_pointer = eval_pointer;

			tret->target = 0;
			tret->next = NULL;

			//double check to make sure
			if(currentfunction->returnType == (T_IMMEDIATE | T_VOID))
			{
				tret->op1_id = (T_IMMEDIATE | T_VOID);
				tret->op1_pointer = false;
			}

			return appendMIRtoMIR(result, tret);
		}


	////////binary ops
	case T_PLUS:	case T_MINUS:	case T_MULTIPLY:	case T_DIVIDE:	case T_MODULUS:
	case T_SHIFTLEFT:	case T_SHIFTRIGHT:
	case T_BXOR:	case T_BOR:	case T_BAND:
	case T_EQUALS:	case T_LESS:	case T_LEQUAL:	case T_GREATER:	case T_GEQUAL:	case T_NOTEQUAL:
		{
			//create temporary holder for instruction
			MIRInstr *tinstr = new MIRInstr;
			tinstr->instr = curnode->op;
			tinstr->target = 0;
			tinstr->next = NULL;
			
			//get left side
			MIRInstr *l_side = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			tinstr->op1_id = eval_id;
			tinstr->op1_defnum = eval_defnum;
			if((signed long)tinstr->op1_id < (signed long)T_TEMPREG)
			{
				tinstr->op1_id |= T_IMMEDIATE;
				tinstr->op1_data = (void *)curnode->child[0]->data;
			}
			tinstr->op1_pointer = eval_pointer;

			//get right side
			MIRInstr *r_side = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);
			tinstr->op2_id = eval_id;
			tinstr->op2_defnum = eval_defnum;
			if((signed long)tinstr->op2_id < (signed long)T_TEMPREG)
			{
				tinstr->op2_id |= T_IMMEDIATE;
				tinstr->op2_data = (void *)curnode->child[1]->data;
			}
			tinstr->op2_pointer = eval_pointer;

			//create new "temporary" register, and advertise its result
			eval_id = tinstr->dest_id = T_TEMPREG;
			eval_defnum = tinstr->dest_defnum = ++nexttemp;
			eval_pointer = tinstr->dest_pointer = false;

			//put in both sides of symbol, and then append current
			return appendMIRtoMIR(appendMIRtoMIR(l_side, r_side), tinstr);
		}

	case T_MEMBEROF:
		{
			//make sure not a function call (otherwise must be handled differently)
			if(curnode->child[1]->op != T_FUNCTIONCALL && curnode->child[1]->op != T_FUNCTIONCALL)
			{
				//create temporary holder for instruction
				MIRInstr *tinstr = new MIRInstr;
				tinstr->instr = T_PLUS;
				tinstr->target = 0;
				tinstr->next = NULL;
				
				//get left side
				MIRInstr *l_side = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, true);
				tinstr->op1_id = eval_id;
				tinstr->op1_defnum = eval_defnum;
				if((signed long)tinstr->op1_id < (signed long)T_TEMPREG)
				{
					tinstr->op1_id |= T_IMMEDIATE;
					tinstr->op1_data = (void *)curnode->child[0]->data;
				}
				tinstr->op1_pointer = false;

				//get right side -addition to current address, so we want a pointer
				MIRInstr *r_side = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, true);
				tinstr->op2_id = eval_id;
				tinstr->op2_defnum = eval_defnum;
				if((signed long)tinstr->op2_id < (signed long)T_TEMPREG)
				{
					tinstr->op2_id |= T_IMMEDIATE;
					tinstr->op2_data = (void *)curnode->child[1]->data;
				}
				tinstr->op2_pointer = false;	//we requested a pointer and don't want to dereference it

				//create new "temporary" register, and advertise its result
				eval_id = tinstr->dest_id = T_TEMPREG;
				eval_defnum = tinstr->dest_defnum = ++nexttemp;
				tinstr->dest_pointer = false;
				eval_pointer = true;

				//put in both sides of symbol, and then append current
				return appendMIRtoMIR(appendMIRtoMIR(l_side, r_side), tinstr);
			}
			else //member function call
			{

//***MUST ADD SOMETHING FOR PARALLEL CLASSES
				//make new push instr -to pass class pointer to function
				MIRInstr *pushinstr = new MIRInstr;
				pushinstr->instr = T_PUSHPARAMETER;
				pushinstr->target = 0;
				pushinstr->next = NULL;

				//get left side -class pointer
				MIRInstr *classpointer = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
				pushinstr->op1_id = eval_id;
				pushinstr->op1_defnum = eval_defnum;
				if((signed long)pushinstr->op1_id < (signed long)T_TEMPREG)
				{
					pushinstr->op1_id |= T_IMMEDIATE;
					pushinstr->op1_data = (void *)curnode->child[0]->data;
				}
				pushinstr->op1_pointer = false; //we want to pass a POINTER to the function

				//get function call (right side)
				MIRInstr *funccall = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);

				//instruction to put pointer in global space
				MIRInstr *global = new MIRInstr;
				global->instr = T_GLOBALSPACE;
				global->next = NULL;
				global->op1_id = pushinstr->op1_id;
				global->op1_defnum = pushinstr->op1_defnum;
				global->op1_pointer = pushinstr->op1_pointer;

				return appendMIRtoMIR(appendMIRtoMIR(classpointer, appendMIRtoMIR(global, pushinstr)), funccall);
			}
		}

	////////expression ops
	case T_ENOT:
		{
			//make new instr
			MIRInstr *tinstr = new MIRInstr;
			tinstr->instr = T_EQUALS;
			tinstr->target = 0;
			tinstr->next = NULL;

			//get expression to be not'd
			MIRInstr *operand = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			tinstr->op1_id = eval_id;
			tinstr->op1_defnum = eval_defnum;
			if((signed long)tinstr->op1_id < (signed long)T_TEMPREG)
			{
				tinstr->op1_id |= T_IMMEDIATE;
				tinstr->op1_data = (void *)curnode->child[0]->data;
			}
			tinstr->op1_pointer = eval_pointer;

			tinstr->op2_id = T_LONG | T_IMMEDIATE;
			tinstr->op2_defnum = 0;
			tinstr->op2_data = &_zero;
			tinstr->op2_pointer = false;

			//create temp register
			eval_id = tinstr->dest_id = T_TEMPREG;
			eval_defnum = tinstr->dest_defnum = ++nexttemp;
			eval_pointer = tinstr->dest_pointer = false;

			return appendMIRtoMIR(operand, tinstr);
		}

	case T_EAND:
		{
			//get first operand and copy over to a temp variable

			//make new assign instr
			MIRInstr *tempgets = new MIRInstr;
			tempgets->instr = T_PLUS;
			tempgets->target = 0;
			tempgets->next = NULL;

			//get first operand
			MIRInstr *ainstr = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			tempgets->op1_id = eval_id;
			tempgets->op1_defnum = eval_defnum;
			if((signed long)tempgets->op1_id < (signed long)T_TEMPREG)
			{
				tempgets->op1_id |= T_IMMEDIATE;
				tempgets->op1_data = (void *)curnode->child[0]->data;
			}
			tempgets->op1_pointer = eval_pointer;

			tempgets->op2_id = T_LONG | T_IMMEDIATE;
			tempgets->op2_defnum = 0;
			tempgets->op2_data = &_zero;
			tempgets->op2_pointer = false;

			TOKEN_ID result_id = tempgets->dest_id = T_TEMPREG;
			long result_defnum = tempgets->dest_defnum = ++nexttemp;
			bool result_pointer = tempgets->dest_pointer = false;

			//put two together
			ainstr = appendMIRtoMIR(ainstr, tempgets);

			//now have first opperand, do comparison
			//(check if equal to zero, optimizer will take care of rest)
			MIRInstr *tempif = new MIRInstr;
			tempif->instr = T_IF;
			tempif->dest_id = T_NOTEQUAL;

			tempif->op1_id = result_id;
			tempif->op1_defnum = result_defnum;
			tempif->op1_pointer = result_pointer;

			tempif->op2_id = T_LONG | T_IMMEDIATE;
			tempif->op2_defnum = 0;
			tempif->op2_data = &_zero;
			tempif->op2_pointer = false;

			long endlabel = tempif->target = ++nextlabel;
			tempif->next = NULL;

			ainstr = appendMIRtoMIR(ainstr, tempif);

			//check 2nd operand
			//if in this section of code (because of short circuiting), the result gets
			//this value
			//want to move result from second operand into the temp we decided on earlier

			tempgets = new MIRInstr;
			tempgets->instr = T_PLUS;
			tempgets->target = 0;
			tempgets->next = NULL;

			MIRInstr *binstr = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);

			tempgets->op1_id = eval_id;
			tempgets->op1_defnum = eval_defnum;
			if((signed long)tempgets->op1_id < (signed long)T_TEMPREG)
			{
				tempgets->op1_id |= T_IMMEDIATE;
				tempgets->op1_data = (void *)curnode->child[1]->data;
			}
			tempgets->op1_pointer = eval_pointer;

			tempgets->op2_id = T_LONG | T_IMMEDIATE;
			tempgets->op2_defnum = 0;
			tempgets->op2_data = &_zero;
			tempgets->op2_pointer = false;

			//report the result, and put as destination
			eval_id = tempgets->dest_id = result_id;
			eval_defnum = tempgets->dest_defnum = result_defnum;
			eval_pointer = tempgets->dest_pointer = result_pointer;

			binstr = appendMIRtoMIR(binstr, tempgets);

			//binstr contains the results if taken, append onto the first half
			ainstr = appendMIRtoMIR(ainstr, binstr);

			//create label "instruction".  re-use tempif pointer for heck of it
			tempif = new MIRInstr;
			tempif->instr = T_LABEL;
			tempif->target = endlabel;

			tempif->next = NULL;

			return appendMIRtoMIR(ainstr, tempif);
		}

	case T_EOR:
		{
			//get first operand and copy over to a temp variable

			//make new assign instr
			MIRInstr *tempgets = new MIRInstr;
			tempgets->instr = T_PLUS;
			tempgets->target = 0;
			tempgets->next = NULL;

			//get first operand
			MIRInstr *ainstr = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			tempgets->op1_id = eval_id;
			tempgets->op1_defnum = eval_defnum;
			if((signed long)tempgets->op1_id < (signed long)T_TEMPREG)
			{
				tempgets->op1_id |= T_IMMEDIATE;
				tempgets->op1_data = (void *)curnode->child[0]->data;
			}
			tempgets->op1_pointer = eval_pointer;

			tempgets->op2_id = T_LONG | T_IMMEDIATE;
			tempgets->op2_defnum = 0;
			tempgets->op2_data = &_zero;
			tempgets->op2_pointer = false;

			TOKEN_ID result_id = tempgets->dest_id = T_TEMPREG;
			long result_defnum = tempgets->dest_defnum = ++nexttemp;
			bool result_pointer = tempgets->dest_pointer = false;

			//put two together
			ainstr = appendMIRtoMIR(ainstr, tempgets);

			//now have first opperand, do comparison
			//(check if not equal to zero, optimizer will take care of rest)
			MIRInstr *tempif = new MIRInstr;
			tempif->instr = T_IF;
			tempif->dest_id = T_NOTEQUAL;

			tempif->op1_id = result_id;
			tempif->op1_defnum = result_defnum;
			tempif->op1_pointer = result_pointer;

			tempif->op2_id = T_LONG | T_IMMEDIATE;
			tempif->op2_defnum = 0;
			tempif->op2_data = &_zero;
			tempif->op2_pointer = false;

			long endlabel = tempif->target = ++nextlabel;
			tempif->next = NULL;

			ainstr = appendMIRtoMIR(ainstr, tempif);

			//check 2nd operand
			//if in this section of code (because of short circuiting), the result gets
			//this value
			//want to move result from second operand into the temp we decided on earlier

			tempgets = new MIRInstr;
			tempgets->instr = T_PLUS;
			tempgets->target = 0;
			tempgets->next = NULL;

			MIRInstr *binstr = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);

			tempgets->op1_id = eval_id;
			tempgets->op1_defnum = eval_defnum;
			if((signed long)tempgets->op1_id < (signed long)T_TEMPREG)
			{
				tempgets->op1_id |= T_IMMEDIATE;
				tempgets->op1_data = (void *)curnode->child[1]->data;
			}
			tempgets->op1_pointer = eval_pointer;

			tempgets->op2_id = T_LONG | T_IMMEDIATE;
			tempgets->op2_defnum = 0;
			tempgets->op2_data = &_zero;
			tempgets->op2_pointer = false;

			//report the result, and put as destination
			eval_id = tempgets->dest_id = result_id;
			eval_defnum = tempgets->dest_defnum = result_defnum;
			eval_pointer = tempgets->dest_pointer = result_pointer;

			binstr = appendMIRtoMIR(binstr, tempgets);

			//binstr contains the results if taken, append onto the first half
			ainstr = appendMIRtoMIR(ainstr, binstr);

			//create label "instruction".  re-use tempif pointer for heck of it
			tempif = new MIRInstr;
			tempif->instr = T_LABEL;
			tempif->target = endlabel;

			tempif->next = NULL;

			return appendMIRtoMIR(ainstr, tempif);
		}

	case T_IF:
		{
			//make new assign instr
			MIRInstr *tempgets = new MIRInstr;
			tempgets->instr = T_PLUS;

			//get first operand (condition)
			MIRInstr *condinstr = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);

			//now have first opperand, do comparison
			//(check if equal to zero, optimizer will take care of rest)
			MIRInstr *tempif = new MIRInstr;
			tempif->instr = T_IF;
			tempif->dest_id = T_EQUALS;

			tempif->op1_id = eval_id;	//results gotten from translation on condition
			tempif->op1_defnum = eval_defnum;
			if((signed long)tempif->op1_id < (signed long)T_TEMPREG)
			{
				tempif->op1_id |= T_IMMEDIATE;
				tempif->op1_data = (void *)curnode->child[0]->data;
			}
			tempif->op1_pointer = eval_pointer;

			tempif->op2_id = T_LONG | T_IMMEDIATE;
			tempif->op2_defnum = 0;
			tempif->op2_data = &_zero;
			tempif->op2_pointer = false;

			long elselabel = tempif->target = ++nextlabel;
			tempif->next = NULL;

			condinstr = appendMIRtoMIR(condinstr, tempif);

			//get 2nd operand (statements if condition)
			MIRInstr *ifinstr = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);

			//goto end instr
			MIRInstr *tempgoto = new MIRInstr;
			tempgoto->instr = T_GOTO;
			long endlabel = tempgoto->target = ++nextlabel;
			tempgoto->next = NULL;

			ifinstr = appendMIRtoMIR(ifinstr, tempgoto);

			//ifinstr contains the results if taken, append onto the first half
			ifinstr = appendMIRtoMIR(condinstr, ifinstr);

			//create label "instruction" for else.  re-use tempif pointer for heck of it
			tempif = new MIRInstr;
			tempif->instr = T_LABEL;
			tempif->target = elselabel;
			tempif->next = NULL;

			//put label for else section
			ifinstr = appendMIRtoMIR(ifinstr, tempif);

			//get else block
			tempif = translateASTtoMIR(curnode->child[2], eval_id, eval_defnum, eval_pointer, false);

			ifinstr = appendMIRtoMIR(ifinstr, tempif);

			//now put end label on
			tempif = new MIRInstr;
			tempif->instr = T_LABEL;
			tempif->target = endlabel;
			tempif->next = NULL;

			return appendMIRtoMIR(ifinstr, tempif);
		}

	case T_WHILE:
		{
			//create beginning label
			MIRInstr *winstr = new MIRInstr;
			winstr->instr = T_LABEL;
			long whilestart = winstr->target = ++nextlabel;
			winstr->next = NULL;

			//get conditional and put after label
			MIRInstr *condinstr = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);
			winstr = appendMIRtoMIR(winstr, condinstr);

			//build condition test (if)
			condinstr = new MIRInstr;
			condinstr->instr = T_IF;
			condinstr->dest_id = T_NOTEQUAL;

			condinstr->op1_id = eval_id; //results from condition
			condinstr->op1_defnum = eval_defnum;
			if((signed long)condinstr->op1_id < (signed long)T_TEMPREG)
			{
				condinstr->op1_id |= T_IMMEDIATE;
				condinstr->op1_data = (void *)curnode->child[0]->data;
			}
			condinstr->op1_pointer = eval_pointer;

			condinstr->op2_id = T_LONG | T_IMMEDIATE;
			condinstr->op2_defnum = 0;
			condinstr->op2_data = &_zero;
			condinstr->op2_pointer = false;

			long whileend = condinstr->target = ++nextlabel;
			condinstr->next = NULL;

			//put test on front
			winstr = appendMIRtoMIR(winstr, condinstr);

			//grab block
			condinstr = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);

			//append block to condition
			winstr = appendMIRtoMIR(winstr, condinstr);

			//instruction to go back to beginning of loop
			condinstr = new MIRInstr;
			condinstr->instr = T_GOTO;
			condinstr->target = whilestart;
			condinstr->next = NULL;

			winstr = appendMIRtoMIR(winstr, condinstr);

			//append end label
			condinstr = new MIRInstr;
			condinstr->instr = T_LABEL;
			condinstr->target = whileend;
			condinstr->next = NULL;

			winstr = appendMIRtoMIR(winstr, condinstr);

			//update any continue/break statements
			replaceLoopControls(winstr);
			return winstr;
		}

	case T_BREAK:
	case T_CONTINUE:
		{
			//just add instruction -loop will replace it with correct goto
			MIRInstr *br = new MIRInstr;
			br->instr = curnode->op;
			br->target = 0;
			br->next = NULL;
			return br;
		}

	case T_LBRACKET:
		{
			//get index, multiply by size, get base address, add together, dereference
			MIRInstr *index = translateASTtoMIR(curnode->child[1], eval_id, eval_defnum, eval_pointer, false);

			//scale
			MIRInstr *scaleindex = new MIRInstr;
			scaleindex->instr = T_MULTIPLY;
			scaleindex->target = 0;
			scaleindex->next = NULL;

			scaleindex->op1_id = eval_id;
			scaleindex->op1_defnum = eval_defnum;
			if((signed long)scaleindex->op1_id < (signed long)T_TEMPREG)
			{
				scaleindex->op1_id |= T_IMMEDIATE;
				scaleindex->op1_data = (void *)curnode->child[1]->data;
			}
			scaleindex->op1_pointer = eval_pointer;

			scaleindex->op2_defnum = 0;
			scaleindex->op2_pointer = false;
			//get size of array unit
			ASTNode *scopenode = curnode->child[0];
			while(scopenode != NULL)
			{
				if(scopenode->op == T_MEMBEROF)
					scopenode = scopenode->child[1];
				else if(scopenode->op == T_LBRACKET)
					scopenode = scopenode->child[0];
				else
					break;
				//can assume this loop is good, because checked by parser
			}
			
			//add get corresponding size
			if(scopenode->type & A_RESIZEABLE)
			{
				scaleindex->op2_id = T_ADDRESS | T_IMMEDIATE;
				scaleindex->op2_defnum = WORDSIZE;
			}
			if(scopenode->type & A_BASETYPES)
			{
				scaleindex->op2_id = T_ADDRESS | T_IMMEDIATE;
				long elementsize;
				switch(symboltable->getType(scopenode->op, scopenode->defnum) & A_BASETYPES)
				{
				case T_CHAR:
					elementsize = SIZE_CHAR;
					break;
				case T_SHORT:
					elementsize = SIZE_SHORT;
					break;
				case T_LONG:
					elementsize = SIZE_LONG;
					break;
				case T_FLOAT:
					elementsize = SIZE_FLOAT;
					break;
				case T_DOUBLE:
					elementsize = SIZE_DOUBLE;
					break;
				case T_ADDRESS:
					elementsize = SIZE_LONG;
					break;
				default:
					elementsize = 0;
					break;
				}

				scaleindex->op2_defnum = elementsize;
			}
			else if(scopenode->type & A_USERTYPE)
			{
				TOKEN_ID scopeparent = symboltable->getUserType(scopenode->op, scopenode->defnum); //type contains TOKEN_ID of [p]class
				if(symboltable->getType(scopeparent, symboltable->getClassDefnum(scopeparent)) == T_PCLASS)
				{
					//set to size of pclass reference
					scaleindex->op2_id = T_ADDRESS | T_IMMEDIATE;
					scaleindex->op2_defnum = SIZE_RPC;
				}
				else
				{
					//get size of class
					scaleindex->op2_id = T_ADDRESS | T_IMMEDIATE;
					scaleindex->op2_defnum = symboltable->getSize(scopeparent, symboltable->getClassDefnum(scopeparent));
				}
			}

			TOKEN_ID scaledindex_id = scaleindex->dest_id = T_TEMPREG;
			long scaledindex_defnum = scaleindex->dest_defnum = ++nexttemp;
			bool scaledindex_pointer = scaleindex->dest_pointer = false;

			index = appendMIRtoMIR(index, scaleindex);


			//get base address
			MIRInstr *pointer = translateASTtoMIR(curnode->child[0], eval_id, eval_defnum, eval_pointer, false);

			index = appendMIRtoMIR(index, pointer);

			//make instruction to add index and base
			MIRInstr *ibadd = new MIRInstr;
			ibadd->instr = T_PLUS;
			ibadd->target = 0;
			ibadd->next = NULL;

			ibadd->op1_id = scaledindex_id;
			ibadd->op1_defnum = scaledindex_defnum;
			if((signed long)ibadd->op1_id < (signed long)T_TEMPREG)
			{
				ibadd->op1_id |= T_IMMEDIATE;
				ibadd->op1_data = (void *)curnode->child[0]->data;
			}
			ibadd->op1_pointer = scaledindex_pointer;

			ibadd->op2_id = eval_id;
			ibadd->op2_defnum = eval_defnum;
			if((signed long)ibadd->op2_id < (signed long)T_TEMPREG)
			{
				ibadd->op2_id |= T_IMMEDIATE;
				ibadd->op2_data = (void *)curnode->child[1]->data;
			}
			ibadd->op2_pointer = eval_pointer;

			eval_id = ibadd->dest_id = T_TEMPREG;
			eval_defnum = ibadd->dest_defnum = ++nexttemp;
			ibadd->dest_pointer = false;

			eval_pointer = true;

			return appendMIRtoMIR(index, ibadd);
		}

	case T_PFUNCTIONCALL:
	case T_FUNCTIONCALL:
		{
			//have list of parameters and pushes (for putting on stack)
			//append together

			MIRInstr *call;	//must have calling instruction (or waiting instr)

			//if in parallel, must wait instead
			//but don't want to mess up parallel class calls, so skip over them
			if(!inparallel || curnode->op == T_PFUNCTIONCALL)
			{
				MIRInstr *parameters = NULL;
				MIRInstr *pushes = NULL;

				//if a member of a class calling a member of its class, then need to push itself
				if(symboltable->getParent(curnode->funccall->op, curnode->funccall->defnum) != GLOBALSCOPE
					&& curnode->parent->op != T_MEMBEROF)
				{
					pushes = new MIRInstr;
					pushes->next = NULL;
					pushes->instr = T_PUSH_THIS;
				}

				//go through each of the parameters, going in correct order
				for(int i = 0; i < curnode->funccall->numparameters; i++)
				{
					//get first operand (condition)
					MIRInstr *param = translateASTtoMIR(curnode->funccall->parameters[i], eval_id, eval_defnum, eval_pointer, false);
					parameters = appendMIRtoMIR(param, parameters);

					//put operand push
					param = new MIRInstr;
					param->instr = T_PUSHPARAMETER;
					param->op1_id = eval_id;	//results gotten from translation
					param->op1_defnum = eval_defnum;
					if((signed long)param->op1_id < (signed long)T_TEMPREG)
					{
						param->op1_id |= T_IMMEDIATE;
						param->op1_data = (void *)curnode->funccall->parameters[i]->data;
					}
					param->op1_pointer = eval_pointer;

					param->target = 0;
					param->next = NULL;

					pushes = appendMIRtoMIR(param,	pushes);
				}

				//create call instruction
				call = new MIRInstr;
				call->instr = curnode->op;
				call->dest_id = curnode->funccall->op;
				call->dest_defnum = curnode->funccall->defnum;
				call->target = 0;
				call->next = NULL;

				parameters = appendMIRtoMIR(parameters, pushes);
				call = appendMIRtoMIR(parameters, call);

				//check to see if return value
				FunctionType *f = symboltable->getFunctionType(curnode->funccall->op, curnode->funccall->defnum);
				if(f->returnType == T_VOID)
				{
					eval_id = 0;
					eval_defnum = 0;
					eval_pointer = 0;
					return call;
				}

				//if parallel class call, wait for return value
				if(curnode->op == T_PFUNCTIONCALL)
				{
					parameters = new MIRInstr;
					parameters->instr = T_WAITRETURN;
					parameters->dest_id = curnode->funccall->op;	//remember name of func waiting for debug info
					parameters->dest_defnum = curnode->funccall->defnum;
					parameters->target = 0; //0 is wait_id for parallel function calls
					parameters->next = NULL;
					call = appendMIRtoMIR(call, parameters);
				}
			}
			else //inparallel
			{
				//check to see if return value (if not, don't bother waiting)
				FunctionType *f = symboltable->getFunctionType(curnode->funccall->op, curnode->funccall->defnum);
				if(f->returnType == T_VOID)
				{
					eval_id = 0;
					eval_defnum = 0;
					eval_pointer = 0;
					return NULL;
				}

				call = new MIRInstr;
				call->instr = T_WAITRETURN;
				call->dest_id = curnode->funccall->op;	//remember name of func waiting for debug info
				call->dest_defnum = curnode->funccall->defnum;
				call->target = curnode->defnum;
				call->next = NULL;
			}

			//need to pull value from stack
			//make instruction to do so
			MIRInstr *ret = new MIRInstr;
			ret->instr = T_PULLPARAMETER;
			ret->target = 0;
			ret->next = NULL;

			eval_id = ret->op1_id = T_TEMPREG;
			eval_defnum = ret->op1_defnum = ++nexttemp;
			eval_pointer = ret->op1_pointer = 0;

			return appendMIRtoMIR(call, ret);
		}


	default:
		return NULL;
	}

	return NULL;
}

//concatenates b to the end of a, and returns the result
//(return is more useful than just 'a' because of NULL's)
MIRInstr *MIRTransform::appendMIRtoMIR(MIRInstr *a, MIRInstr *b)
{
	//just incase either is NULL, our job is easy
	if(a == NULL)
		return b;
	if(b == NULL)
		return a;

	MIRInstr *cur;
	for(cur = a; cur->next != NULL; cur = cur->next);
	cur->next = b;
	return a;
}

//for break and continue, leave in MIR, and call this when come out of
//block of code -will replace all occurences with correct jump
void MIRTransform::replaceLoopControls(MIRInstr *a)
{
	long startlabel = 0, endlabel = 0;

	//get first label
	MIRInstr *cur;
	for(cur = a; cur->next != NULL; cur = cur->next)
	{
		if(cur->instr == T_LABEL)
		{
			startlabel = cur->target;
			break;
		}
	}

	//get last label
	for(cur = a; cur != NULL; cur = cur->next)
	{
		if(cur->instr == T_LABEL)
			endlabel = cur->target;
	}

	//go through changing breaks and continues to goto's
	for(cur = a; cur != NULL; cur = cur->next)
	{
		if(cur->instr == T_BREAK)
		{
			cur->instr = T_GOTO;
			cur->target = endlabel;
		}
		else if(cur->instr == T_CONTINUE)
		{
			cur->instr = T_GOTO;
			cur->target = startlabel;
		}
	}
}

//finds 1-deep function calls in a block and extracts them
//this is used for parallel blocks to start the functions running before
//executing the block

//***NOTE: WILL WANT TO CHECK FOR POINTERS (i.e. member functions) BEFORE CALLING IN PARALLEL!!!
MIRInstr *MIRTransform::translateASTFuncPrecallsToMIR(ASTNode *curnode)
{
	if(curnode == NULL)
		return NULL;

	//traverse tree looking for 1-deep function calls

	switch(curnode->op)
	{

	//unary ops
	case T_BNOT:	case T_ENOT:
	case T_UNARYPLUS:	case T_UNARYMINUS:
		return translateASTFuncPrecallsToMIR(curnode->child[0]);

	//binary ops
	case T_PLUS:	case T_MINUS:	case T_MULTIPLY:	case T_DIVIDE:
	case T_MODULUS:	case T_SHIFTLEFT:	case T_SHIFTRIGHT:
	case T_BXOR:	case T_BOR:	case T_BAND:
	case T_EQUALS:	case T_LESS:	case T_LEQUAL:	case T_GREATER:
	case T_GEQUAL:	case T_NOTEQUAL:
	case T_EOR:	case T_EAND:
	case T_GETS:	case T_ARRAYRESIZE:
	case T_MEMBEROF:	case T_LBRACKET:
	case T_SEMICOLON:
		return appendMIRtoMIR(translateASTFuncPrecallsToMIR(curnode->child[0]), translateASTFuncPrecallsToMIR(curnode->child[1]));

	case T_FUNCTIONCALL:
		{
			curnode->defnum = ++current_wait_id; //need to remember to wait for this function call

			MIRInstr *parameters = NULL;
			MIRInstr *pushes = NULL;
			MIRInstr *call;

			TOKEN_ID eval_id = 0;
			long eval_defnum = 0;
			bool eval_pointer = 0;

			//go through each of the parameters, going in correct order
			for(int i = 0; i < curnode->funccall->numparameters; i++)
			{
				//get first operand (condition)
				MIRInstr *param = translateASTtoMIR(curnode->funccall->parameters[i], eval_id, eval_defnum, eval_pointer, false);
				parameters = appendMIRtoMIR(param, parameters);

				//put operand push
				param = new MIRInstr;
				param->instr = T_PUSHPARAMETER;
				param->op1_id = eval_id;	//results gotten from translation
				param->op1_defnum = eval_defnum;
				if((signed long)param->op1_id < (signed long)T_TEMPREG)
				{
					param->op1_id |= T_IMMEDIATE;
					param->op1_data = (void *)curnode->funccall->parameters[i]->data;
				}
				param->op1_pointer = eval_pointer;

				param->target = 0;
				param->next = NULL;

				pushes = appendMIRtoMIR(param, pushes);
			}

			//create call instruction
			call = new MIRInstr;
			call->instr = T_PFUNCTIONCALL;
			call->dest_id = curnode->funccall->op;
			call->dest_defnum = curnode->funccall->defnum;
			call->target = curnode->defnum;
			call->next = NULL;

			parameters = appendMIRtoMIR(parameters, pushes);
			return appendMIRtoMIR(parameters, call);
		}
		break;

	default:
		break;
	}

	return NULL;
}


//does final polishing after MIR translation (such as backwards linking the list, data manip, etc.)
void MIRTransform::PolishMIR(MIRInstr *m)
{
	if(m == NULL)
		return;

	for(MIRInstr *mir = m; mir->next != NULL; mir = mir->next)
	{
		//chains mir instructions so 2-way linked
		mir->next->prev = mir;

		//copies address to data section
		if(mir->op1_id == (T_IMMEDIATE | T_ADDRESS) )
		{
			mir->op1_data = (void *)new unsigned long;
			*(unsigned long *)mir->op1_data = mir->op1_defnum;
		}

		if(mir->op2_id == (T_IMMEDIATE | T_ADDRESS) )
		{
			mir->op2_data = (void *)new unsigned long;
			*(unsigned long *)mir->op2_data = mir->op2_defnum;
		}
	}
}

