
#include <stdio.h>

//mid level intermediate representation
//translates parse tree into this form

class MIRTransform
{
public:
	MIRTransform(StringTable *st, SymbolTable *yt);
	~MIRTransform();

	void Transform(FunctionType *func);

	//printing MIR code
	void PrintListing(FILE *outfile, MIRInstr *func);
	void MirPrint3PartInstr(MIRInstr *func, char *buffer, char *instr);

private:
	StringTable *stringtable;
	SymbolTable *symboltable;

	//does bulk of AST to MIR translation
	MIRInstr *translateASTtoMIR(ASTNode *curnode, TOKEN_ID &eval_id, long &eval_defnum, bool &eval_pointer, bool request_pointer);

	//finds 1-deep function calls in a block and extracts them
	//this is used for parallel blocks to start the functions running before
	//executing the block
	MIRInstr *translateASTFuncPrecallsToMIR(ASTNode *curnode);

	//does final polishing after MIR translation (such as backwards linking the list, data manip, etc.)
	void PolishMIR(MIRInstr *mir);


	//"globals" used by translateASTtoMIR() -must be set before begin calling
	ASTNode *curfunction;
	long nexttemp;	//defnum of next temporary register (temporary registers use id T_TEMPREG)
	long nextlabel;	//defnum of next avaliable label for control

	long current_wait_id;	//used for marking parallel function calls, so they can find their return values

	bool inparallel;	//specifies translator whether or not in parallelizing block

	double _zero;	//set to zero for using a reference to zero
	unsigned long _neg_one;	//set to all ones

	//concatonates a to b, and returns the result (return is more useful than just a because of NULL's)
	MIRInstr *appendMIRtoMIR(MIRInstr *a, MIRInstr *b);

	//replaces break and continue instructions within loop to goto's (last and first label respectively)
	void replaceLoopControls(MIRInstr *a);

	FunctionType *currentfunction;	//function currently transforming

};

