
//linked list node of ASTNode's
struct OpDef
{
	ASTNode *astnode;
	long parendepth;
	long bracketdepth;
	OpDef *next;
};

//stack of operators for parser
class OpStack
{
public:
	OpStack();
	~OpStack();

	void pushOp(ASTNode *n, long parendepth, long bracketdepth);
	void pushOp(ASTNode *n); //used for non-operators
	ASTNode *popOp();
	ASTNode *peekOp();
	long peekParenDepth();
	long peekBracketDepth();

private:
	OpDef *root;
};



//signals where end of current scope is in stack
#define SCOPECHANGE 0
//invalid defnum
#define INVALIDDEFNUM -1
//linked list node for variable definition scope stack
struct VarDef
{
	TOKEN_ID id;
	long curdefnum;
	VarDef *next;
};


//stack for finding correct variable to use (from highest scope)
class VarDefStack
{
public:
	VarDefStack();
	~VarDefStack();

	void pushVarDef(TOKEN_ID id, long curdefnum);
	void increaseScope();
	void decreaseScope();	//flushes variables when exiting scope

	long findFirstDef(TOKEN_ID id);	//returns first defnum
	bool ExistInCurScope(TOKEN_ID id);	//returns if exists in current scope

private:
	VarDef *root;
};


class Parser
{
public:
	Parser(char *file, StringTable *st, SymbolTable *yt);
	bool Parse();

	~Parser();

private:
	StringTable *stringtable;
	SymbolTable *symboltable;
	Scanner *scanner;
	Constrainer *constrainer;

	//used internally to clean up
	bool EndParse();

	int curScopeDepth;	//0 = global

	OpStack *opstack, *varstack;
	VarDefStack *vardefstack;

	long globalvarsize;	//total space of global variables


	/////////////////////////////
	//parser broken up for maintanability
	//procedures to break up parser into smaller blocks
	bool ParseGlobalTypeDefinition();
	bool ParseGlobalVarDefinition();
	bool ParseGlobalVarDefinitionInit();
	bool ParseGlobalArrayDefinition();

	bool ParseGlobalClassDefinition();
	bool ParseGlobalUserTypeDefinition();

	bool ParseGlobalFunction();
	bool ParseGlobalClassScopeFunction();

	bool ParseFunction();

	//PF = parse function
	bool PFStatement(ASTNode *&tree, bool neednexttoken, bool mult_statements);
	bool PFTypeDefinition(ASTNode *&tree);
	bool PFVarDefinition();
	bool PFArrayDefinition();
	bool PFExpression(ASTNode *&tree);
	void PFCondenseStack(TOKEN_ID compairtoken);
	bool PFUserTypeDefinition(long curclassdefnum);

	void PFConvertMemberFunctionCalls(ASTNode *tree);


	//globals for parser simplification procedures
	TOKEN_ID curtoken;
	TOKEN_ID nexttoken;
	TOKEN_ID prevtoken;
	ATTRIB_TYPE curattrib;
	TOKEN_ID curparent;	//used for within classes in declarations
	long curparentdefnum;
	bool inclass;
	bool symbol_already_exists;
	bool require_func_definition;
	FunctionType *curfunc;
	TOKEN_ID functid;
	long funcdefnum;	//possible overloading

	long curparendepth;
	long returnparendepth;	//the depth of which to return (normally 0)
	long curbracketdepth;
	bool inparallel;	//set if in parallel block
	bool inloop;	//set if in loop
};

