#include "common.h"

//must check for overloading (allowed)
//if inclass, then must add hidden arguement of class instantiation
bool Parser::ParseGlobalFunction()
{
	//function declaration
	
	curfunc = new FunctionType;
	curfunc->returnType = curattrib;
	curfunc->numarguements = 0; //initialize to 0
	curfunc->function = NULL;
	curfunc->curvarpos = -SIZE_LONG;		//start at beginning of stack -start below return address
	curfunc->mircode = NULL;

	StringTrieNode *funcname = scanner->getEndNameofLastGet();
	functid = curtoken;

	long functionstack = 2*SIZE_LONG;	//mem address of each variable -skip return address and saved stack pointer

	//set to default
	curattrib = A_SIGNED;
	vardefstack->increaseScope();	//for within function
	//get arguements
	do
	{
		curtoken = scanner->getNextToken();
		switch(curtoken)
		{
		case T_CHAR:
		case T_SHORT:
		case T_LONG:
		case T_FLOAT:
		case T_DOUBLE:
			{
				//save type
				curattrib |= curtoken;

				//get name
				curtoken = scanner->getNextToken();

				//make sure not error
				if(curtoken == T_ENDOFFILE)
				{
					constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
					return false;
				}
				else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
					constrainer->reportError(ERROR_INVALIDIDENTIFIER);

				//check for duplicates
				if(vardefstack->ExistInCurScope(curtoken))
					constrainer->reportError(ERROR_SYMBOLDEFINED);

				long defnum = symboltable->getNextDefNum(curtoken);
				vardefstack->pushVarDef(curtoken, defnum);
				
				//////////////////
				//check for array declaration
				nexttoken = scanner->getNextToken();
				//check for array modifier
				if(nexttoken == T_LBRACKET)
				{
					nexttoken = scanner->getNextToken();
					if(nexttoken == T_ENDOFFILE)
					{
						constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
						return false;
					}
					else if(nexttoken != T_RBRACKET)
						constrainer->reportError(ERROR_INVALIDSYMBOL);

					curattrib |= A_ARRAY;
					nexttoken = scanner->getNextToken();
				}

				switch(nexttoken)
				{
				case T_COMMA:
				case T_RPAREN:	//will exit from loop
					//add variable to list
					curfunc->arguements[curfunc->numarguements++] = curattrib;
					symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(), curattrib, 0,
						functionstack, (unsigned char *)"\0\0\0\0\0\0\0\0", functid);
					//update function calling stack
					if(curattrib & A_ARRAY)
					{
				//***user types and arrays are not allowed yet -add this
				constrainer->reportError(ERROR_USERTYPESNOTALLOWED);
						functionstack += SIZE_LONG;
					}
					else
					{
						switch(curattrib & A_BASETYPES)
						{
						case T_CHAR:
							functionstack += SIZE_CHAR;
							break;
						case T_SHORT:
							functionstack += SIZE_SHORT;
							break;
						case T_LONG:
							functionstack += SIZE_LONG;
							break;
						case T_FLOAT:
							functionstack += SIZE_FLOAT;
							break;
						case T_DOUBLE:
							functionstack += SIZE_DOUBLE;
							break;
						default:
							functionstack += SIZE_MAXDATA;
							break;
						}
					}
					curtoken = nexttoken;	//to check at end of loop if done
					break;

				case T_ENDOFFILE:
					constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
					return false;
				default:
					constrainer->reportError(ERROR_INVALIDSYMBOL);
					break;
				} //switch ender (',' ')' etc.)

				//return to default
				curattrib = A_SIGNED;

			}	//arguement
			break;

		//qualifiers
		case T_CONST:
			if(curattrib == A_SIGNED)
				curattrib = A_CONST;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;

		case T_UNSIGNED:
			if((curattrib & A_UNSIGNED) != A_UNSIGNED)
				curattrib |= A_UNSIGNED;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;

		case T_RPAREN:
			break;

		case T_ENDOFFILE:
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;

		default:
			{
				//check to see if instatiation of user type
				long curclassdefnum = vardefstack->findFirstDef(curtoken);
				TOKEN_ID curclass = curtoken;
				if(curclassdefnum
					&& (symboltable->getType(curclass, curclassdefnum) == T_CLASS || symboltable->getType(curclass, curclassdefnum) == T_PCLASS) )
				{

					//get name
					curtoken = scanner->getNextToken();

					//make sure not error
					if(curtoken == T_ENDOFFILE)
					{
						constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
						return false;
					}
					else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
						constrainer->reportError(ERROR_INVALIDIDENTIFIER);

					//check for duplicates
					if(vardefstack->ExistInCurScope(curtoken))
						constrainer->reportError(ERROR_SYMBOLDEFINED);

					long defnum = symboltable->getNextDefNum(curtoken);
					vardefstack->pushVarDef(curtoken, defnum);
					
					//////////////////
					//check for array declaration
					nexttoken = scanner->getNextToken();
					//check for array modifier
					if(nexttoken == T_LBRACKET)
					{
						nexttoken = scanner->getNextToken();
						if(nexttoken == T_ENDOFFILE)
						{
							constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
							return false;
						}
						else if(nexttoken != T_RBRACKET)
							constrainer->reportError(ERROR_INVALIDSYMBOL);

						curattrib |= A_ARRAY;
						nexttoken = scanner->getNextToken();
					}

					switch(nexttoken)
					{
					case T_COMMA:
					case T_RPAREN:	//will exit from loop
				//***user types and arrays are not allowed yet -add this
				constrainer->reportError(ERROR_USERTYPESNOTALLOWED);
						//add variable to list
						curfunc->arguements[curfunc->numarguements++] = A_USERTYPE;
						curfunc->arguements[curfunc->numarguements++] = curclass;
						symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(), A_USERTYPE, curclass,
							functionstack, (unsigned char *)"\0\0\0\0\0\0\0\0", functid);
						//update function calling stack
						if(curattrib & A_ARRAY)
							functionstack += SIZE_LONG;
						else
						{
							functionstack += symboltable->getSize(curclass, curclassdefnum);
						}
						curtoken = nexttoken;	//to check at end of loop if done
						break;

					case T_ENDOFFILE:
						constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
						return false;
					default:
						constrainer->reportError(ERROR_INVALIDSYMBOL);
						break;
					} //switch ender (',' ')' etc.)
				} //class arguement
				else
					constrainer->reportError(ERROR_INVALIDSYMBOL);

				//return to default
				curattrib = A_SIGNED;
			}
			break;
		}
	} while(curtoken != T_RPAREN);

	curfunc->parameterlength = functionstack;

	//check to see if prototype, etc.
	curtoken = scanner->getNextToken();
	switch(curtoken)
	{
	case T_SEMICOLON: //prototype -add function declaration
		{
			funcdefnum = symboltable->getMatchingFuncDefnum(functid, curfunc, curparent);
			if(!require_func_definition && funcdefnum == INVALIDDEFNUM) //doesn't exist, so create it
			{
				funcdefnum = symboltable->getNextDefNum(functid);
				symboltable->addSymbol(functid, funcdefnum, funcname, A_FUNCTION, 0,
					0 /*use 0 as init addr for funcs*/, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);
				symboltable->setFunctionType(functid, funcdefnum, curfunc);
			}
			else //already exists
			{
				constrainer->reportError(ERROR_DUPLICATEPROTOTYPE);
			}
			vardefstack->decreaseScope();	//get rid of all function's parameters
		}
		break;
	case T_LBRACE:
		if(curfunc->returnType & A_EXTERNAL)
		{
			constrainer->reportError(ERROR_INITIALIZEDEXTERNAL);
			return false;
		}

		funcdefnum = symboltable->getMatchingFuncDefnum(functid, curfunc, curparent);
		if(funcdefnum == INVALIDDEFNUM) //doesn't exist, so create it
		{
			//must exist in class
			if(require_func_definition)
				constrainer->reportError(ERROR_NOTCLASSMEMBER);

			funcdefnum = symboltable->getNextDefNum(functid);
			symboltable->addSymbol(functid, funcdefnum, funcname, A_FUNCTION, 0,
				0 /*use 0 as init addr for funcs*/, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);
			symboltable->setFunctionType(functid, funcdefnum, curfunc);
		}
		else //it already exists
		{
			delete curfunc; //created for no reason
			curfunc = symboltable->getFunctionType(functid, funcdefnum);

			if(curfunc->function != NULL)
			{
				constrainer->reportError(ERROR_FUNCTALREADYDEFINED);
			}
		}

		//the big part... parse the function and put it into curfunc->function
		if(!ParseFunction())
			return false;
		break;
	case T_ENDOFFILE:
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	default:
		constrainer->reportError(ERROR_INVALIDSYMBOL);
		break;
	}

	return true;
}

//have parent class identifier
bool Parser::ParseGlobalClassScopeFunction()
{
	//make sure not within another class
	if(curparent != GLOBALSCOPE)
		constrainer->reportError(ERROR_SCOPEDFUNCDEFNOTGLOBAL);

	curparent = curtoken;	//set to class from scope reference
	//make sure legitimate class
	if(curparent != GLOBALSCOPE && symboltable->getClassDefnum(curparent) == INVALIDDEFNUM)
		constrainer->reportError(ERROR_INVALIDIDENTIFIER);

	curtoken = scanner->getNextToken();
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}

	//make sure function name not symbol
	if(curtoken < T_BUILTINFUNCSTART)
		constrainer->reportError(ERROR_INVALIDIDENTIFIER);

	//get left parentheses
	nexttoken = scanner->getNextToken();
	if(nexttoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	else if(nexttoken != T_LPAREN)
	{
		constrainer->reportError(ERROR_INVALIDSYMBOL);
	}

	require_func_definition = true;
	bool retval = ParseGlobalFunction();
	require_func_definition = false;

	curparent = GLOBALSCOPE;
	return retval;
}

