#include "common.h"


//parse the function and put it into curfunc->function
//curtoken is assumed to be T_LBRACE
bool Parser::ParseFunction()
{
	inparallel = false;	//start not in parallel code
	inloop = false;		//start not in loop
	//start with blank function
	curfunc->function = NULL;
	bool retval = PFStatement(curfunc->function, false, false);
	vardefstack->decreaseScope();	//increased again when going into function -need to even out
	if(retval)
		PFConvertMemberFunctionCalls(curfunc->function);
	return retval;
}


bool Parser::PFStatement(ASTNode *&tree, bool neednexttoken, bool mult_statements)
{
	ASTNode *curnode = NULL;

	bool getnexttoken = neednexttoken;	//keeps local copy

	curattrib = A_SIGNED;

	do //while in function
	{
		if(getnexttoken)
			curtoken = scanner->getNextToken();
		getnexttoken = true;	//set to true for next case

		switch(curtoken)
		{
		//declaration
		case T_CHAR:
		case T_SHORT:
		case T_LONG:
		case T_FLOAT:
		case T_DOUBLE:
			if(inparallel)
				constrainer->reportError(ERROR_PARALLELVARIABLEDECLARED);
			if(!PFTypeDefinition(tree))
				return false;
			break;

		//*******
		//qualifiers
		case T_CONST:
			if(curattrib == A_SIGNED)
				curattrib = A_CONST;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;

		case T_UNSIGNED:
			if((curattrib & A_UNSIGNED) != A_UNSIGNED)
				curattrib |= A_UNSIGNED;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;

		case T_IF:
			{
				//make sure no qualifiers
				if(curattrib != A_SIGNED)
				{
					constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
					curattrib = A_SIGNED;	//so don't pick up error more than once
				}
				//can't have control statements in parallel blocks
				if(inparallel)
					constrainer->reportError(ERROR_PARALLELCONTROLSTATEMENTS);

				//get left parentheses
				curtoken = scanner->getNextToken();
				if(curtoken == T_ENDOFFILE)
				{
					constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
					return false;
				}
				if(curtoken != T_LPAREN)
					constrainer->reportError(ERROR_MISSINGLEFTPARENTHESES);

				//get expression within if(...)
				curparendepth = 0;
				returnparendepth = 0;
				curbracketdepth = 0;
				curtoken = scanner->getNextToken();
				if(!PFExpression(curnode))
					return false;
				if(curtoken != T_RPAREN)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				else if(curbracketdepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

				//begin building if condition
				ASTNode *ifcond = new ASTNode;
				ifcond->op = T_IF;
				ifcond->type = T_VOID;
				ifcond->funccall = NULL;
				ifcond->child[0] = curnode;
				if(ifcond->child[0] != NULL)
					ifcond->child[0]->parent = ifcond;
				else
					constrainer->reportError(ERROR_EMPTYLANGUAGECONSTRUCT);

				//if's true branch
				curnode = NULL;
				PFStatement(curnode, true, false);

				//put expression on child[1] of if node
				ifcond->child[1] = curnode;
				if(ifcond->child[1] != NULL)
					ifcond->child[1]->parent = ifcond;

				//check to see if next token is T_ELSE
				curtoken = scanner->getNextToken();
				if(curtoken != T_ELSE)
				{
					ifcond->child[2] = NULL;
				}
				else //else statement
				{
					curnode = NULL;
					PFStatement(curnode, true, false);

					//put expression on child[2] of if node (else statement)
					ifcond->child[2] = curnode;
					ifcond->child[2]->parent = ifcond;
				}
				getnexttoken = false; //checked to see if else was there, don't need next

				//attach if statement to rest of tree via semicolon
				//put semicolon (statement separator) on top
				ASTNode *tempnode = new ASTNode;
				tempnode->op = T_SEMICOLON;
				tempnode->type = T_VOID;
				tempnode->funccall = NULL;
				//put new expression to right, old tree to left
				tempnode->child[0] = tree;
				tempnode->child[1] = ifcond;
				tempnode->child[2] = NULL;
				if(tempnode->child[0] != NULL)
					tempnode->child[0]->parent = tempnode;
				if(tempnode->child[1] != NULL)
					tempnode->child[1]->parent = tempnode;
				tree = tempnode;
			}
			break;

		case T_WHILE:
			{
				//make sure no qualifiers
				if(curattrib != A_SIGNED)
				{
					constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
					curattrib = A_SIGNED;	//so don't pick up error more than once
				}
				//can't have control statements in parallel blocks
				if(inparallel)
					constrainer->reportError(ERROR_PARALLELCONTROLSTATEMENTS);
				//get left parentheses
				curtoken = scanner->getNextToken();
				if(curtoken == T_ENDOFFILE)
				{
					constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
					return false;
				}
				if(curtoken != T_LPAREN)
					constrainer->reportError(ERROR_MISSINGLEFTPARENTHESES);

				//get expression within while(...)
				curparendepth = 0;
				returnparendepth = 0;
				curbracketdepth = 0;
				curtoken = scanner->getNextToken();
				if(!PFExpression(curnode))
					return false;
				if(curtoken != T_RPAREN)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				else if(curbracketdepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

				//begin building while loop
				ASTNode *wloop = new ASTNode;
				wloop->op = T_WHILE;
				wloop->type = T_VOID;
				wloop->funccall = NULL;
				wloop->child[0] = curnode;
				if(wloop->child[0] != NULL)
					wloop->child[0]->parent = wloop;
				else
					constrainer->reportError(ERROR_EMPTYLANGUAGECONSTRUCT);

				//remember whether or not in loop
				bool loopsave = inloop;
				inloop = true;

				//while loop contains expression
				curnode = NULL;
				PFStatement(curnode, true, false);
				inloop = loopsave;

				//put expression on child[1] of while loop node
				wloop->child[1] = curnode;
				if(wloop->child[1] != NULL)
					wloop->child[1]->parent = wloop;

				wloop->child[2] = NULL;

				//attach while loop to rest of tree via semicolon
				//put semicolon (statement separator) on top
				ASTNode *tempnode = new ASTNode;
				tempnode->op = T_SEMICOLON;
				tempnode->type = T_VOID;
				tempnode->funccall = NULL;
				//put new expression to right, old tree to left
				tempnode->child[0] = tree;
				tempnode->child[1] = wloop;
				tempnode->child[2] = NULL;
				if(tempnode->child[0] != NULL)
					tempnode->child[0]->parent = tempnode;
				if(tempnode->child[1] != NULL)
					tempnode->child[1]->parent = tempnode;
				tree = tempnode;
			}
			break;

		case T_FOR:
			{
				//make sure no qualifiers
				if(curattrib != A_SIGNED)
				{
					constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
					curattrib = A_SIGNED;	//so don't pick up error more than once
				}
				//can't have control statements in parallel blocks
				if(inparallel)
					constrainer->reportError(ERROR_PARALLELCONTROLSTATEMENTS);
				//get left parentheses
				curtoken = scanner->getNextToken();
				if(curtoken == T_ENDOFFILE)
				{
					constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
					return false;
				}
				if(curtoken != T_LPAREN)
					constrainer->reportError(ERROR_MISSINGLEFTPARENTHESES);

				//grab next character after left parentheses
				curtoken = scanner->getNextToken();

				//get first expression within for(...)
				curparendepth = 0;
				returnparendepth = 0;
				curbracketdepth = 0;
				if(!PFExpression(curnode))
					return false;
				
				if(curtoken != T_SEMICOLON)
					constrainer->reportError(ERROR_INVALIDSYMBOL);

				if(curparendepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				if(curbracketdepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

				//put semicolon (statement separator) on top
				ASTNode *fornode = new ASTNode;
				fornode->op = T_SEMICOLON;
				fornode->type = T_VOID;
				fornode->funccall = NULL;
				//put first expression to left, then put a while loop on right
				fornode->child[0] = curnode;
				fornode->child[2] = NULL;	//just in case
				if(fornode->child[0] != NULL)
					fornode->child[0]->parent = fornode;

				//get condition of 'for' loop
				curparendepth = 0;
				returnparendepth = 0;
				curbracketdepth = 0;
				curtoken = scanner->getNextToken();
				if(!PFExpression(curnode))
					return false;
				if(curtoken != T_SEMICOLON)
					constrainer->reportError(ERROR_INVALIDSYMBOL);

				if(curparendepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				if(curbracketdepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

				//begin building while loop
				ASTNode *wloop = new ASTNode;
				wloop->op = T_WHILE;
				wloop->type = T_VOID;
				wloop->funccall = NULL;
				wloop->child[0] = curnode;
				if(wloop->child[0] != NULL)
					wloop->child[0]->parent = wloop;
				else
					constrainer->reportError(ERROR_EMPTYLANGUAGECONSTRUCT);

				fornode->child[1] = wloop;
				if(fornode->child[1] != NULL)
					fornode->child[1]->parent = fornode;

				//get last expression within for(...) (iterator)
				ASTNode *iterator;
				curparendepth = 0;
				returnparendepth = 0;
				curbracketdepth = 0;
				curtoken = scanner->getNextToken();
				if(!PFExpression(iterator))
					return false;
				if(curtoken != T_RPAREN)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				else if(curbracketdepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

				//remember whether or not in loop
				bool loopsave = inloop;
				inloop = true;

				//while loop contains expression
				curnode = NULL;
				PFStatement(curnode, true, false);
				inloop = loopsave;

				//put iterator on end of loop
				ASTNode *forbody = new ASTNode;
				forbody->op = T_SEMICOLON;
				forbody->type = T_VOID;
				forbody->funccall = NULL;
				//put first expression to left, then put a while loop on right
				forbody->child[0] = curnode;
				forbody->child[1] = iterator;
				forbody->child[2] = NULL;	//just in case
				if(forbody->child[0] != NULL)
					forbody->child[0]->parent = forbody;
				if(forbody->child[1] != NULL)
					forbody->child[1]->parent = forbody;

				//put expression on child[1] of while loop node
				wloop->child[1] = forbody;
				if(wloop->child[1] != NULL)
					wloop->child[1]->parent = wloop;

				wloop->child[2] = NULL;

				//attach while loop to rest of tree via semicolon
				//put semicolon (statement separator) on top
				ASTNode *tempnode = new ASTNode;
				tempnode->op = T_SEMICOLON;
				tempnode->type = T_VOID;
				tempnode->funccall = NULL;
				//put new expression to right, old tree to left
				tempnode->child[0] = tree;
				tempnode->child[1] = fornode;
				tempnode->child[2] = NULL;
				if(tempnode->child[0] != NULL)
					tempnode->child[0]->parent = tempnode;
				if(tempnode->child[1] != NULL)
					tempnode->child[1]->parent = tempnode;
				tree = tempnode;
			}
			break;

		case T_RETURN:
			{
				//make sure no qualifiers
				if(curattrib != A_SIGNED)
				{
					constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
					curattrib = A_SIGNED;	//so don't pick up error more than once
				}
				//can't have control statements in parallel blocks
				if(inparallel)
					constrainer->reportError(ERROR_PARALLELCONTROLSTATEMENTS);

				//get expression after return keyword
				curparendepth = 0;
				returnparendepth = 0;
				curbracketdepth = 0;
				curtoken = scanner->getNextToken(); //start in expression
				if(!PFExpression(curnode))
					return false;
				if(curtoken != T_SEMICOLON)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				else if(curparendepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
				else if(curbracketdepth != 0)
					constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

				//put retnode and curnode on tree
				//put return expression on return
				ASTNode *retnode = new ASTNode;

				retnode->op = T_RETURN;

				retnode->type = curfunc->returnType;
				retnode->funccall = NULL;
				retnode->child[0] = curnode;
				retnode->child[1] = NULL;
				retnode->child[2] = NULL;
				if(curnode != NULL)
					curnode->parent = retnode;
				//join retnode and tree with semicolon parent of both
				curnode = new ASTNode;
				curnode->op = T_SEMICOLON;
				curnode->type = T_VOID;
				curnode->funccall = NULL;
				curnode->parent = NULL; //putting on top
				curnode->child[0] = tree;
				curnode->child[1] = retnode;
				curnode->child[2] = NULL;	//just in case
				if(curnode->child[0] != NULL)
					curnode->child[0]->parent = curnode;
				if(curnode->child[1] != NULL)
					curnode->child[1]->parent = curnode;
				tree = curnode;
			}
			break;

		case T_BREAK:
		case T_CONTINUE:
			{
				if(!inloop)
				{
					constrainer->reportError(ERROR_CONTROLNOTINLOOP);
				}
				//can't have control statements in parallel blocks
				if(inparallel)
					constrainer->reportError(ERROR_PARALLELCONTROLSTATEMENTS);

				nexttoken = scanner->getNextToken();
				if(nexttoken != T_SEMICOLON)
					constrainer->reportError(ERROR_ILLEGALOPPERATORUSE);
				ASTNode *breaknode = new ASTNode;
				breaknode->op = curtoken;
				breaknode->type = 0;
				breaknode->funccall = NULL;
				breaknode->child[0] = NULL;
				breaknode->child[1] = NULL;
				breaknode->child[2] = NULL;
				breaknode->parent = NULL;
				//join breaknode and tree with semicolon parent of both
				curnode = new ASTNode;
				curnode->op = T_SEMICOLON;
				curnode->type = T_VOID;
				curnode->funccall = NULL;
				curnode->parent = NULL; //putting on top
				curnode->child[0] = tree;
				curnode->child[1] = breaknode;
				curnode->child[2] = NULL;	//just in case
				if(curnode->child[0] != NULL)
					curnode->child[0]->parent = curnode;
				if(curnode->child[1] != NULL)
					curnode->child[1]->parent = curnode;
				tree = curnode;
			}
			break;

		case T_SEMICOLON:
			//make sure no qualifiers
			if(curattrib != A_SIGNED)
			{
				constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
				curattrib = A_SIGNED;	//so don't pick up error more than once
			}
			break;

		case T_LBRACE:
			//make sure no qualifiers
			if(curattrib != A_SIGNED)
			{
				constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
				curattrib = A_SIGNED;	//so don't pick up error more than once
			}
			//code block
			vardefstack->increaseScope();
			PFStatement(tree, true, true);
			break;

		case T_RBRACE:
			//make sure no qualifiers
			if(curattrib != A_SIGNED)
			{
				constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
				curattrib = A_SIGNED;	//so don't pick up error more than once
			}
			vardefstack->decreaseScope();
			return true;
			break;

		case T_PARALLEL:
			//make sure no qualifiers
			if(curattrib != A_SIGNED)
			{
				constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
				curattrib = A_SIGNED;	//so don't pick up error more than once
			}
			if(!inparallel)
			{
				vardefstack->increaseScope();
				inparallel = true;
				curnode = NULL;
				PFStatement(curnode, true, true);
				ASTNode *pnode; //for parallel op
				//put above tree (T_PARALLEL: child0 = execute first, child1 = parallelize)
				pnode = new ASTNode;
				pnode->op = T_PARALLEL;
				pnode->parent = NULL;
				pnode->type = T_VOID;
				pnode->funccall = NULL;
				pnode->child[0] = tree;
				pnode->child[1] = curnode;
				pnode->child[2] = NULL;	//just in case
				if(pnode->child[0] != NULL)
					pnode->child[0]->parent = pnode;
				if(pnode->child[1] != NULL)
					pnode->child[1]->parent = pnode;
				tree = pnode;

				//outside of block now
				inparallel = false;
				vardefstack->decreaseScope();
			}
			else
			{
				//done with parallel section
				//return up to calling depth
				return true;
			}
			break;

		case T_INVALID:
			constrainer->reportError(ERROR_INVALIDIDENTIFIER);
			break;

		case T_ENDOFFILE:
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;


		//left hand side
		default:
			{
				//make sure no qualifiers
				if(curattrib != A_SIGNED)
				{
					constrainer->reportError(ERROR_PRECEDINGQUALIFIER);
					curattrib = A_SIGNED;	//so don't pick up error more than once
				}
				//check to see if instatiation of user type
				long curclassdefnum = vardefstack->findFirstDef(curtoken);
				if(curclassdefnum
					&& (symboltable->getType(curtoken, curclassdefnum) == T_CLASS || symboltable->getType(curtoken, curclassdefnum) == T_PCLASS) )
				{
					//can't have declarations in parallel blocks
					if(inparallel)
						constrainer->reportError(ERROR_PARALLELCONTROLSTATEMENTS);
					if(!PFUserTypeDefinition(curclassdefnum))
						return false;
				}
				else
				{
					//evaluate expression
					curparendepth = 0;
					returnparendepth = 0;
					curbracketdepth = 0;
					if(!PFExpression(curnode))
						return false;

					if(curtoken == T_ENDOFFILE)
					{
						constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
						return false;
					}
					if(curtoken != T_SEMICOLON)
						constrainer->reportError(ERROR_INVALIDSYMBOL);
					else if(curparendepth != 0)
						constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
					else if(curbracketdepth != 0)
						constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

					//put semicolon (statement separator) on top
					ASTNode *tempnode = new ASTNode;
					tempnode->op = T_SEMICOLON;
					tempnode->type = T_VOID;
					tempnode->funccall = NULL;
					//put new expression to right, old tree to left
					tempnode->child[0] = tree;
					tempnode->child[1] = curnode;
					tempnode->child[2] = NULL;	//just in case
					if(tempnode->child[0] != NULL)
						tempnode->child[0]->parent = tempnode;
					if(tempnode->child[1] != NULL)
						tempnode->child[1]->parent = tempnode;
					tree = tempnode;
				} //else evaluate expression
			} //token not in list (default case)
			break;
		} //switch curtoken
	} while(mult_statements); //for all

	return true;
}


bool Parser::PFTypeDefinition(ASTNode *&tree)
{
	//save type
	curattrib |= curtoken;

	//get name
	curtoken = scanner->getNextToken();

	//make sure not error
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
	{
		constrainer->reportError(ERROR_INVALIDIDENTIFIER);
		return true;	//continue on parsing
	}

	symbol_already_exists = vardefstack->ExistInCurScope(curtoken);

	switch(nexttoken = scanner->getNextToken())
	{
	//************
	//variable declaration
	case T_SEMICOLON:
		if(!PFVarDefinition())
			return false;
		break;

	//***********
	//initialized variable
	case T_GETS:
		{
			//add variable
			if(!PFVarDefinition())
				return false;

			//put token on top of stack, then call expression
			ASTNode *curnode = new ASTNode;
			curnode->parent = NULL;
			curnode->funccall = NULL;
			curnode->op = curtoken;
			curnode->defnum = vardefstack->findFirstDef(curtoken);
			curnode->type = symboltable->getType(curnode->op, curnode->defnum);
			curnode->child[0] = NULL;
			curnode->child[1] = NULL;	//just in case
			curnode->child[2] = NULL;	//just in case
			varstack->pushOp(curnode);

			curparendepth = 0;
			returnparendepth = 0;
			curbracketdepth = 0;
			curtoken = T_GETS;	//set to current operator
			if(!PFExpression(curnode))
				return false;

			if(curtoken != T_SEMICOLON)
				constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
			else if(curparendepth != 0)
				constrainer->reportError(ERROR_UNBALANCEDPARENTHESES);
			else if(curbracketdepth != 0)
				constrainer->reportError(ERROR_UNBALANCEDBRACKETS);

			//put semicolon (statement separator) on top
			ASTNode *tempnode = new ASTNode;
			tempnode->op = T_SEMICOLON;
			tempnode->type = T_VOID;
			tempnode->funccall = NULL;
			//put new expression to right, old tree to left
			tempnode->child[0] = tree;
			tempnode->child[1] = curnode;
			tempnode->child[2] = NULL;	//just in case
			if(tempnode->child[0] != NULL)
				tempnode->child[0]->parent = tempnode;
			tempnode->child[1]->parent = tempnode;
			tree = tempnode;
		}
		break;

	//**********
	//array
	case T_LBRACKET:
		if(!PFArrayDefinition())
			return false;
		break; //array

	case T_ENDOFFILE:
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;

	default:
		constrainer->reportError(ERROR_INVALIDSYMBOL);
		break;
	}

	//return to default
	curattrib = A_SIGNED;
	return true;
}

bool Parser::PFVarDefinition()
{
	if(symbol_already_exists)
		constrainer->reportError(ERROR_SYMBOLDEFINED);

	//make sure not a constant (because constants should be initialized
	if(curattrib & A_CONST)
		constrainer->reportError(ERROR_CONSTANTNOTINIT);

	long defnum = symboltable->getNextDefNum(curtoken);
	vardefstack->pushVarDef(curtoken, defnum);
	symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
		curattrib, 0, curfunc->curvarpos, (unsigned char *)"\0\0\0\0\0\0\0\0", functid);
	curfunc->curvarpos -= symboltable->getSize(curtoken, defnum);

	return true;
}

bool Parser::PFArrayDefinition()
{
	curattrib |= A_ARRAY;

	if(symbol_already_exists)
	{
		constrainer->reportError(ERROR_SYMBOLDEFINED);
	}

	long defnum = symboltable->getNextDefNum(curtoken);
	vardefstack->pushVarDef(curtoken, defnum);
	symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
		curattrib, 0, curfunc->curvarpos, (unsigned char *)"\0\0\0\0\0\0\0\0", functid);

	nexttoken = scanner->getNextToken();
	long arraylength;

	if(!(nexttoken & T_IMMEDIATE))
		constrainer->reportError(ERROR_CONSTANTREQUIRED);

	//remove immediate flag
	nexttoken &= (~(TOKEN_ID)T_IMMEDIATE);

	//check for type
	if(nexttoken != T_CHAR && nexttoken != T_SHORT && nexttoken != T_LONG)
		constrainer->reportError(ERROR_INVALIDTYPE);

	//get length of array
	arraylength = scanner->getImmediateLong();

	//get trailing bracket
	nexttoken = scanner->getNextToken();
	//make sure not error
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	if(nexttoken != T_RBRACKET)
		constrainer->reportError(ERROR_BRACKETREQUIRED);

	nexttoken = scanner->getNextToken();
	//make sure semicolon
	if(nexttoken != T_SEMICOLON)
		constrainer->reportError(ERROR_INVALIDSYMBOL);

	if(arraylength != 0)
	{
		long arraysize = arraylength * symboltable->getSize(curtoken, defnum);
		symboltable->setSize(curtoken, defnum, arraysize);
		curfunc->curvarpos -= arraysize;
	}
	else //resizeable -save only pointer
	{
		symboltable->setType(curtoken, defnum, symboltable->getType(curtoken, defnum) | A_RESIZEABLE);
		curfunc->curvarpos -= SIZE_LONG;	//size of pointer
	}

	return true;
}


bool Parser::PFUserTypeDefinition(long curclassdefnum)
{

	//instantiation

	//remember current class
	TOKEN_ID curclass = curtoken;
	curtoken = scanner->getNextToken();

	//make sure not error
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
	{
		constrainer->reportError(ERROR_INVALIDIDENTIFIER);
		return true;	//continue on parsing
	}

	//check to make sure exists
	if(vardefstack->ExistInCurScope(curtoken))
	{
		constrainer->reportError(ERROR_SYMBOLDEFINED);
	}

	//make sure terminated right
	nexttoken = scanner->getNextToken();
	
	//check to see if array
	if(nexttoken == T_LBRACKET)
	{
		long arraylength;
		nexttoken = scanner->getNextToken();
		if(nexttoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}
		else if(!(nexttoken & T_IMMEDIATE))
			constrainer->reportError(ERROR_CONSTANTREQUIRED);

		//remove immediate flag
		nexttoken &= (~(TOKEN_ID)T_IMMEDIATE);

		//check for type
		if(nexttoken != T_CHAR && nexttoken != T_SHORT && nexttoken != T_LONG)
			constrainer->reportError(ERROR_INVALIDTYPE);

		//get length of array
		arraylength = scanner->getImmediateLong();

		//get trailing bracket
		nexttoken = scanner->getNextToken();
		//make sure not error
		if(nexttoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}
		if(nexttoken != T_RBRACKET)
			constrainer->reportError(ERROR_BRACKETREQUIRED);

		//add array to symbol table
		long defnum = symboltable->getNextDefNum(curtoken);
		vardefstack->pushVarDef(curtoken, defnum);
		symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
			(arraylength != 0) ? (A_ARRAY | A_USERTYPE) : (A_ARRAY | A_RESIZEABLE),
			curclass, curfunc->curvarpos, (unsigned char *)"\0\0\0\0\0\0\0\0", functid);

		//set size of class
		symboltable->setSize(curtoken, defnum, (arraylength != 0) ? (arraylength * symboltable->getSize(curclass, curclassdefnum)) : SIZE_LONG);
		curfunc->curvarpos -= symboltable->getSize(curtoken, defnum);

		nexttoken = scanner->getNextToken();
	}
	else //non-array declaration
	{
		//add variable to symbol table
		long defnum = symboltable->getNextDefNum(curtoken);
		vardefstack->pushVarDef(curtoken, defnum);
		symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
			A_USERTYPE, curclass, curfunc->curvarpos, (unsigned char *)"\0\0\0\0\0\0\0\0", functid);
		curfunc->curvarpos -= symboltable->getSize(curclass, symboltable->getClassDefnum(curclass));
	}

	if(nexttoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	else if(nexttoken != T_SEMICOLON)
		constrainer->reportError(ERROR_INVALIDSYMBOL);

	//return to default
	curattrib = A_SIGNED;

	return true;
}


bool Parser::PFExpression(ASTNode *&tree)
{
	ASTNode *curnode;
	prevtoken = T_SEMICOLON;	//just states that last expression not included

	//curtoken = scanner->getNextToken(); -called with curtoken set
	for(;;)	//while in expression
	{
		//must get next token sometime before loop (in each branch)

		if(curtoken & T_IMMEDIATE) //constant
		{
			curnode = new ASTNode;
			curnode->parent = NULL;
			curnode->op = (curtoken & A_BASETYPES);
			curnode->defnum = 0;
			curnode->type = (curtoken & A_BASETYPES);
			curnode->funccall = NULL;
			curnode->child[0] = NULL;
			curnode->child[1] = NULL;	//just in case
			curnode->child[2] = NULL;	//just in case

			switch(curtoken & A_BASETYPES)
			{
			case T_CHAR:
				((char *)curnode->data)[0] = (char)scanner->getImmediateLong();
				break;
			case T_LONG:
				((long *)curnode->data)[0] = scanner->getImmediateLong();
				break;
			case T_DOUBLE:
				((double *)curnode->data)[0] = scanner->getImmediateDouble();
				break;
			case T_IMMEDIATE_STR:
				((char **)curnode->data)[0] = scanner->getImmediateString();
				break;
			default:
				constrainer->reportError(ERROR_INVALIDIMMEDIATE);
				break;
			}

			varstack->pushOp(curnode);

			//remember last token
			prevtoken = curtoken;
			curtoken = scanner->getNextToken();
		}
		else if(curtoken >= T_BUILTINFUNCSTART)
		{
			TOKEN_ID scopeparent;
			//lookahead
			nexttoken = scanner->getNextToken();

			curnode = new ASTNode;
			curnode->parent = NULL;
			curnode->funccall = NULL;
			curnode->op = curtoken;
			curnode->child[0] = NULL;
			curnode->child[1] = NULL;	//just in case
			curnode->child[2] = NULL;	//just in case


			scopeparent = curparent;	//start from wherever (in a member function, global, etc.)
			if(prevtoken == T_MEMBEROF)
			{
				//must set scopeparent to whatever previous parent was
				//if memberof, go to right child
				//if array, go to left child

				ASTNode *scopenode = varstack->peekOp();
				while(scopenode != NULL && scopenode->op < T_IDENTIFIERSTART)
			//		&& (scopenode->type & A_USERTYPE) != A_USERTYPE)
				{
					if(scopenode->op == T_MEMBEROF)
						scopenode = scopenode->child[1];
					else if(scopenode->op == T_LBRACKET)
						scopenode = scopenode->child[0];
					else
					{
						constrainer->reportError(ERROR_ILLEGALMEMBEROF);
						break;
					}
				}
				if(scopenode == NULL)
					constrainer->reportError(ERROR_ILLEGALMEMBEROF);

				scopeparent = symboltable->getUserType(scopenode->op, scopenode->defnum); //type contains TOKEN_ID of [p]class

				curnode->defnum = symboltable->getDefnum(curnode->op, scopeparent);
				if(curnode->defnum == INVALIDDEFNUM)
				{
					constrainer->reportError(ERROR_UNDEFINEDIDENTIFIER);
					return true;
				}

				//must check for variables attemted to be accessed from a pclass
				if(nexttoken != T_LPAREN
						&& symboltable->getType(scopeparent, symboltable->getClassDefnum(scopeparent)) == T_PCLASS)
					constrainer->reportError(ERROR_PCLASSATTRIBUTEACCESS);
			}
			else if(nexttoken != T_LPAREN) //global scope, not a function
			{
				//get defnum
				curnode->defnum = vardefstack->findFirstDef(curtoken);
				if(curnode->defnum == INVALIDDEFNUM || curnode->defnum == 0)
				{
					//check to see if parsing a member function of a class
					//and variable is a member of that class
					if(curparent != GLOBALSCOPE)
					{
						curnode->defnum = symboltable->getDefnumOfMemberVariable(curtoken, curparent);
					}

					if(curnode->defnum == INVALIDDEFNUM || curnode->defnum == 0)
					{
						constrainer->reportError(ERROR_UNDEFINEDIDENTIFIER);
						return true;
					}
				}

			}

			//scopeparent contains TOKEN_ID of class of parent of type

			//see if special case (funciton call, array reference, etc.)
			if(nexttoken == T_LPAREN) //function
			{
				curnode->funccall = new ASTFuncCall;
				curnode->funccall->op = curnode->op;	//move function id to call
				curnode->op = T_FUNCTIONCALL;			//designate as function call

				//adjust parenthese depth
				curparendepth++;

				long parendepthsave = returnparendepth;
				returnparendepth = curparendepth;

				//get function parameters
				for(curnode->funccall->numparameters = 0; true; /*curnode->funccall->numparameters++*/)
				{
					prevtoken = curtoken;
					curtoken = scanner->getNextToken();

					//incase no parameters
					if(curtoken == T_RPAREN)
					{
						curparendepth--;
						break;
					}

					//get next parameter and store in function
					if(!PFExpression(curnode->funccall->parameters[curnode->funccall->numparameters]))
						return false;

					//if an expression exists, count it
					if(curnode->funccall->parameters[curnode->funccall->numparameters] != NULL)
						curnode->funccall->numparameters++;

					if(curtoken == T_ENDOFFILE)
					{
						constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
						return false;
					}
					
					if(curtoken == T_COMMA)
						continue;

					if(curtoken == T_RPAREN)
						break;

					constrainer->reportError(ERROR_INVALIDSYMBOL);
				}

				//put back to normal
				returnparendepth = parendepthsave;

				//convert ASTFunctionCall to FunctionType to find correct function match
				FunctionType ft;
				ft.numarguements = curnode->funccall->numparameters;
				//copy parameters over to FunctionType ft
				for(int i = 0, j = 0; i < curnode->funccall->numparameters; i++, j++)
				{
					if(curnode->funccall->parameters[i]->type != A_USERTYPE)
						ft.arguements[j] = curnode->funccall->parameters[i]->type;
					else //need to write A_USERTYPE and write id of type
					{
						ft.arguements[j++] = A_USERTYPE;
						ft.arguements[j] = symboltable->getUserType(curnode->funccall->parameters[i]->op, curnode->funccall->parameters[i]->defnum);
						ft.numarguements++; //must account for double parameter
					}
				}

				//ft contains parameter types, curnode->funccall->op contains function's ID

				long funcdefnum;
				//find from within specified class
				if(scopeparent != curparent)
				{
					funcdefnum = symboltable->getClosestFuncDefnum(curnode->funccall->op, &ft, scopeparent);
					if(symboltable->getType(scopeparent, symboltable->getClassDefnum(scopeparent)) == T_PCLASS)
						curnode->op = T_PFUNCTIONCALL;

					if(funcdefnum == INVALIDDEFNUM)
						constrainer->reportError(ERROR_UNDEFINEDFUNCTION);
				}
				else //first check parent class if exist, then global
				{
					funcdefnum = symboltable->getClosestFuncDefnum(curnode->funccall->op, &ft, curparent);
					if(funcdefnum == INVALIDDEFNUM)
					{
						funcdefnum = symboltable->getClosestFuncDefnum(curnode->funccall->op, &ft, GLOBALSCOPE);
						//if still haven't found
						if(funcdefnum == INVALIDDEFNUM)
							constrainer->reportError(ERROR_UNDEFINEDFUNCTION);
					}
				}

				curnode->funccall->defnum = funcdefnum;

				//set return type
				if(funcdefnum != INVALIDDEFNUM)
				{
					FunctionType *f = symboltable->getFunctionType(curnode->funccall->op, curnode->funccall->defnum);
					curnode->type = f->returnType;
				}

				varstack->pushOp(curnode);

				//remember last token
				prevtoken = curtoken;
				curtoken = scanner->getNextToken();
			}
			else if(nexttoken == T_LBRACKET) //array index
			{
				//set type
				curnode->type = symboltable->getType(curtoken, curnode->defnum);

				//make sure array type
				if(!(curnode->type & A_ARRAY) && !(curnode->type & A_RESIZEABLE))
					constrainer->reportError(ERROR_INDEXEDNONARRAYTYPE);

				//put array identifier on stack
				varstack->pushOp(curnode);

				//build tree node
				curnode = new ASTNode;
				curnode->parent = NULL;
				curnode->funccall = NULL;
				curnode->op = nexttoken;
				curnode->defnum = 0;
				curnode->child[0] = NULL;
				curnode->child[1] = NULL;
				curnode->child[2] = NULL;
				curnode->child[3] = NULL;

				opstack->pushOp(curnode, curparendepth, curbracketdepth);
				//adjust bracket depth
				curbracketdepth++;

				prevtoken = nexttoken;
				curtoken = scanner->getNextToken();
			}
			else //regular variable
			{
				//set type
				curnode->type = symboltable->getType(curtoken, curnode->defnum);
				
				varstack->pushOp(curnode);

				//remember last token
				prevtoken = curtoken;
				curtoken = nexttoken;	//update
			}
		}
		else if(curtoken == T_INVALID)
		{
			constrainer->reportError(ERROR_INVALIDSYMBOL);
			curtoken = scanner->getNextToken();
		}
		else if(curtoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}
		else if(curtoken == T_SEMICOLON)
		{
			//finish setting up tree
			PFCondenseStack(curtoken);
			tree = varstack->popOp();
			return true;
		}
		else if(curtoken == T_COMMA)
		{
			//finish setting up tree
			PFCondenseStack(curtoken);
			tree = varstack->popOp();
			return true;
		}
		else if(curtoken == T_LPAREN)
		{
			//adjust parenthese depth
			curparendepth++;
			prevtoken = curtoken;
			curtoken = scanner->getNextToken();
		}
		else if(curtoken == T_RPAREN)
		{
			//condense stack within last set of parentheses, then decrease depth
			PFCondenseStack(curtoken);
			curparendepth--;
			//check to see if exiting expression -useful for controls and functions
			if(curparendepth < returnparendepth)
			{
				tree = varstack->popOp();
				return true;
			}
			prevtoken = curtoken;
			curtoken = scanner->getNextToken();
		}
		else if(curtoken == T_RBRACKET)
		{
			if(prevtoken == T_LBRACKET)
				constrainer->reportError(ERROR_MISMATCHEDBRACKETS);
			//condense stack within last set of parentheses, then decrease depth
			PFCondenseStack(curtoken);
			curbracketdepth--;
			//check to see if exiting expression -useful for controls and functions
			if(curparendepth < 0)
				constrainer->reportError(ERROR_MISMATCHEDBRACKETS);

			//remember last token
			prevtoken = curtoken;
			curtoken = scanner->getNextToken();

		}
		else //operator
		{
			//build tree node
			curnode = new ASTNode;
			curnode->parent = NULL;
			curnode->op = curtoken;
			curnode->funccall = NULL;
			curnode->defnum = 0;
			curnode->child[0] = NULL;
			curnode->child[1] = NULL;
			curnode->child[2] = NULL;
			curnode->child[3] = NULL;

			//check plus and minus to see if binary or unary
			if(prevtoken < T_BUILTINFUNCSTART && !(prevtoken & T_IMMEDIATE)
				&& prevtoken != T_RPAREN && prevtoken != T_RBRACKET)
			{
				if(curnode->op == T_PLUS)
					curnode->op = T_UNARYPLUS;
				else if(curnode->op == T_MINUS)
					curnode->op = T_UNARYMINUS;
			}

			PFCondenseStack(curnode->op);

			opstack->pushOp(curnode, curparendepth, curbracketdepth);

			//remember last token
			prevtoken = curtoken;
			curtoken = scanner->getNextToken();
		}

	} //for all of expression

	return true;
}

//function converts everything in the tree that is x.func(...) to 
//func(..., x)
void Parser::PFConvertMemberFunctionCalls(ASTNode *tree)
{
	if(tree == NULL)
		return;

	if(tree->op != T_MEMBEROF)
	{
		PFConvertMemberFunctionCalls(tree->child[0]);		
		PFConvertMemberFunctionCalls(tree->child[1]);
		PFConvertMemberFunctionCalls(tree->child[2]);
		return;
	}

	//check parent to see if member of opperator
	if(tree->child[1] == NULL || tree->child[1]->op != T_FUNCTIONCALL)
	{
		PFConvertMemberFunctionCalls(tree->child[0]);		
		PFConvertMemberFunctionCalls(tree->child[1]);
		PFConvertMemberFunctionCalls(tree->child[2]);
		return;
	}

	//conversion case (member function)

	//cut function call out
	ASTNode *func = tree->child[1];
	tree->child[1] = NULL;

	//now have redundant memberof operator -remove
	ASTNode *temp = tree->child[1], *parenttemp;
	parenttemp = tree->parent;
	memcpy(tree, tree->child[0], sizeof(ASTNode));
	tree->parent = parenttemp;
	delete temp;

	//now find where it should be placed
	ASTNode *curnode = tree;
	while(curnode->parent != NULL && (curnode->parent->op == T_MEMBEROF || curnode->parent->op == T_LBRACKET))
	{
		curnode = curnode->parent;
	}

	//check to make sure nothing weird happened
	if(curnode->parent == NULL)
		return; //something weird happened and shouldn't have

	//curnode->parent is where class pointer arithmetic begins
	int childnum;
	for(childnum = 0; childnum < 4; childnum++)
		if(curnode->parent->child[childnum] == curnode)
			break;

	//put function in place of pointer arithmetic
	temp = curnode;
	curnode = curnode->parent;
	curnode->child[childnum] = func;
	func->parent = curnode;

	//put pointer arithmetic as last parameter to function call
	func->funccall->parameters[func->funccall->numparameters++] = temp;
	temp->parent = func;
}

void Parser::PFCondenseStack(TOKEN_ID compairtoken)
{

//************add warning checking
	ASTNode *curnode;
	long minprecedence = getPrecedence(compairtoken);
	
	while(opstack->peekOp() != NULL
		&& curparendepth == opstack->peekParenDepth() && curbracketdepth == opstack->peekBracketDepth()
		&& minprecedence < getPrecedence(opstack->peekOp()->op))
	{
		curnode = opstack->popOp();

		switch(curnode->op)
		{

		//unary ops
		case T_BNOT:	case T_ENOT:
		case T_UNARYPLUS:	case T_UNARYMINUS:
			curnode->child[0] = varstack->popOp();
			if(curnode->child[0] != NULL)
				curnode->child[0]->parent = curnode;
			if(curnode->op != T_ENOT)
				curnode->type = curnode->child[0]->type;	//propagate type upwards
			else
				curnode->type = T_BOOL;	//expression not should be boolean
			varstack->pushOp(curnode);
			break;

		//binary ops
		case T_PLUS:	case T_MINUS:	case T_MULTIPLY:	case T_DIVIDE:
		case T_MODULUS:	case T_SHIFTLEFT:	case T_SHIFTRIGHT:
		case T_BXOR:	case T_BOR:	case T_BAND:
		case T_EQUALS:	case T_LESS:	case T_LEQUAL:	case T_GREATER:
		case T_GEQUAL:	case T_NOTEQUAL:
		case T_EOR:	case T_EAND:
		case T_GETS:	case T_ARRAYRESIZE:
		case T_MEMBEROF:	case T_LBRACKET:
			curnode->child[1] = varstack->popOp();
			if(curnode->child[1] != NULL)
				curnode->child[1]->parent = curnode;
			else
			{
				constrainer->reportError(ERROR_ILLEGALOPPERATORUSE);
				break;
			}
			curnode->child[0] = varstack->popOp();
			if(curnode->child[0] != NULL)
				curnode->child[0]->parent = curnode;
			else
			{
				constrainer->reportError(ERROR_ILLEGALOPPERATORUSE);
				break;
			}

			//choose evaluation type -propagate upwards
			switch(curnode->op)
			{
			case T_SHIFTLEFT:	case T_SHIFTRIGHT:
				if((curnode->child[1]->op & A_BASETYPES) > T_LONG)
					constrainer->reportError(ERROR_INVALIDTYPE);
			case T_MODULUS:
			case T_BXOR:	case T_BOR:	case T_BAND:
				if(!( (curnode->child[0]->type & A_BASETYPES) <= T_LONG
						&& (curnode->child[1]->type & A_BASETYPES) <= T_LONG))
					constrainer->reportError(ERROR_INVALIDTYPE);
			case T_PLUS:	case T_MINUS:	case T_MULTIPLY:	case T_DIVIDE:
				curnode->type = getGeneralType(curnode->child[0]->type, curnode->child[1]->type);
				break;

			case T_EQUALS:	case T_LESS:	case T_LEQUAL:	case T_GREATER:
			case T_GEQUAL:	case T_NOTEQUAL:
			case T_EOR:	case T_EAND:
				curnode->type = T_BOOL;
				break;

			case T_GETS:
				curnode->type = curnode->child[0]->type;	//choose type of what assigning to
				break;

			case T_LBRACKET:
				//make sure array index is of an integer type
				if(curnode->child[0] != NULL && curnode->child[1] != NULL)
				{
					curnode->type = symboltable->getType(curnode->child[0]->op, curnode->child[0]->defnum);

					ATTRIB_TYPE cnodetype = (curnode->child[1]->type & A_BASETYPES);
					if(cnodetype != T_CHAR && cnodetype != T_SHORT && cnodetype != T_LONG)
						constrainer->reportError(ERROR_NONINTEGERARRAYINDEX);
				}
				break;

			case T_ARRAYRESIZE:
				if(!(curnode->child[1]->type <= T_LONG))
					constrainer->reportError(ERROR_INVALIDTYPE);
				curnode->type = curnode->child[0]->type;
				break;

			default:
				curnode->type = T_LONG;
				break;
			}

			varstack->pushOp(curnode);
			break;

		default:
			//somethin' goofy happened
			constrainer->reportError(ERROR_ILLEGALUSEOFSYMBOL);
			break;

		} //switch top of stack

		if(curnode->type == T_IMMEDIATE_STR)
			constrainer->reportError(ERROR_ILLEGALUSEOFSTRING);

	} //while condensing stack
}

