
/////////////////////
//String table

//total number of characters accepted
#define ST_NUM_ACCEPT_CHARS 63

struct StringTrieNode
{
	StringTrieNode *list[63];
	TOKEN_ID id;
	StringTrieNode *parent;
};

//controls symbol table, and string table
class StringTable
{
public:
	StringTable();

	~StringTable();

	//////////////
	//string table
	//adds string if does not exist, returns -1 if invalid
	TOKEN_ID getId(const char *str);

	//returns string from id
	StringTrieNode *getEndNameofLastGet();
	char *getString(StringTrieNode *endname);

	void addReservedWord(const char *str, TOKEN_ID id);
	
private:
	//returns letter position in list, -1 if not in list
	int letterPosition(const char c);

	//deletes node and children
	void DeleteNode(StringTrieNode *n);

	StringTrieNode *root;
	StringTrieNode *lastentry;
	TOKEN_ID nextId;
};

/////////////////////
//Symbol table

//extra info needed for functions in symbol table
struct FunctionType
{
	ATTRIB_TYPE returnType;
	ATTRIB_TYPE arguements[MAXPARAMETERS];	//user types take up 2 arguements, one for A_USERTYPE,
											//and the one following for the TOKEN_ID of the usertype
	int numarguements;

	int curvarpos;					//stack size for function variables
									//goes down from -SIZE_LONG (with respect to stack)

	int parameterlength;			//stack size for entire parameter list (includes return address)

	//abstract syntax tree of function
	ASTNode *function;
	MIRInstr *mircode;

	long enumeration;	//number of function (used for calling remotely)
};

#define IN_STACK -1

struct SymbolTableNode
{
	TOKEN_ID id;
	long defnum;	//definition number for use with different scopes and depths -first def is 1

	SymbolTableNode *next;	//next in hash table

	ATTRIB_TYPE type;
	TOKEN_ID usertype;			//if type == A_USERTYPE, then contains id of symbol
	ATTRIB_ADDR address;		//heap address for destination program, IN_STACK if not global
								//if member of class, then offset of variable into class
	unsigned char initval[8];	//initial value for global values
								//if IN_STACK, position in stack
								//if constant, then value of constant
	long size;		//size of symbol (if class, keep adding size, if instantiation, set to class size)
					//if array, length of array

	StringTrieNode *nameend;	//traverse string tree upward for name
	TOKEN_ID parent;		//used to designate class belongs to
	FunctionType *functype;	//used for information about functions (parameters, return, ast, etc.)
};


class SymbolTable
{
public:
	SymbolTable();
	~SymbolTable();
	void DeleteNode(ASTNode *n); //used for deleting AST for function entries

	void addSymbol(TOKEN_ID id, long defnum, StringTrieNode *nameend, ATTRIB_TYPE type, TOKEN_ID usertype, ATTRIB_ADDR address, unsigned char *initval, TOKEN_ID parent);
	bool isSymbolDefined(TOKEN_ID id, long defnum);
	long getNextDefNum(TOKEN_ID id);	//get next usable defnum (for newer identifier)

	//set attributes of symbol
	void setType(TOKEN_ID id, long defnum, ATTRIB_TYPE type);
	void setUserType(TOKEN_ID id, long defnum, TOKEN_ID usertype);
	void setAddress(TOKEN_ID id, long defnum, ATTRIB_ADDR addr);
	void setInitVal(TOKEN_ID id, long defnum, unsigned char *initval, ATTRIB_TYPE type);
	void setParent(TOKEN_ID id, long defnum, TOKEN_ID parent);
	void setFunctionType(TOKEN_ID id, long defnum, FunctionType *f);
	void setSize(TOKEN_ID id, long defnum, long size);	//useful for arrays
	void addSize(TOKEN_ID id, long defnum, long size);	//useful for classse

	//get attributes of symbol
	ATTRIB_TYPE getType(TOKEN_ID id, long defnum);
	TOKEN_ID getUserType(TOKEN_ID id, long defnum);
	ATTRIB_ADDR getAddress(TOKEN_ID id, long defnum);
	unsigned char *getInitVal(TOKEN_ID id, long defnum);
	TOKEN_ID getParent(TOKEN_ID id, long defnum);
	StringTrieNode *getNameLookup(TOKEN_ID id, long defnum);
	FunctionType *getFunctionType(TOKEN_ID id, long defnum);
	long getSize(TOKEN_ID id, long defnum);
	long getDefnum(TOKEN_ID id, TOKEN_ID parent);	//useful for getting IDs of functions of a class or global
	long getMatchingFuncDefnum(TOKEN_ID id, FunctionType* f, TOKEN_ID parent);	//returns defnum, otherwise 0 if none
	long getClosestFuncDefnum(TOKEN_ID id, FunctionType* f, TOKEN_ID parent);	//returns defnum, otherwise 0 if none
	long getClassDefnum(TOKEN_ID id);	//gets defnum for class with respective id
	long getDefnumOfMemberVariable(TOKEN_ID id, TOKEN_ID class_id);	//gets defnum of the variable in the class

	//for fast access
	SymbolTableNode *getSymbolTableEntry(TOKEN_ID id, long defnum);

	//for getting functions
	FunctionType *getFirstFunction();
	FunctionType *getNextFunction();
	TOKEN_ID getCurFunctionID();
	long getCurFunctionDefnum();

	//for getting globals
	void resetGlobalSelection();	//resets selection -call selectNextGlobal afterwards
	bool selectNextGlobal();	//returns true if exists
	TOKEN_ID getCurGlobalID();
	long getCurGlobalDefnum();

private:
	SymbolTableNode *findNode(TOKEN_ID id, long defnum);

	//actual table
	SymbolTableNode *hashTable[SYMBOL_TABLE_HASH_SIZE];

	//used for getNextFunction()
	int CurFuncHashPos;
	SymbolTableNode *CurFuncNode;

	//used for getNextGlobal()
	int CurGlobalHashPos;
	SymbolTableNode *CurGlobalNode;	

};

