#define TOKEN_ID long

//////////////////////////
//reserved tokens
//take up ID's 0->0x200
//////////////////////////

#define MAX_INCLUDE_RECURSION 200

enum Token
{
	//token			number	representation
	T_INVALID		= 0xFFFFFFFF,	//invalid token
	T_RECURSEINCLUDE= 0xFFFFFFFE,	//recursive includes
	
	T_ENDOFFILE		= 0x00,			//end of source file

	T_IMMEDIATE		= 0x80000000,	//flag for immediate value

	T_IDENTIFIERSTART = 0x200,		//where identifiers start
									//(after all stuff built into language)
	T_BUILTINFUNCSTART = 0x100,		//where built in functions begin

	T_TYPE_MAX = 0x08,				//maximum value a type can have (currently T_ADDRESS)

	//data types
	T_CHAR			= 0x01,	//char
	T_SHORT			= 0x02,	//short
	T_LONG			= 0x03,	//long
	T_FLOAT			= 0x04,	//float
	T_DOUBLE		= 0x05,	//double
	T_BOOL			= 0x01,	//bool		//same as char
	T_INT			= 0x03,	//int		//same as long
	T_VOID			= 0x06,	//void
	T_IMMEDIATE_STR	= 0x07,	//"..."		//immediate string type
	T_ADDRESS		= 0x08,	//			//used in MIR

	//type modifiers
	T_CONST			= 0x10,	//const
	T_UNSIGNED		= 0x11,	//unsigned

	//class
	T_CLASS			= 0x20,	//class
	T_PCLASS		= 0x21,	//pclass
	T_MEMBEROF		= 0x22,	// .
	T_SCOPE_RES		= 0x23,	// ::

	//program control
	T_IF			= 0x30,	//if(
	T_ELSE			= 0x31,	//else
	T_WHILE			= 0x32,	//while(
	T_BREAK			= 0x33,	//break;
	T_CONTINUE		= 0x34,	//continue
	T_FOR			= 0x35,	//for(
	T_RETURN		= 0x40,	//return
	T_GOTO			= 0x41,	//goto		//used for MIR
	T_PRETURN		= 0x42,	//preturn	//parallel return -used for MIR

	//arithmatic
	T_PLUS			= 0x50,	// +
	T_MINUS			= 0x51,	// -
	T_MULTIPLY		= 0x52,	// *
	T_DIVIDE		= 0x53,	// /
	T_MODULUS		= 0x54,	// %
	T_SHIFTLEFT		= 0x55, // <<
	T_SHIFTRIGHT	= 0x56,	// >>
	T_UNARYPLUS		= 0x57,	// +
	T_UNARYMINUS	= 0x58,	// -

	//bitwise
	T_BXOR			= 0x60,	// ^
	T_BOR			= 0x61,	// |
	T_BAND			= 0x62,	// &
	T_BNOT			= 0x63,	// ~

	//comparison
	T_EQUALS		= 0x70,	// ==
	T_LESS			= 0x71,	// <
	T_LEQUAL		= 0x72,	// <=
	T_GREATER		= 0x73,	// >
	T_GEQUAL		= 0x74,	// >=
	T_NOTEQUAL		= 0x75,	// !=

	//expression
	T_EOR			= 0x80,	// ||
	T_EAND			= 0x81,	// &&
	T_ENOT			= 0x82,	// !
	T_LPAREN		= 0x83,	// (
	T_RPAREN		= 0x84,	// )

	//statement
	T_SEMICOLON		= 0x90,	// ;
	T_COMMA			= 0x91,	// ,
	T_FUNCTIONCALL	= 0x92,	// funcname(...)
	T_PFUNCTIONCALL	= 0x93,	// funcname(...) execute parallel call
	T_PCFUNCTIONCALL= 0x94,	// xxx.funcname(...) execute parallel class call
	T_WAITRETURN	= 0x95,	// waits for the return value of a parallel call
	T_PUSHPARAMETER	= 0x96,	// push  (currently only used by MIR for calling)
	T_PULLPARAMETER	= 0x97,	// pull	 (currently only used by MIR for getting parameters)


	//blocks
	T_LBRACE		= 0xA0,	// {
	T_RBRACE		= 0xA1,	// }
	T_LABEL			= 0xA2, // name: (currently only used by MIR)

	//parallel
	T_PARALLEL		= 0xB0,	// @
	
	//asignment
	T_GETS			= 0xC0,	// =

	//arrays
	T_ARRAYRESIZE	= 0xD0, // :=
	T_LBRACKET		= 0xD1,	// [
	T_RBRACKET		= 0xD2,	// ]
	T_ADDRESSOF		= 0xD3,	// &   (currently only used by MIR)
	T_GLOBALSPACE	= 0xD4,	// converts pointer to global space -used for passing pointer parameters to functions
	T_PUSH_THIS		= 0xD5,	// for calling another member of the same class -pushes itself

	//entry point of program
	T_MAIN			= 0x110,	//main

	//external flag
	T_EXTERN		= 0x120,	//extern

	//often compair values to this to see if immediate
	//if marked with immediate flag, then it will also be less than this
	//if compaired using signed comparisons (because immediates will be negative)
	T_TEMPREG		= 0x1FF		//temporary "register" used in MIR

};

//attribute types
typedef long ATTRIB_TYPE;
//constants for attributes
enum ATTRIB
{
	A_BASETYPES		= 0xFF,		//used for masking out base types
	A_SIGNED		= 0,
	A_UNSIGNED		= 0x01000,	//keep flags above type info
	A_CONST			= 0x02000,
	A_FUNCTION		= 0x04000,
	A_ARRAY			= 0x08000,	//for arrays -init value can store pointer to list of members when initialized
	A_RESIZEABLE	= 0x10000,	//for resizeable arrays
	A_EXTERNAL		= 0x20000,	//for external/imported stuff
	A_USERTYPE		= 0xF0000000 //for user defined classes, pclasses, etc.
};


#define GLOBALSCOPE 0

const int numReservedWords = 23;

//reserved words type
struct ReservedWord
{
	char rw[32];
	TOKEN_ID id;
};

const ReservedWord reservedWords[numReservedWords] =
{
	//string	token
	{"char",	T_CHAR},
	{"short",	T_SHORT},
	{"long",	T_LONG},
	{"float",	T_FLOAT},
	{"double",	T_DOUBLE},
	{"bool",	T_BOOL},	//same as char
	{"int",		T_INT},	//same as long
	{"void",	T_VOID},

	{"const",	T_CONST},
	{"unsigned",T_UNSIGNED},


	{"class",	T_CLASS},
	{"pclass",	T_PCLASS},

	{"if",		T_IF},
	{"else",	T_ELSE},
	{"while",	T_WHILE},
	{"for",		T_FOR},
	{"return",	T_RETURN},
	{"break",	T_BREAK},
	{"continue",T_CONTINUE},

	{"main",	T_MAIN},
	{"extern",	T_EXTERN},
};

